/*
  Copyright 2004, 2005 Jean-Baptiste Note

  This file is part of prism54usb.

  prism54usb is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  prism54usb is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with prism54usb; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

*/

#include "isl_sm.h"
#include "islsm_protocol.h"
#include "wlan-monitor.h"

#include <linux/netdevice.h>

#ifdef MADWIFI
#include <linux/version.h>
#include "net80211/if_media.h"
#include "net80211/ieee80211_var.h"
#endif /* MADWIFI */


static unsigned int islsm_ref_to_chan(unsigned int ref);
static void islsm_data_debug(struct net_device *netdev, void *_data, int len);
static void islsm_input_fault(struct net_device *netdev, struct sk_buff *skb);

/*
 * Data packet input functions
 */

static void
input_data_frame(struct net_device *netdev, struct sk_buff *skb)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct islsm_rx_packet_header *data_head = (struct islsm_rx_packet_header *) (skb->data);
	unsigned int trim_length;
#ifdef MADWIFI
	struct ieee80211com *ic = &islsm->sc_ic;
	struct ieee80211_node *ni;
#endif /* MADWIFI */
	/* we keep this information and put it in the avc header, 
	   where it will be meaningless, but there we'll be able 
	   to see it in ethereal */
	u16 magic1;
	u16 unknown1;
	u16 unknown2;

	islsm_dbg("data packet");

	magic1 = le16_to_cpu(((struct islsm_id *) (skb->data))->magic1);
	unknown1 = le16_to_cpu(data_head->unknown1);
	unknown2 = le16_to_cpu(data_head->unknown2);

	/* problem : I don't know how to report the FCS in monitor
	 * mode ! */
	trim_length = le16_to_cpu( ((struct islsm_id *) skb->data)->length );
	skb_pull(skb,SIZE_RX_HEADER);
	skb_trim(skb,trim_length);
	/* strip FCS. Needed for madwifi. How to deal with it in
	 * monitor mode ? */
	skb_trim(skb,skb->len - 4);
	
	/* catch monitor mode */
	if (
#ifdef MADWIFI
		ic->ic_opmode == IEEE80211_M_MONITOR
#else 
		islsm->iw_mode == IW_MODE_MONITOR
#endif /* MADWIFI */
		)
        {
#if 1
		struct avs_80211_1_header *avs;
		struct sk_buff **skb_ref = &skb;
		/* time scale is at 1MHZ. maybe there's a fractional part */
		u32 clock = le32_to_cpu(data_head->timestamp);
		u16 freq = le16_to_cpu(data_head->freq);
		u8 rssi = data_head->signal_strength;
		
		if (skb_headroom(skb) < sizeof (struct avs_80211_1_header)) {
			struct sk_buff *newskb = skb_copy_expand(skb,
								 sizeof (struct avs_80211_1_header),
								 0, GFP_ATOMIC);
			if (newskb) {
				dev_kfree_skb_irq(skb);
				*skb_ref = newskb;
			} else
			{
				islsm_dbg("dropped skb in monitor mode");
				return;
			}
			/* This behavior is not very subtile... */
		}
		
		/* make room for the new header and fill it. */
		avs = (struct avs_80211_1_header *) skb_push(*skb_ref,
							     sizeof (struct avs_80211_1_header));
			
		avs->version = cpu_to_be32(P80211CAPTURE_VERSION);
		avs->length = cpu_to_be32(sizeof (struct avs_80211_1_header));
		avs->mactime = cpu_to_be64((u64)clock);
		avs->hosttime = cpu_to_be64(jiffies);
		avs->phytype = cpu_to_be32(6);	/* OFDM: 6 for (g), 8 for (a) */
		avs->channel = cpu_to_be32(islsm_ref_to_chan(freq));
		/* Warning : those are given a bogus meaning for the time being ! */
		avs->datarate = cpu_to_be32((u32)unknown1); /* unknown */
		avs->antenna = cpu_to_be32((u32)unknown2);	/* unknown */
		avs->priority = cpu_to_be32((u32)magic1);	/* unknown */
		avs->ssi_type = cpu_to_be32(3);	/* 2: dBm, 3: raw RSSI */
		avs->ssi_signal = cpu_to_be32(rssi); /* why put & 0x7f here ? what does the upper bit mean in original prism54 code ? */
		avs->ssi_noise = cpu_to_be32(0);
		avs->preamble = cpu_to_be32(0);	/*unknown */
		avs->encoding = cpu_to_be32(0);	/*unknown */
#endif /* 1 */
		(*skb_ref)->dev = netdev;
#define ETH_P_80211_RAW 0x0019
		(*skb_ref)->protocol = htons(ETH_P_80211_RAW);
		(*skb_ref)->mac.raw = (*skb_ref)->data;
		(*skb_ref)->pkt_type = PACKET_OTHERHOST;
		(*skb_ref)->ip_summed = CHECKSUM_NONE;
		
		netif_rx(*skb_ref);
		return;
	}

	/*
	 * print magic ; Remove packets with bad FCS
	 */
	islsm_dbg("receive packet with magic %i, unknown1 %i, unknown2 %i",magic1, unknown1, unknown2);
	if (!(magic1 & ISLSM_ID_FCS_OK))
	{
		islsm_dbg("packet with bad FCS");
		/* should be dropped. For now, we keep it, as long as
		   we're not sure */
	}
	
#ifdef MADWIFI
#define	IS_CTL(wh) \
	((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) == IEEE80211_FC0_TYPE_CTL)
	/*
	 * From this point on we assume the frame is at least
	 * as large as ieee80211_frame_min; verify that.
	 */
	if (skb->len < IEEE80211_MIN_LEN) {
		islsm_dbg("short packet %d",skb->len);
		dev_kfree_skb(skb);
		return;
	}
	
	skb->protocol = ETH_P_CONTROL;  /* XXX */
	
	/*
	 * Locate the node for sender, track state, and then
	 * pass the (referenced) node up to the 802.11 layer
	 * for its use.  We are required to pass some node so
	 * we fall back to ic_bss when this frame is from an
	 * unknown sender.  The 802.11 layer knows this means the
	 * sender wasn't in the node table and acts accordingly.
	 * Note also that by convention we do not reference
	 * count ic_bss, only other nodes (ic_bss is never free'd).
	 */
	if (ic->ic_opmode != IEEE80211_M_STA) {
		struct ieee80211_frame_min *wh =
			(struct ieee80211_frame_min *) skb->data;
		if (IS_CTL(wh))
			ni = ieee80211_find_node(ic, wh->i_addr1);
		else
			ni = ieee80211_find_node(ic, wh->i_addr2);
		if (ni == NULL)
			ni = ic->ic_bss;
	} else
		ni = ic->ic_bss;
	
	ieee80211_input(ic, skb, ni,
			data_head->signal_strength,
			le32_to_cpu(data_head->timestamp));
	
	/*
	 * Reclaim node reference.
	 */
	if (ni != ic->ic_bss)
		ieee80211_free_node(ic, ni);
#undef IS_CTL
#else /* MADWIFI */
	dev_kfree_skb(skb);
#endif /* MADWIFI */
#undef DATA_HEAD
	return;
}

static 
void islsm_input_stats(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm_control_stats * stats = (struct islsm_control_stats *) skb->data;
	islsm_dbg("stats report : rx_pkt %i, %i, %i, %i, %i, %i, time %i, %i, noise %i",
		 le32_to_cpu(stats->rx_pkt),
		 le32_to_cpu(stats->stat1),
		 le32_to_cpu(stats->stat2),
		 le32_to_cpu(stats->stat3),
		 le32_to_cpu(stats->stat4),
		 le32_to_cpu(stats->stat5),
		 le32_to_cpu(stats->timestamp),
		 le32_to_cpu(stats->stat6),
		 le32_to_cpu(stats->noise));
	dev_kfree_skb(skb);
	return;
}

static 
void islsm_input_txack(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm_control_queueemptied *free = (struct islsm_control_queueemptied *) skb->data;
	islsm_dbg("tx ack : status %x, ack strength %x, sequence %04x, %x",
		 free->status,
		 le16_to_cpu(free->ack_strength),
		 free->sequence,
		 free->unknown);
	dev_kfree_skb(skb);
	return;
}

static
void islsm_input_freqdone(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm_control_freqchange *freq = (struct islsm_control_freqchange *) skb->data;
	islsm_dbg("frequency sucessfully set to %i MHz", le16_to_cpu(freq->frequency));
	dev_kfree_skb(skb);
}


/* first pass only copies the buffer */
static 
void input_mgmt_frame(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct islsm_control_eeprom *txe = (struct islsm_control_eeprom *) skb->data;

        /* the copy buffer must be allocated */
	BUG_ON( islsm->mgmt_frame ==  0 );
	BUG_ON( le16_to_cpu(txe->offset) + le16_to_cpu(txe->len) > ISLSM_MGMT_MEM_SIZE );

	/* signal we saw this frame */
	complete(&islsm->dev_init_comp);

	memcpy(&islsm->mgmt_frame[le16_to_cpu(txe->offset)],
	       &txe->data[0],
	       le16_to_cpu(txe->len));

	dev_kfree_skb(skb);
	return;
}

/* second level input function (along with data_input at top) */

static 
void islsm_input_islheader(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm_tx_control *rx = (struct islsm_tx_control *) skb->data;
	unsigned int trim_length = le16_to_cpu(rx->id.length);
	unsigned int type = le16_to_cpu(rx->type);
	u32 id = rx->req_id;

	skb_pull(skb, SIZE_TX_CONTROL);
	skb_trim(skb, trim_length);

//	islsm_dbg("packet magic1 %08x with standard payload length %i, req_id %08x, type %04x",
//		  magic1, le16_to_cpu(rx->id.length),rx->req_id,type);

	// free the associated queue, if any
	(void) islsm_txqueue_free(netdev, id);
	
	switch ( type )
	{
	case ISLSM_TX_CONTROL_TYPE_FREQDONE:
		islsm_input_freqdone(netdev,skb);
		break;
	case ISLSM_TX_CONTROL_TYPE_TXDONE:
		islsm_input_txack(netdev,skb);
		break;
	case ISLSM_TX_CONTROL_TYPE_STAT_READBACK:
		islsm_input_stats(netdev, skb);
		break;
	case ISLSM_TX_CONTROL_TYPE_EEPROM_READBACK:
		input_mgmt_frame(netdev, skb);
		break;
	default:
		islsm_dbg("unknown control packet length %i, req_id %08x, type %04x",
			  le16_to_cpu(rx->id.length),rx->req_id,type);
		islsm_data_debug(netdev, skb->data, skb->len);
		/* signal something went wrong */
		dev_kfree_skb(skb);
	}
	return;
}

/*
 * first level input function, dispatches on id
 */

void islsm_data_input(struct net_device *netdev, struct sk_buff *skb)
{
	unsigned int magic1;

	/* please note : on the mgmt endpoint, packets with the
	 * mgmt_response flag not set should not generate responses on
	 * this pipe ! We have no way of making the distinction here
	 */

	/* add this for debug, temporarily */
//	islsm_data_debug(netdev, 0x81, skb->data, skb->len);
	if (skb->len == 0)
	{
		islsm_dbg("data_input : null packet");
		dev_kfree_skb(skb);
		return;
	}

	skb->dev = netdev;

	magic1 =  le16_to_cpu(((struct islsm_id *) (skb->data))->magic1);

	islsm_dbg("islsm input packet magic %i", magic1);
	/* actually i should just check on an and with 0x8000, but until
	 * the fields are known, i'd rather not to... */
	switch ( magic1 & ISLSM_ID_MASK )
	{
	case ISLSM_ID_DATA_PKT:
		input_data_frame(netdev,skb);
		break;
		
	case ISLSM_ID_EEPROM_READBACK:
		islsm_input_islheader(netdev,skb);
		break;
		
	case ISLSM_ID_FAULT:
		islsm_input_fault(netdev,skb);
		break;

	default:
		islsm_dbg("unknown magic1 %i packet on data pipe",magic1);
		islsm_data_debug(netdev, skb->data, skb->len);
		dev_kfree_skb(skb);
	};

	return;
}
EXPORT_SYMBOL(islsm_data_input);

static
void islsm_input_fault(struct net_device *netdev, struct sk_buff *skb)
{
	char *buffer = skb->data;
        while(!strncmp(buffer, "!MOR", 4))
	{
		printk("%.4s",buffer);
		skb_pull(skb,4);
		buffer = skb->data;
	}
#define DESCR_OFFSET 0x43c
	printk("%s",buffer + DESCR_OFFSET);
	islsm_data_debug(netdev, skb->data, skb->len);
	/* more to be written */
}

void islsm_bootup_input(struct net_device *netdev, struct sk_buff * skb)
{
	struct islsm *islsm = netdev_priv(netdev);
	char *buffer = (char*) skb->data;

	/* don't do anything fancy for now */
	islsm_dbg("bootup input :");
	islsm_data_debug(netdev, skb->data, skb->len);

	/* When receiving g, we're done ! 
	 * Maybe we should split the receive OK / send go / receive g process
	 * Someone tests this ?
	 */
        if(!strncmp(buffer, "ERROR", 5)) {
		islsm_dbg("FW Upload not successful: (%s)\n", buffer);
        } else if(!strncmp(buffer, "OK", 2)) {
		islsm_dbg("FW Upload successful: (%s)\n", buffer);
	} else if (!strncmp(buffer, "g", 1)) {
		islsm_dbg("Device initialized");
		complete(&islsm->dev_init_comp);
	}
	dev_kfree_skb(skb);
	return;
}
EXPORT_SYMBOL(islsm_bootup_input);

/*
 * lmac readback function
 */
/*
  This function should be able to check that the firmware is correct for
  the device. It also should set the islsm parameters accordingly, maybe
  looking into this header.
  For instance, this affectation :
  islsm->config_reg = ISLSM_TX_QUEUE_CONFIG;
  islsm->datatx_reg = ISLSM_TX_QUEUE_ZERO;

  should be based on what we can see here.

  Now, let's try this function on some various firmwares to how we can
  recognize the various species of firmware...

  This doesn't quite work on the 1.0.4.3 hardmac firmware that i have.

*/
int
islsm_parse_firmware(struct net_device *netdev, void* data, unsigned int len)
{
	unsigned int run_length = 0;
	unsigned int max_offset = len;
	short end = 0;
	islsm_dbg("parsing firmware");

	// offset is wrong in case of X2. X2 should accomodate for this
	/* skip the signature of the file */
	run_length += 2*sizeof(u32);

	/* find the beginning of umac header */
	while (run_length < max_offset)
	{
		int info = *((u32 *)(data + run_length));
		if (info)
			break;
		run_length += sizeof(u32);
	}
	
	islsm_dbg("found start of umac description at offset %i",run_length);

	while (!end && run_length < max_offset)
	{
		struct islsm_fw_option *option = (struct islsm_fw_option *) (data+run_length);
		u32 option_magic = le32_to_cpu(option->magic);
		unsigned int next_option_offset = le32_to_cpu(option->length) * sizeof(u32);
		islsm_info("option magic %08x", option_magic);
		islsm_info("option length %08x", next_option_offset);
		islsm_info("option flags %04x", (option_magic >> 16));

		if (run_length + next_option_offset > max_offset)
		{
			islsm_info("option too long, aborting");
			break;
		}

		switch( option_magic & 0xffff ) {
		case LMAC_OPT_1 :
			/* should completely abort */
			if (strncmp(option->data,"LM86",4))
				islsm_info("Trying to upload a non-softmac firmware !");
		case LMAC_OPT_2 :
			islsm_info("string option :");
			islsm_info("%.*s",next_option_offset,(char *) option->data);
			break;

		case LMAC_OPT_FINAL :
			islsm_info("final option");
			end = 1;
		case LMAC_OPT_3 :
		case LMAC_OPT_4 :
		default :
			islsm_info("option values :");
			islsm_data_debug(netdev,option->data,next_option_offset);
		}
		run_length += sizeof(*option) + next_option_offset;
	}

	// we've reached the version string, print it
	if (end)
		islsm_info("Version string for firmware is : %s", (char*) (data + run_length));

	islsm_dbg("parsing firmware done.");
	return 0;
}
EXPORT_SYMBOL(islsm_parse_firmware);

/*
 * eeprom readback input functions
 */

/* This slow function is now called at the end of the mgmt readback */
/* maybe this should be done automagically... */
void
islsm_parse_mgmt_frame(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct islsm_mgmt_rx_header *mgmt_h = (struct islsm_mgmt_rx_header *) islsm->mgmt_frame;
	struct islsm_mgmt_option *mgmt;
#ifdef MADWIFI
	struct ieee80211com *ic = &islsm->sc_ic;
#endif
	unsigned int next_option_offset;
	unsigned int run_length = 0;
	// Maximal observed frame size
	unsigned int total_length = 0x400;
	
	next_option_offset = le16_to_cpu(mgmt_h->preamble_length);
	/* This first offset is expressed in bytes */
	islsm_dbg("mgmt frame header is :");
	islsm_data_debug(netdev,mgmt_h,next_option_offset+12);
	mgmt = (struct islsm_mgmt_option *) &(((u8 *)&(mgmt_h->data))[next_option_offset]);

	while(run_length < total_length)
	{
		unsigned str_length;
		unsigned int option_magic = le16_to_cpu(mgmt->magic);
		next_option_offset = le16_to_cpu(mgmt->length);
		str_length = (next_option_offset-1) * 2;
		run_length += next_option_offset + 1;

		switch ( option_magic )
		{
		case MGMT_NAME_OPTION:
			islsm_dbg("Device identifies itself as  : %.*s",str_length,mgmt->data);
			break;
		case MGMT_SERIAL_OPTION:
			islsm_dbg("Device reports following S/N : %.*s",str_length,mgmt->data);
			break;
		case MGMT_HWTYPE_OPTION:
		{
			unsigned int hwtype = le32_to_cpu( *(u32*)mgmt->data );
			struct islsm_hw_option *option = (struct islsm_hw_option *) mgmt->data;
			islsm_info("Device reports hardware type : %08x", hwtype);
			islsm_info("Device is an isl38%02x", option->isl);
			if ( option->unk1 != HW_UNK1 || le16_to_cpu(option->unk2) != HW_UNK2 )
				islsm_info("Unusual hardware information. Please contact the module author with the above information.");
			break;
		}
		case MGMT_MAC_OPTION :
			islsm_dbg("Device reported its MAC");
#ifdef MADWIFI
			IEEE80211_ADDR_COPY(ic->ic_myaddr, mgmt->data);
			IEEE80211_ADDR_COPY(netdev->dev_addr, mgmt->data);
#else
			memcpy(netdev->dev_addr, mgmt->data, ETH_ALEN);
#endif /* MADWIFI */
			break;
		case MGMT_END_OPTION :
			islsm_dbg("End of options after %i bytes processed without a final option seen, please report", run_length);
			run_length = total_length;
			break;
		case MGMT_FREQ_3_OPTION :
			islsm_dbg("radio freq3 option");
			{
				struct islsm_mgmt_freq3_option *dat = (struct islsm_mgmt_freq3_option *) mgmt->data;
				unsigned int nfreqs_max = sizeof(islsm->finfo3) / sizeof(islsm->finfo3[0]);
				unsigned int j;
				unsigned int nfreqs = dat->nelem;
				islsm_data_debug(netdev,&mgmt->magic,next_option_offset*sizeof(u16));
				islsm_dbg("%i channels", nfreqs);
				for ( j = 0 ; j < nfreqs ; j++)
				{
					unsigned int channel = islsm_ref_to_chan(le16_to_cpu(dat->data[j].freq));
					
					if (channel >= nfreqs_max || channel <= 0)
					{
						islsm_dbg("channel %i out of range!",channel);
						continue;
					}
					islsm->finfo3[channel] = dat->data[j];
				}
			}
			break;
		case MGMT_FREQ_4_OPTION :
			islsm_dbg("radio freq4 option");
			{
				struct islsm_mgmt_freq4_option *dat = (struct islsm_mgmt_freq4_option *) mgmt->data;
				struct islsm_freq_option4 *entry = (struct islsm_freq_option4 *) dat->data;

				unsigned int nfreqs = dat->nfreqs;
				unsigned int nfreqs_max = sizeof(islsm->finfo4) / sizeof(islsm->finfo4[0]);

				unsigned int j,i;
				unsigned int extended = (dat->flags & FREQ4_TYPE_EXTENDED);
				unsigned int nsettings = (dat->nsettings);

				islsm_dbg("%i channels, %i settings per channels", nfreqs,nsettings);
				
				islsm_data_debug(netdev,&mgmt->magic,next_option_offset*sizeof(u16));
				BUG_ON(nsettings > ISLSM_CHANNEL_DEFS);
				islsm->option4_nelems = nsettings;
				for ( j = 0 ; j < nfreqs ; j++)
				{
					/* FIXME : depends on */
					unsigned int channel = islsm_ref_to_chan(le16_to_cpu(entry->freq));
					if (channel >= nfreqs_max || 0 >= channel)
					{
						islsm_dbg("channel %i out of range!",channel);
						break;
					}
					/* store the frame and do the calculations already */
					if ( !extended )
					{
						struct islsm_freq_option4_s *sparse = 
							(struct islsm_freq_option4_s *) entry->data;
						for(i = 0; i < nsettings; i++)
						{
							struct islsm_freq_option4_f * command = &islsm->finfo4[channel][i].command;
#define SUB(x,y) (u8)((x) - (y)) > (x) ? 0 : (x) - (y)
							command->head=sparse[i].head;
							command->tail=sparse[i].tail;
							command->subtails[3]=SUB(command->tail,12);
							command->subtails[2]=SUB(command->subtails[3],12);
							command->subtails[1]=SUB(command->subtails[2],12);
							command->subtails[0]=SUB(command->subtails[1],14);
#undef SUB
							islsm->finfo4[channel][i].padding = 0;
						}
						entry = (struct islsm_freq_option4 *) &sparse[nsettings];
					}
					else /* we have the complete
						stack */
					{
						char *data;
						struct islsm_freq_option4_f *full = 
							(struct islsm_freq_option4_f *) entry->data;
						for(i = 0; i < nsettings; i++)
						{
							islsm->finfo4[channel][i].command = full[i];
							islsm->finfo4[channel][i].padding = 0;
						}
						// would be so easy with skb...
						data = (char *) &full[nsettings];
						data++;
						/* this will work with my device, but is this
						 * solid ? -- skip one padding
						 * byte */
						entry = (struct islsm_freq_option4 *) data;
					}
				}
			}
//		case MGMT_FREQ_5_OPTION :
//			islsm_dbg("islsm freq5 option");
			// data ?
			break;
		case MGMT_FREQ_6_OPTION :
			islsm_dbg("islsm freq6 option");
			{
				struct islsm_freq_struct3 *dat = (struct islsm_freq_struct3 *) mgmt->data;
				unsigned int nfreqs_max = sizeof(islsm->finfo6) / sizeof(islsm->finfo6[0]);
				unsigned int j;
				unsigned int nfreqs = ((next_option_offset-1) << 1) / sizeof(islsm->finfo6[0]);
				islsm_data_debug(netdev,&mgmt->magic,next_option_offset*sizeof(u16));
				islsm_dbg("islsm freq6 array has %i elements", nfreqs);
				for ( j = 0 ; j < nfreqs ; j++)
				{
					unsigned int channel = islsm_ref_to_chan(le16_to_cpu(dat[j].freq));
					if (channel >= nfreqs_max || channel <= 0)
					{
						islsm_dbg("channel %i out of range!",channel);
						continue;
					}
					islsm->finfo6[channel] = dat[j];
				}
			}
			break;
		case MGMT_FINALCRC_OPTION :
			/* display CRC ? */
			islsm_dbg("End of options after %i bytes processed", run_length);
			run_length = total_length;
			break;
		default :
			islsm_dbg("Unknown option length %i, magic %02x,please report. data :",next_option_offset, option_magic);
			islsm_data_debug(netdev,&mgmt->magic,next_option_offset*sizeof(u16));
		}
		/* skip to next option : This option offset is
		   expressed in 16-bits words */
		mgmt = (struct islsm_mgmt_option *) (((u16 *)&mgmt->magic) + next_option_offset);
	}
}
EXPORT_SYMBOL(islsm_parse_mgmt_frame);

/*
 *  tx queues management
 */

void islsm_txqueue_init(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct tx_queue *tx_queue_array = &islsm->tx_queues[0];
	u8 i;
	for(i=0; i<ISLSM_NR_TX_QUEUES; i++) {
		tx_queue_array[i].id = (i+6);
		tx_queue_array[i].usb_id = 0;
		tx_queue_array[i].ptr = 0;
	}
}

u8 islsm_txqueue_alloc(struct net_device *netdev, void* ptr) {
	/* race on alloc, should use atomic "free" field */
	struct islsm *islsm = netdev_priv(netdev);
	struct tx_queue *tx_queue_array = &islsm->tx_queues[0];
	u8 i;

	if (ptr == 0) {
		islsm_dbg("alloc null ptr not authorized");
		return 0;
	}

	islsm_dbg("alloc for id %p",ptr);
	
	for(i=0; i<ISLSM_NR_TX_QUEUES; i++) {
		if (tx_queue_array[i].ptr == 0) {
			tx_queue_array[i].ptr = ptr;
			tx_queue_array[i].usb_id  = tx_queue_array[i].id;
			return tx_queue_array[i].id;
		}
	}

	islsm_dbg("no free queue !");
	return 0;
}

void *islsm_txqueue_free(struct net_device *netdev, u32 usb_id) {
	struct islsm *islsm = netdev_priv(netdev);
	struct tx_queue *tx_queue_array = &islsm->tx_queues[0];
	u8 i;
	void *ptr = 0;
	
	if (usb_id == 0) {
		return 0;
	}

//	islsm_dbg("trying to free queue %08x",usb_id);

	for(i=0;i<ISLSM_NR_TX_QUEUES;i++) {
		//islsm_dbg("trying with %08x",tx_queue_array[i].usb_id);
		if (tx_queue_array[i].usb_id == usb_id ) {
			islsm_dbg("queue %08x freed",tx_queue_array[i].id);
			ptr = tx_queue_array[i].ptr;
			tx_queue_array[i].ptr = 0;
			tx_queue_array[i].usb_id = 0;
			break;
		}
	}

	if (i==ISLSM_NR_TX_QUEUES)
		islsm_dbg("no matching queue found for %08x !",usb_id);

	return ptr;
}

/*
 * helper functions
 */

static unsigned int 
islsm_ref_to_chan(unsigned int ref) {
#ifdef MADWIFI
	return ieee80211_mhz2ieee(ref, IEEE80211_CHAN_2GHZ);
#else
	/* kludgy as hell */
	unsigned int out;	
#define ISLSM_REF_CHAN0 0x0967
	if (ref <= ISLSM_REF_CHAN0)
	{
		islsm_info("Problem requesting channel of ref %i : too small", ref); 
		return 0;
	}
	out = (ref-ISLSM_REF_CHAN0) / 5; 
	if (out > 14)
	{
		if (ref == 2484)
			return 14;
		islsm_info("Problem requesting channel of ref %i : channel %i, too big",ref,out); 
		return 0;
	}
	return out;
#undef ISLSM_REF_CHAN0
#endif /* MADWIFI */
}

static
void islsm_data_debug(struct net_device *netdev, void *_data, int len)
{
	unsigned char *data = _data;
	int k, l;
	unsigned long int time = jiffies;
	if(sm_debug > 1) {
		printk(KERN_CRIT "[IN +%-6ld]\n", time);
		for(k = 0; k < len; k+= 16) {
			printk(KERN_CRIT "%08x:", k);
			for (l = 0; (l < 16) && (k + l < len); ++l) {
				printk(" %02x", ((unsigned char *)data)[k + l]);
			}
			printk("\n");
		}
	}
}
