/*
  Copyright 2004 Feyd
  Copyright 2004, 2005 Jean-Baptiste Note

  This file is part of prism54usb.

  prism54usb is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  prism54usb is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with prism54usb; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/skbuff.h>
#include <asm/uaccess.h>

#include "isl_38xx.h"
#include "islusb_dev.h"
#include "isl_sm.h"

void p54u_mdelay(int ms)
{
	int t = (ms * HZ + 500) / 1000;
	
	if(!t && ms)
		t = 1;
	
	set_current_state(TASK_UNINTERRUPTIBLE);
	while(t) {
		//printk("t: %i\n", t);
		t = schedule_timeout(t);
	}
	
	return;
}


static
void p54u_data_debug(struct net_device *netdev, unsigned int ep, void *_data, int len)
{
	unsigned char *data = _data;
	int k, l, in;
	unsigned long int time = jiffies;
	in = ep & USB_DIR_IN;
	if(islusb_debug > 1) {
		printk(KERN_CRIT "[%s >%-6lld<%-6lld+%-6ld] %02x\n", in?"IN ":"OUT", 0LL, 0LL, time, ep);
		for(k = 0; k < len; k+= 16) {
			printk(KERN_CRIT "%08x:", k);
			for (l = 0; (l < 16) && (k + l < len); ++l) {
				printk(" %02x", ((unsigned char *)data)[k + l]);
			}
			printk("\n");
		}
	}
}

int p54u_bulk_msg(struct net_device *netdev, unsigned int ep, void *_data, int len)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;
	unsigned char *data = _data;
	size_t alen;
	int pipe, err;

	p54u_data_debug(netdev, ep, data, len);
	
	pipe = usb_sndbulkpipe(usbdev, ep);
	err = usb_bulk_msg(usbdev, pipe, data, len, &alen, 2 * HZ);
	if(err)
		islusb_info("bulk submit failed: %i\n", err);
	
	return err;
}

int p54u_submit_urb(struct net_device *netdev, struct urb *urb, int flags)
{
	int err;

	p54u_data_debug(netdev, 0x00, urb->transfer_buffer, urb->transfer_buffer_length);

	urb->transfer_flags |= URB_ASYNC_UNLINK;
	urb->actual_length = 0;	
	err = usb_submit_urb(urb,flags);

	if(err)
	{
		islusb_info("urb submit failed: %i\n", err);
	}

	return err;
}

/* RX QUEUE MANAGEMENT */

/* callback common to all bulk endpoints */
static
void p54u_urb_callback(struct urb *urb, struct pt_regs *regs);
static
void p54u_nil_urb_callback(struct urb *urb, struct pt_regs *regs);
static int
p54u_alloc_urb(struct usb_device *usbdev, struct p54u_pipe *queue, struct urb **urb_ref);

/* Should be called under lock ? */
static
int rx_queue_add (struct net_device *netdev, struct p54u_pipe *queue) 
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;
	struct sk_buff *skb;
	int err;
	unsigned int index;
	struct urb **urb_ref;

	index = queue->p % queue->len;
	urb_ref = &queue->ringbuf[index].data_urb;
	
	/* realloc urb if needed */
	/* needed only in non-atomic context */
	err=p54u_alloc_urb(usbdev,queue,urb_ref);
	if (err)
		return err;
	
	/* realloc skb -- skb recycling not authorized for now */
	BUG_ON(queue->ringbuf[index].skb != (struct sk_buff *) 0);
	skb = dev_alloc_skb(ISLSM_MAX_FRAME_SIZE);

	if (skb == (struct sk_buff *) 0)
	{
		islusb_dbg("could not allocate rx skb");
		return -ENOMEM;
	}

	/* realign buffer */
	skb_push(skb, skb_headroom(skb));
	skb_trim(skb, 0);

	queue->ringbuf[index].skb = skb;

	(*urb_ref)->transfer_buffer = (queue->ringbuf[index].skb)->data;
	(*urb_ref)->transfer_buffer_length = queue->buffer_size;

	err = usb_submit_urb(*urb_ref, GFP_KERNEL);
	if (err)
	{
		/* unalloc everything we allocd... Just in case we run
		 * again (maybe recycling would be good here, and
		 * usefull: for instance, we may recover in case there's
		 * a bandwidth problem */
		islusb_dbg("could not submit rx urb : %i",err);
		dev_kfree_skb(queue->ringbuf[index].skb);
		queue->ringbuf[index].skb = 0;
	}

	return err;
}

/* needed for ver1 devices */

static inline
void remove_net2280_header(struct sk_buff *skb) {
	/* trim the packet and get off its net2280 header */
	unsigned int trim_length = le16_to_cpu( ((struct islsm_rx *) skb->data)->length );
	skb_pull(skb, SIZE_RX);
	skb_trim(skb, trim_length);
	return;
}

static inline
int rx_queue_process (struct net_device *netdev, struct p54u_pipe *queue) {
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	/* don't go further than this. We don't miss any if we're a
	 * tasklet. Check for others. */
	unsigned long final = queue->c;
	int err = 0;
	/* should lock, no problem on !smp */
	/* queue_lock */
	/* need serialization against ourselves. no big deal if queue->c
	   changes from under us ; big problem if we change or decide to
	   change queue->f at the same time. Maybe with atomic
	   inc_and_compare we could be set... investigate this ! */
	/* for all acked packets in the queue */
	BUG_ON(queue->f > final);
	while ( queue->f != final )
	{
		unsigned int index = queue->f % queue->len;
		usb_command_t *cmd = &queue->ringbuf[index];
		struct sk_buff *skb = cmd->skb;
		if (!cmd->data_urb->status) {
			(void) skb_put(skb,cmd->data_urb->actual_length);
			
			if ((p54u->device_version == P54U_DEVICE_VER2) && 
			    (p54u->state == P54U_BOOT))
				islsm_bootup_input(netdev,skb);

			else if (p54u->state == P54U_RUN)
			{
				if (p54u->device_version == P54U_DEVICE_VER1)
					remove_net2280_header(skb);
				islsm_data_input(netdev,skb);
			}
			else
			{
				islusb_dbg("dropping received packet while not running");
				dev_kfree_skb(skb);
			}
		}
		else {
			/* urb was fucked up */
			islusb_dbg("rx urb bad status : %i",cmd->data_urb->status);
			err = -EIO;
			dev_kfree_skb(skb);
		}
		queue->ringbuf[index].skb = (struct sk_buff *) 0;
		queue->f++;
	}
	return err;
}

int rx_queue_refill (struct net_device *netdev, struct p54u_pipe *queue) {
	/* iterates and refill buffer */
	/* needs serialization against ourselves */
	/* needs only to have a consistent queue->f if we're serialized
	 * against ourselves. It's no big problem if queue->f changes
	 * from under us ; it is if queue->p is changed ! */
	/* produce-lock */
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	int err = 0;
	if (p54u->state == P54U_SHUTDOWN)
	{
		islusb_dbg("not refilling queue %02x in shutdown state",queue->addr);
		return 0;
	}

	BUG_ON( queue->p - queue->f > queue->len);
	while ( (queue->p - queue->f) != queue->len)
	{
		if ( (err = rx_queue_add(netdev,queue)) )
			break;
		queue->p++;
	}
	return err;
}

void
p54u_rx_bh(void *data) {
	struct net_device *netdev = (struct net_device *) data;
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	int err;

	err = rx_queue_process(netdev,&p54u->data_rx);
	if (!err)
		rx_queue_refill(netdev,&p54u->data_rx);

	// FIXME : Does the mgmt queue even though it's not initialized
	// !!! -- not good
	// FIXME : same problem on tx. dissociate with per-queue bh...
	if (p54u->device_version == P54U_DEVICE_VER1)
	{
		err = rx_queue_process(netdev,&p54u->mgmt_rx);
		if (!err)
			rx_queue_refill(netdev,&p54u->mgmt_rx);
	}
	return;
}

static inline 
void rx_queue_cancel (struct p54u_pipe *queue, unsigned int index) {
	index = index % queue->len;
	usb_unlink_urb(queue->ringbuf[index].data_urb);
}

/* only these are exported */
void rx_queue_empty(struct p54u_pipe *queue) {
	unsigned int start = queue->c;
	unsigned int end = queue->p;
	BUG_ON(start > end);
	/* queue lock should not be needed */
	/* do this with irq disabled */
	while ( start < end )
		rx_queue_cancel(queue, start++);
	
	// volatile ?
	//while (queue->c < queue->p)
	//	p54u_mdelay(10);
}

/* TODO : queue unalloc function */

/* TX QUEUE MANAGEMENT */

/* called in non-atomic context */
static int
p54u_alloc_urb(struct usb_device *usbdev, struct p54u_pipe *queue, struct urb **urb_ref)
{
	if (*urb_ref == (struct urb *) 0)
	{
		*urb_ref = usb_alloc_urb(0,GFP_KERNEL);
		if (*urb_ref == (struct urb *) 0)
		{
			islusb_dbg("could not allocate rx/tx data urb");
			return -ENOMEM;
		}

		(*urb_ref)->transfer_flags |= URB_ASYNC_UNLINK;

		switch(queue->type) {
		case USB_ENDPOINT_XFER_BULK:
			usb_fill_bulk_urb(*urb_ref, usbdev, queue->endp, 0, queue->buffer_size,
					  p54u_urb_callback, queue);			
			break;
		case USB_ENDPOINT_XFER_INT:
			usb_fill_int_urb(*urb_ref, usbdev, queue->endp, 0, queue->buffer_size,
					 p54u_urb_callback, queue, queue->interval);
			break;
		default:
			islusb_err("unrecognized queue type");
		}
	}
	return 0;
}

static int
p54u_alloc_descr_urb(struct net_device *netdev, struct p54u_pipe *queue, struct urb **urb_ref)
{
	/* TODO : fix size for version 2 devices */
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;
	unsigned int size = 16;
	void *buf;

	buf = *urb_ref;
	if (*urb_ref == (struct urb *) 0)
	{
		*urb_ref = usb_alloc_urb(0,GFP_KERNEL);
		if (*urb_ref == (struct urb *) 0)
		{
			islusb_dbg("could not allocate descr urb");
			return -ENOMEM;
		}
		
		(*urb_ref)->transfer_flags |= URB_ASYNC_UNLINK;
	}

	/* urb was not allocated */
	if ( !buf )
	{
		(*urb_ref)->transfer_flags |= (URB_NO_TRANSFER_DMA_MAP | URB_ASYNC_UNLINK);

		/* fill in the urb when in version 1 device ? */
		buf = usb_buffer_alloc(usbdev, size, GFP_KERNEL, &(*urb_ref)->transfer_dma);
		if (buf == (void *)0)
		{
			islusb_dbg("could not allocate tx descr buf");
			usb_free_urb(*urb_ref);
			return -ENOMEM;
		}

		/* the descr urb is very different from one device from
		 * the other */
		if (p54u->device_version == P54U_DEVICE_VER1) {
			int endp = usb_sndbulkpipe(usbdev, P54U_PIPE_DEV & USB_ENDPOINT_NUMBER_MASK);
			usb_fill_bulk_urb(*urb_ref, usbdev, endp, buf, 16,
					  p54u_nil_urb_callback, 0);
		}
		else {
			usb_fill_bulk_urb(*urb_ref, usbdev, queue->endp, buf, 4,
					  p54u_nil_urb_callback, 0);
		}
	}

	return 0;
}

/* Should be called under lock ? */
static
int tx_queue_add (struct net_device *netdev, struct p54u_pipe *queue, unsigned int index) {
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;
	int err = 0;
	struct urb **urb_descr_ref, **urb_data_ref;

	urb_descr_ref = &queue->ringbuf[index].descr_urb;
	urb_data_ref = &queue->ringbuf[index].data_urb;
	
	/* alloc urb needed */
	err = p54u_alloc_urb(usbdev, queue, urb_data_ref);
	if (!err)
		p54u_alloc_descr_urb(netdev, queue, urb_descr_ref);

	/* add something */

	return err;
}

static inline
void tx_queue_process (struct net_device *netdev, struct p54u_pipe *queue) {
	unsigned long final = queue->c;
	/* should lock, no problem on !smp */
	/* queue_lock */
	/* need serialization against ourselves. no big deal if queue->c
	   changes from under us ; big problem if we change or decide to
	   change queue->f at the same time. Maybe with atomic
	   inc_and_compare we could be set... investigate this ! */
	/* for all acked packets in the queue */
	BUG_ON(queue->f > final);
	while ( queue->f != final )
	{
		unsigned int index = queue->f % queue->len;
		struct sk_buff *skb = queue->ringbuf[index].skb;
		/* free can be done unatomically */
		islusb_dbg("freeing skb %p, index %i",skb,queue->f);
		queue->ringbuf[index].skb = (struct sk_buff *) 0;
		dev_kfree_skb(skb);
		queue->f++;
	}
}

void
p54u_usb_tx_bh(void *data) {
	struct net_device *netdev = (struct net_device *) data;
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	tx_queue_process(netdev,&p54u->data_tx);
	if (p54u->device_version == P54U_DEVICE_VER1)
	    tx_queue_process(netdev,&p54u->mgmt_tx);
	return;
}

/* INTERRUPT QUEUE. MUST BE OF SIZE ONE ! */
/* disable interrupt in ISL */
/* The prevents us from going to tasklet style for the int queue */
static
void isl_process_interrupt(struct net_device *netdev)
{
	u32 reg;
	struct islsm *islsm = netdev_priv(netdev);
	p54u_dev_writel(netdev, ISL38XX_INT_EN_REG, 0);
	do {
		/* ack the interrupts */
		reg = p54u_dev_readl(netdev, ISL38XX_INT_IDENT_REG);
		islusb_dbg("isl int vector: %08x", reg);
		if (reg & ISL38XX_INT_IDENT_INIT)
		{
			islusb_dbg("isl init flag set, chip has booted !");
			complete(&islsm->dev_init_comp);
		}
		if (reg & ~(0x80000000))
			p54u_dev_writel(netdev, ISL38XX_INT_ACK_REG, reg);
	}
	while(reg & ~(0x80000000));
	
	/* clear the interrupt at net2280 level
	   the net2280 is in host mode */
	p54u_brg_writel(netdev, NET2280_IRQSTAT1, NET2280_PCI_INTA_INTERRUPT);
	
	/* reenable interrupts in ISL */
	p54u_dev_writel(netdev, ISL38XX_INT_EN_REG, 0x00004004);
}

void p54u_int_bh(void *data) 
{
	struct net_device *netdev = (struct net_device *) data;
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct p54u_pipe *queue = &p54u->int_rx;
	struct sk_buff *skb;
	usb_command_t *cmd;
	int pending = 0;
	int err = 0;

	while ( queue->f != queue->c )
	{
		cmd = &queue->ringbuf[queue->f % queue->len];
		skb = cmd->skb;
		// unlikely
		if (cmd->data_urb->status || cmd->data_urb->actual_length < 4)
		{
			islusb_dbg("bad int urb : status %i, length %i",
				 cmd->data_urb->status,
				 cmd->data_urb->actual_length);
			err = -EIO;
		}
		else
		{
			(void) skb_put(skb,cmd->data_urb->actual_length);

			pending = le32_to_cpu(*(int *)skb->data);
			islusb_dbg("interrupt vector %08x", pending);
			
			if (pending & NET2280_PCI_INTA_INTERRUPT)
 				isl_process_interrupt(netdev);
		}
		/* recycle the same buffer */
		skb_trim(skb,0);
		queue->f++;
	}

	// likely
	if (err == 0 && p54u->state != P54U_SHUTDOWN)
		while(queue->p != queue->f + queue->len)
		{
			cmd = &queue->ringbuf[queue->p % queue->len];
			skb = cmd->skb;
			if (p54u_submit_urb(netdev,cmd->data_urb,GFP_KERNEL))
				break;
			queue->p++;
		}
	return;
}

/* QUEUE MANAGEMENT */
/* Agnostic functions */
static
void p54u_urb_callback(struct urb *urb, struct pt_regs *regs) {
	struct p54u_pipe *queue = (struct p54u_pipe *) urb->context;
        /* TODO : spin lock for c, or atomic. Check in what context the
	 * callback is called. Is it serialized against callbacks on the
	 * same endpoint ? */
	queue->c++;
//	islusb_dbg("queue %02x c -> %i", queue->addr, queue->c);
	/* schedule bottom half */
	schedule_work(&queue->bh);
	
	return;
}

static
void p54u_nil_urb_callback(struct urb *urb, struct pt_regs *regs) {
	return;
}

static inline 
void p54u_queue_cancel (struct p54u_pipe *queue, unsigned int index) {
	index = index % queue->len;
	usb_unlink_urb(queue->ringbuf[index].data_urb);
	if (queue->ringbuf[index].descr_urb)
		usb_unlink_urb(queue->ringbuf[index].descr_urb);
}

static
void p54u_queue_empty(struct p54u_pipe *queue) {
	unsigned int start = queue->c;
	unsigned int end = queue->p;
	/* queue lock should not be needed */
	islusb_dbg("emptying the queue %02x from %i to %i",queue->addr,start,end);
	while ( start < end )
		p54u_queue_cancel(queue, start++);
	
	// volatile ?
	//while (queue->c < queue->p)
	//	p54u_mdelay(10);
}

/* only these are exported */
int p54u_queue_init (struct net_device *netdev, struct usb_endpoint_descriptor *desc, 
		   struct p54u_pipe *queue) 
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;

	/* some needed fields init. Maybe I should just have the desc
	 * saved instead of copying all its fields. */
	/* please note : investigate what this wMaxPacketSize is worth
	 * and _use_it_ instead of a P54U_MAX_bidule 
	 * -- it's worth nothing
	 *  
	 */
	queue->type = desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
	queue->addr = desc->bEndpointAddress;
	queue->interval = desc->bInterval;

	switch(queue->type) {
	case USB_ENDPOINT_XFER_BULK:
		queue->len = P54U_QUEUE_LEN;
		queue->buffer_size = P54U_MAX_FRAME_SIZE;
		/* Maybe we should limit ourselves to only one bh for BULK type */
		if(queue->addr & USB_DIR_IN) {
			queue->endp = usb_rcvbulkpipe(usbdev, queue->addr & USB_ENDPOINT_NUMBER_MASK);
			INIT_WORK(&queue->bh, p54u_rx_bh, netdev);
		}
		else {
			queue->endp = usb_sndbulkpipe(usbdev, queue->addr & USB_ENDPOINT_NUMBER_MASK);
			INIT_WORK(&queue->bh, p54u_usb_tx_bh, netdev);
		}
		break;
	case USB_ENDPOINT_XFER_INT:
		// Should understand better what i'm doing here
		queue->len = 1;
		queue->buffer_size = 4;
		queue->interval = 6;
		queue->endp = usb_rcvintpipe(usbdev, queue->addr & USB_ENDPOINT_NUMBER_MASK);
		INIT_WORK(&queue->bh, p54u_int_bh, netdev);
		break;
	}

	/* queue proper init */
	queue->f = queue->c = queue->p = 0;

	/* initialize arrays */
	queue->ringbuf = kmalloc(sizeof(*queue->ringbuf) * queue->len, GFP_KERNEL);
	if (queue->ringbuf == 0)
	{
		islusb_err("could not allocate queue ring buffer");
		return -ENOMEM;
	}
	/* zero out the thing */
	memset(queue->ringbuf, 0, sizeof(*queue->ringbuf) * queue->len);

	/* In case of an OUT endpoint, we allocate the data, because
	 * we'd better do this in interruptible context, whereas
	 * submission will be done in atomic context */
	if( !(queue->addr & USB_DIR_IN) )
	{
		unsigned int i;
		islusb_dbg("allocating urbs for out endpoint");
		for (i=0 ; i < queue->len ; i++)
			tx_queue_add(netdev,queue,i);
	}

	return 0;
}

void p54u_queue_destroy (struct p54u_pipe *queue) 
{
	unsigned int i;

	if (!queue || !(queue->ringbuf))
		return;

	p54u_queue_empty(queue);

	/* wait for the urb to become free. Maybe i shouldn't set the
	 * async unlink flag ? */
	while(queue->c != queue->p)
	{
		islusb_dbg("waiting for urbs in [%i %i[, queue %02x to return",queue->c, queue->p, queue->addr);
		p54u_mdelay(10);
	}

	/* maybe flush_scheduled_work will work here ? */
//	flush_scheduled_work();

	/* wait for the last bh to do its job */
	/* upping f should will happen in the bh 
	 */
	while(queue->f != queue->c)
	{
		islusb_dbg("waiting for urbs in [%i %i[, queue %02x to be bh'd",queue->f, queue->c, queue->addr);
		schedule_work(&queue->bh);
		p54u_mdelay(10);
	}		


	/* free the alloc'd urbs. This is simple, 
	   usb_free_urb and dev_kfree_skb can be called on nulls. */
	for(i=0; i<queue->len; i++)
	{
		usb_command_t *cmd = &queue->ringbuf[i];
		usb_free_urb(cmd->data_urb);
		usb_free_urb(cmd->descr_urb);
		if (cmd->skb)
			dev_kfree_skb(cmd->skb);
	}

	kfree(queue->ringbuf);
	queue->ringbuf = 0;

	return;
}

/* more factorization needed */
static
int tx_queue_submit (struct net_device *netdev, struct p54u_pipe *queue, 
		     struct sk_buff *skb, u32 address);

int tx_submit_ver1 (struct net_device *netdev, struct sk_buff *skb, u32 lmac_address)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) (islsm->hardware);
	int len = skb->len;
	int err = 0;	
	struct islsm_tx *islsm_tx;
		
	/* headroom should have been allocated by upper layers */
	BUG_ON(skb_headroom(skb) < SIZE_NET2280_H);

	islsm_tx = (struct islsm_tx *) skb_push(skb, SIZE_NET2280_H);
	memset(skb->data, 0, SIZE_NET2280_H);
	islsm_tx->queue  = cpu_to_le32(lmac_address);
	islsm_tx->length = cpu_to_le16(len);
	
	/* this is the limit payload that will trigger a 
	   split */
#define LIMIT 0x400
	if (len + SIZE_NET2280_H < LIMIT)
	{
		islsm_tx->follower = cpu_to_le16(0);
		err = tx_queue_submit (netdev, &p54u->data_tx, skb, lmac_address);
	}
	else
	{
		/* we overflow the max size, we must split the frame */
		islsm_tx->follower = cpu_to_le16(ISLSM_MGMT_FOLLOWUP);
		// This one is not announced ! f* !
		// Should take care of this in the lower
		// functions ?
		err = p54u_bulk_msg(netdev, P54U_PIPE_MGMT, skb->data, SIZE_NET2280_H);
		//tx_queue_submit (netdev, &p54u->mgmt_tx, skb, queue);
		if (!err)
		{
			skb_pull(skb,SIZE_NET2280_H);
			err = tx_queue_submit (netdev, &p54u->mgmt_tx, skb, lmac_address);
		}
	}
	return err;
}

int tx_submit_ver2 (struct net_device *netdev, struct sk_buff *skb, u32 lmac_address) {
	/* version 2 devices needs nothing */
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) (islsm->hardware);	
	int err;
	err = tx_queue_submit (netdev, &p54u->data_tx, skb, lmac_address);	
	return err;
}

static
int tx_queue_submit (struct net_device *netdev, struct p54u_pipe *queue, 
		     struct sk_buff *skb, u32 address) 
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	usb_command_t *command;
	int err;
	/* lock */
	/* the function replaces the announcement messages */
	/* use the skb as data */
	/* lock until submitted, to make sure submittion is in correct
	 * order */
	if (p54u->state != P54U_RUN)
	{
		islusb_dbg("dismissed packet received while not running");
		return -EIO;
	}

	BUG_ON(queue->p > queue->f + queue->len);
	if (queue->p == queue->f + queue->len)
	{
		islusb_dbg("tx queue full");
		return -ENOSPC;
	}

	command = &queue->ringbuf[queue->p % queue->len];
	/* take care of the announcement message */
	/* in case of version 1 */
	if (p54u->device_version == P54U_DEVICE_VER2)
	{
		/* BUG_ON here ? */
		* ((u32*)command->descr_urb->transfer_buffer) = cpu_to_le32(address);
		/* should be taken care of during allocation */
		command->descr_urb->transfer_buffer_length = 4;
	}
	else
	{
		/* ho-ho, more complicated -- do this here or in
		 * allocation ? */
		struct p54u_reg reg;
		int val;

		if (queue->addr == P54U_PIPE_DATA)
		{
			val = ISL38XX_DEV_INT_DATA;
		}
		else if (queue->addr == P54U_PIPE_MGMT)
		{
			val = ISL38XX_DEV_INT_MGMT;
		}
		else
		{
			/* for now */
			islusb_dbg("unknown queue");
			val = ISL38XX_DEV_INT_DATA;
		}

		reg.port = cpu_to_le16(P54U_PORT_DEV_U32);
		reg.addr = cpu_to_le32(P54U_DEV_BASE | ISL38XX_DEV_INT_REG);
		reg.val = cpu_to_le32(val);
		
		memcpy(command->descr_urb->transfer_buffer,&reg,sizeof(reg));
		command->descr_urb->transfer_buffer_length = sizeof(reg);
	}

	/* Check alignment of the frame */
	BUG_ON( ( (long)skb->data & 0x03) );

	/* take care of the data urb */
	/* padding to 4; only do this for version 1 devices */
	if (p54u->device_version == P54U_DEVICE_VER1) {
		if (skb->len % 4)
		{
			/* FIXME : is there enough room ? */
			/* needed for isl3880 fullmac, not for softmac
			   chipsets, it seems */
			unsigned int padding = 4 - (skb->len % 4);
			void * ref = skb_put(skb,padding);
			memset(ref,0,padding);
		}
	}

	command->data_urb->transfer_buffer = skb->data;
	command->data_urb->transfer_buffer_length = skb->len;

	/* submit the urbs, for now with debug */
	err = p54u_submit_urb(netdev,command->descr_urb,GFP_ATOMIC);
	if (!err)
		err = p54u_submit_urb(netdev,command->data_urb,GFP_ATOMIC);
	if (!err)
	{
		command->skb = skb;
		queue->p++;
	}
	/* unlock */
	return err;
}

/* SPECIFIC FUNCTIONS */

u32 p54u_reg_rw(struct net_device *netdev, int write, int ep, int port, u32 addr, u32 val)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct p54u *p54u = (struct p54u *) islsm->hardware;
	struct usb_device *usbdev = p54u->usbdev;
	struct p54u_reg reg;
	unsigned int pipe;
	size_t len,retlen;
	int err;
		
	reg.port = cpu_to_le16(port);
	reg.addr = cpu_to_le32(addr);
	reg.val = ((port & P54U_PORT_U32) == P54U_PORT_U32) ? cpu_to_le32(val) : cpu_to_le32(val&0xffff);
	len = write ? P54U_REG_WSIZE : P54U_REG_RSIZE;
	pipe = usb_sndbulkpipe(usbdev, ep & USB_ENDPOINT_NUMBER_MASK);
	
	err = usb_bulk_msg(usbdev, pipe, &reg, len, &retlen, HZ);
		
	p54u_data_debug(netdev, ep, &reg, len);

	if(err) {
		islusb_err("%s %02x %04x %08x %08x: failed: %i", write ? "Write": "Read", ep, port, addr, val, err);
		//p54u->err = err;
		if(write && (ep == P54U_PIPE_DEV ) && (len != retlen)) {
			/* something bad happened, but we know how to cope with this case. We should do it for reads, too */
			islusb_err("expected to write %i but wrote %i on dev pipe, resetting",len,retlen);
			/* clear stall */
			usb_clear_halt(usbdev,pipe);
			/* resubmit only once */
			err = usb_bulk_msg(usbdev, pipe, &reg, len, &retlen, HZ);
			if(err) {
				islusb_err("%s %02x %04x %08x %08x: failed: %i", write ? "Write": "Read", ep, port, addr, val, err);
				p54u->err = err;
			}
		}
		else {
			p54u->err = err;
		}
		return 0;
	}
	
	if(write) {
//		udelay(ISL38XX_WRITEIO_DELAY);
		return 0;
	}
	
	pipe = usb_rcvbulkpipe(usbdev, ep & USB_ENDPOINT_NUMBER_MASK);
	len = sizeof(reg.val);
	
	err = usb_bulk_msg(usbdev, pipe, &reg.val, len, &len, HZ);
		
	if(err) {
		islusb_err("Register read %02x %04x %08x %08x: failed: %i", ep, port, addr, val, err);
		p54u->err = err;
		return 0;
	}
	
	p54u_data_debug(netdev, ep | USB_DIR_IN, &reg.val, len);
	
	return ((port & P54U_PORT_U32) == P54U_PORT_U32) ? le32_to_cpu(reg.val) : (le32_to_cpu(reg.val)&0xffff);
}
