/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_adapter.c

	Abstract: Initialization and register update routines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/if_arp.h>

#include <net/iw_handler.h>
#include <ieee802_11.h>

#include "rt2x00_adapter.h"

void rt2x00_init_adapter_eeprom_antenna(struct _rt2x00_adapter* adapter, u16 data)
{
	struct _eeprom_antenna	*antenna = (struct _eeprom_antenna*)&data;

	/*
	 * Load antenna data.
	 */
	if(adapter->chip_adapter == RT2400)
		adapter->chip_rf_type = antenna->rt2400_rf_type;
	else
		adapter->chip_rf_type = antenna->rt2500_rf_type;

	if(antenna->tx_default_antenna == 2)
		adapter->config->antenna |= ANTENNA_TX_B;
	else if(antenna->tx_default_antenna == 1)
		adapter->config->antenna |= ANTENNA_TX_A;
	else
		adapter->config->antenna |= ANTENNA_TX_DIV;

	if(antenna->rx_default_antenna == 2)
		adapter->config->antenna |= ANTENNA_RX_B;
	else if(antenna->rx_default_antenna == 1)
		adapter->config->antenna |= ANTENNA_RX_A;
	else
		adapter->config->antenna |= ANTENNA_RX_DIV;
}

void rt2x00_init_adapter_eeprom_txpower(struct _rt2x00_adapter* adapter, u16 *data)
{
	struct _eeprom_txpower	*tx_power = NULL;
	u8			index;
	u8			channel = 0;

	for(index = 0; index < 7; index++){
		tx_power = (struct _eeprom_txpower*)&data;
		adapter->capability->channels[channel++].tx_power = tx_power->tx_power1;
		adapter->capability->channels[channel++].tx_power = tx_power->tx_power2;
		data++;
	}

	/*
	 * Only 14 tx powers are stored in the eeprom.
	 * In case of 802.11a capable devices we should write to the remaining channels the tx powers for channel 1.
	 */
	for(index = 14; index < MAX_FREQUENCIES; index++)
		adapter->capability->channels[index].tx_power = adapter->capability->channels[0].tx_power;
}


char rt2x00_init_adapter_eeprom(struct _rt2x00_adapter* adapter)
{
	u16			eeprom[26];
	struct _eeprom_bbp	*bbp = NULL;
	u8			index;
	u8			max;

	/*
	 * Set the range of the EEPROM data that has to be read.
	 */
	if(adapter->chip_adapter == RT2400){
		index = 0x16;
		max = 15;	/* 8 PPB values + 7 tx power values. */
	} else {
		index = 0x20;
		max = 26;	/* 19 PPB values + 7 tx power values. */
	}

	write_lock(&adapter->config_lock);

	/*
	 * Read EEPROM into array.
	 */
	rt2x00_eeprom_regread(adapter, index, max, &eeprom[0]);

	/*
	 * Print contents of EEPROM to log file.
	 */
	INFO("Dumping EEPROM contents...\n");
	INFO("%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
		eeprom[0], eeprom[1], eeprom[2], eeprom[3], eeprom[4], eeprom[5], eeprom[6], eeprom[7]);
	INFO("%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
		eeprom[8], eeprom[9], eeprom[10], eeprom[11], eeprom[12], eeprom[13], eeprom[14], eeprom[15]);
	if(adapter->chip_adapter == RT2500){
		INFO("%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
			eeprom[16], eeprom[17], eeprom[18], eeprom[19], eeprom[20], eeprom[21], eeprom[22], eeprom[23]);
		INFO("%04x:%04x\n",
			eeprom[24], eeprom[25]);
	}
	INFO("...End of EEPROM contents.\n");

	/*
	 * The first value contains the antenna data.
	 */
	rt2x00_init_adapter_eeprom_antenna(adapter, eeprom[0]);

	/*
	 * We continue with the BBP section of the EEPROM into the registers.
	 * For RT devices however we should start reading at the 3rd entry.
	 */
	if(adapter->chip_adapter == RT2400){
		index = 1;
		max = 7;
	} else {
		index = 2;
		max = 17;
	}

	/*
	 * We moved to the correct position in the array, start writing the registers.
	 */
	for(; index <= max; index++){
		if(eeprom[index] != 0xffff && eeprom[index] != 0x0000){
			bbp = (struct _eeprom_bbp*)&eeprom[index];
			rt2x00_bbp_regwrite(adapter, bbp->reg_id, bbp->value);
		}
	}

	/*
	 * Start loading tx_power data.
	 */
	rt2x00_init_adapter_eeprom_txpower(adapter, &eeprom[index]);

	write_unlock(&adapter->config_lock);

	INFO("EEPROM configuration loaded.\n");

	return 0;
}

void rt2x00_init_adapter_capable_rate(struct _rt2x00_adapter* adapter)
{
	/* 
	 * 802.11g WIFI spec disallow OFDM rates in 802.11g ADHOC mode
	 * make sure 1, 2, 5.5 and 11 are the first 4 rates in the array.
	 */
	adapter->capability->bit_rates[0].value = 2;
	adapter->capability->bit_rates[0].plcp = 0x00;
	adapter->capability->bit_rates[1].value = 4;
	adapter->capability->bit_rates[1].plcp = 0x01;
	adapter->capability->bit_rates[2].value = 11;
	adapter->capability->bit_rates[2].plcp = 0x02;
	adapter->capability->bit_rates[3].value = 22;
	adapter->capability->bit_rates[3].plcp = 0x03;

	adapter->capability->size_bit_rates = 4;

	if(adapter->chip_adapter == RT2500){
		adapter->capability->bit_rates[4].value = 12;
		adapter->capability->bit_rates[4].plcp =  0x0b;
		adapter->capability->bit_rates[5].value = 18;
		adapter->capability->bit_rates[5].plcp = 0x0f;
		adapter->capability->bit_rates[6].value = 24;
		adapter->capability->bit_rates[6].plcp = 0x0a;
		adapter->capability->bit_rates[7].value = 36;
		adapter->capability->bit_rates[7].plcp = 0x0e;
		adapter->capability->bit_rates[8].value = 48;
		adapter->capability->bit_rates[8].plcp = 0x09;
		adapter->capability->bit_rates[9].value = 72;
		adapter->capability->bit_rates[9].plcp = 0x0d;
		adapter->capability->bit_rates[10].value = 96;
		adapter->capability->bit_rates[10].plcp = 0x08;
		adapter->capability->bit_rates[11].value = 108;
		adapter->capability->bit_rates[11].plcp = 0x0c;

		adapter->capability->size_bit_rates = 12;
	}
}

void rt2x00_init_adapter_capable_txpower(struct _rt2x00_adapter* adapter)
{
	/* I am reduucing these untill I have had a propper look at what is going on
	my rt2400 card has died and I don't want our code to kill anybody elses card
	(it may have had nothing to do with it. But lets play safe */
	
	adapter->capability->tx_powers[0] = 1;
	adapter->capability->tx_powers[1] = 5;
	adapter->capability->tx_powers[2] = 20;
	adapter->capability->tx_powers[3] = 25;
	adapter->capability->tx_powers[4] = 28;
	adapter->capability->tx_powers[5] = 28;
	adapter->capability->size_tx_powers = 6;
}

void rt2x00_init_adapter_capable_rf_set1(struct _rt2x00_adapter* adapter)
{
	/*
	 * RF register values for RT2400 and RT2522 devices.
	 */
	adapter->capability->channels[0].rf = 0x940c1fda;
	adapter->capability->channels[1].rf = 0x940c1fee;
	adapter->capability->channels[2].rf = 0x940c2002;
	adapter->capability->channels[3].rf = 0x940c2016;
	adapter->capability->channels[4].rf = 0x940c202a;
	adapter->capability->channels[5].rf = 0x940c203e;
	adapter->capability->channels[6].rf = 0x940c2052;
	adapter->capability->channels[7].rf = 0x940c2066;
	adapter->capability->channels[8].rf = 0x940c207a;
	adapter->capability->channels[9].rf = 0x940c208e;
	adapter->capability->channels[10].rf = 0x940c20a2;
	adapter->capability->channels[11].rf = 0x940c20b6;
	adapter->capability->channels[12].rf = 0x940c20ca;
	adapter->capability->channels[13].rf = 0x940c20fa;
}

void rt2x00_init_adapter_capable_rf_set2(struct _rt2x00_adapter* adapter)
{
	/*
	 * RF register values for RT2523 and RT2524 devices.
	 */
	adapter->capability->channels[0].rf = 0x94000c9e;
	adapter->capability->channels[1].rf = 0x94000ca2;
	adapter->capability->channels[2].rf = 0x94000ca6;
	adapter->capability->channels[3].rf = 0x94000caa;
	adapter->capability->channels[4].rf = 0x94000cae;
	adapter->capability->channels[5].rf = 0x94000cb2;
	adapter->capability->channels[6].rf = 0x94000cb6;
	adapter->capability->channels[7].rf = 0x94000cba;
	adapter->capability->channels[8].rf = 0x94000cbe;
	adapter->capability->channels[9].rf = 0x94000d02;
	adapter->capability->channels[10].rf = 0x94000d06;
	adapter->capability->channels[11].rf = 0x94000d0a;
	adapter->capability->channels[12].rf = 0x94000d0e;
	adapter->capability->channels[13].rf = 0x94000d1a;
}

void rt2x00_init_adapter_capable_rf_set3(struct _rt2x00_adapter* adapter)
{
	/*
	 * RF register values for RT2525 devices.
	 */
	adapter->capability->channels[0].rf = 0x94080c9e;
	adapter->capability->channels[1].rf = 0x94080ca2;
	adapter->capability->channels[2].rf = 0x94080ca6;
	adapter->capability->channels[3].rf = 0x94080caa;
	adapter->capability->channels[4].rf = 0x94080cae;
	adapter->capability->channels[5].rf = 0x94080cb2;
	adapter->capability->channels[6].rf = 0x94080cb6;
	adapter->capability->channels[7].rf = 0x94080cba;
	adapter->capability->channels[8].rf = 0x94080cbe;
	adapter->capability->channels[9].rf = 0x94080d02;
	adapter->capability->channels[10].rf = 0x94080d06;
	adapter->capability->channels[11].rf = 0x94080d0a;
	adapter->capability->channels[12].rf = 0x94080d0e;
	adapter->capability->channels[13].rf = 0x94080d1a;
}

void rt2x00_init_adapter_capable_rf_set4(struct _rt2x00_adapter* adapter)
{
	/*
	 * RF register values for RT2525E devices.
	 */
	adapter->capability->channels[0].rf = 0x94081136;
	adapter->capability->channels[1].rf = 0x9408113a;
	adapter->capability->channels[2].rf = 0x9408113e;
	adapter->capability->channels[3].rf = 0x94081182;
	adapter->capability->channels[4].rf = 0x94081186;
	adapter->capability->channels[5].rf = 0x9408118a;
	adapter->capability->channels[6].rf = 0x9408118e;
	adapter->capability->channels[7].rf = 0x94081192;
	adapter->capability->channels[8].rf = 0x94081196;
	adapter->capability->channels[9].rf = 0x9408119a;
	adapter->capability->channels[10].rf = 0x9408119e;
	adapter->capability->channels[11].rf = 0x940811a2;
	adapter->capability->channels[12].rf = 0x940811a6;
	adapter->capability->channels[13].rf = 0x940811ae;
}

void rt2x00_init_adapter_capable_rf_set5(struct _rt2x00_adapter* adapter)
{
	/*
	 * RF register values for RT5222 devices.
	 */
	adapter->capability->channels[0].rf = 0x94001136;
	adapter->capability->channels[1].rf = 0x9400113a;
	adapter->capability->channels[2].rf = 0x9400113e;
	adapter->capability->channels[3].rf = 0x94001182;
	adapter->capability->channels[4].rf = 0x94001186;
	adapter->capability->channels[5].rf = 0x9400118a;
	adapter->capability->channels[6].rf = 0x9400118e;
	adapter->capability->channels[7].rf = 0x94001192;
	adapter->capability->channels[8].rf = 0x94001196;
	adapter->capability->channels[9].rf = 0x9400119a;
	adapter->capability->channels[10].rf = 0x9400119e;
	adapter->capability->channels[11].rf = 0x940011a2;
	adapter->capability->channels[12].rf = 0x940011a6;
	adapter->capability->channels[13].rf = 0x940011ae;
	adapter->capability->channels[14].rf = 0x94018896;
	adapter->capability->channels[15].rf = 0x9401889a;
	adapter->capability->channels[16].rf = 0x9401889e;
	adapter->capability->channels[17].rf = 0x940188a2;
	adapter->capability->channels[18].rf = 0x940188a6;
	adapter->capability->channels[19].rf = 0x940188aa;
	adapter->capability->channels[20].rf = 0x940188ae;
	adapter->capability->channels[21].rf = 0x940188b2;
	adapter->capability->channels[22].rf = 0x94008802;
	adapter->capability->channels[23].rf = 0x94008806;
	adapter->capability->channels[24].rf = 0x9400880a;
	adapter->capability->channels[25].rf = 0x9400880e;	
	adapter->capability->channels[26].rf = 0x94008812;
	adapter->capability->channels[27].rf = 0x94008816;
	adapter->capability->channels[28].rf = 0x9400881a;
	adapter->capability->channels[29].rf = 0x9400881e;
	adapter->capability->channels[30].rf = 0x94008822;
	adapter->capability->channels[31].rf = 0x94008826;
	adapter->capability->channels[32].rf = 0x9400882a;
	adapter->capability->channels[33].rf = 0x940090a6;
	adapter->capability->channels[34].rf = 0x940090ae;
	adapter->capability->channels[35].rf = 0x940090b6;
	adapter->capability->channels[36].rf = 0x940090be;
}

void rt2x00_init_adapter_capable_channel(struct _rt2x00_adapter* adapter)
{
	/* OFDM */
	adapter->capability->channels[0].value = 1;
	adapter->capability->channels[0].freq = 2412;
	adapter->capability->channels[1].value = 2;
	adapter->capability->channels[1].freq = 2417;
	adapter->capability->channels[2].value = 3;
	adapter->capability->channels[2].freq = 2422;
	adapter->capability->channels[3].value = 4;
	adapter->capability->channels[3].freq = 2427;
	adapter->capability->channels[4].value = 5;
	adapter->capability->channels[4].freq = 2432;
	adapter->capability->channels[5].value = 6;
	adapter->capability->channels[5].freq = 2437;
	adapter->capability->channels[6].value = 7;
	adapter->capability->channels[6].freq = 2442;
	adapter->capability->channels[7].value = 8;
	adapter->capability->channels[7].freq = 2447;
	adapter->capability->channels[8].value = 9;
	adapter->capability->channels[8].freq = 2452;
	adapter->capability->channels[9].value = 10;
	adapter->capability->channels[9].freq = 2457;
	adapter->capability->channels[10].value = 11;
	adapter->capability->channels[10].freq = 2462;
	adapter->capability->channels[11].value = 12;
	adapter->capability->channels[11].freq = 2467;
	adapter->capability->channels[12].value = 13;
	adapter->capability->channels[12].freq = 2472;
	adapter->capability->channels[13].value = 14;
	adapter->capability->channels[13].freq = 2484;

	adapter->capability->size_channels = 14;

	/* In 802.11a we have more channels we an connect to. */
	if(adapter->chip_rf_type == RT5222){
		/* UNII */
		adapter->capability->channels[14].value = 36;
		adapter->capability->channels[14].freq = 5180;
		adapter->capability->channels[15].value = 40;
		adapter->capability->channels[15].freq = 5200;
		adapter->capability->channels[16].value = 44;
		adapter->capability->channels[16].freq = 5220;
		adapter->capability->channels[17].value = 48;
		adapter->capability->channels[17].freq = 5240;
		adapter->capability->channels[18].value = 52;
		adapter->capability->channels[18].freq = 5260;
		adapter->capability->channels[19].value = 56;
		adapter->capability->channels[19].freq = 5280;
		adapter->capability->channels[20].value = 60;
		adapter->capability->channels[20].freq = 5300;
		adapter->capability->channels[21].value = 64;
		adapter->capability->channels[21].freq = 5320;
		adapter->capability->channels[22].value = 149;
		adapter->capability->channels[22].freq = 5745;
		adapter->capability->channels[23].value = 153;
		adapter->capability->channels[23].freq = 5765;
		adapter->capability->channels[24].value = 157;
		adapter->capability->channels[24].freq = 5785;
		adapter->capability->channels[25].value = 161;
		adapter->capability->channels[25].freq = 5805;
		/* HiperLAN2 */
		adapter->capability->channels[26].value = 100;
		adapter->capability->channels[26].freq = 5500;
		adapter->capability->channels[27].value = 104;
		adapter->capability->channels[27].freq = 5520;
		adapter->capability->channels[28].value = 108;
		adapter->capability->channels[28].freq = 5540;
		adapter->capability->channels[29].value = 112;
		adapter->capability->channels[29].freq = 5560;
		adapter->capability->channels[30].value = 116;
		adapter->capability->channels[30].freq = 5580;
		adapter->capability->channels[31].value = 120;
		adapter->capability->channels[31].freq = 5600;
		adapter->capability->channels[32].value = 124;
		adapter->capability->channels[32].freq = 5620;
		adapter->capability->channels[33].value = 128;
		adapter->capability->channels[33].freq = 5640;
		adapter->capability->channels[34].value = 132;
		adapter->capability->channels[34].freq = 5660;
		adapter->capability->channels[35].value = 136;
		adapter->capability->channels[35].freq = 5680;
		adapter->capability->channels[36].value = 140;
		adapter->capability->channels[36].freq = 5700;

		adapter->capability->size_channels = 37;
	}
}

char rt2x00_init_adapter_capable(struct _rt2x00_adapter* adapter)
{
	write_lock(&adapter->config_lock);

	rt2x00_init_adapter_capable_rate(adapter);
	rt2x00_init_adapter_capable_txpower(adapter);
	rt2x00_init_adapter_capable_channel(adapter);

	/*
	 * Load RF register values depending on device.
	 */
	if(adapter->chip_adapter == RT2400 || adapter->chip_rf_type == RT2522)
		rt2x00_init_adapter_capable_rf_set1(adapter);
	else if(adapter->chip_rf_type == RT2523 ||adapter->chip_rf_type == RT2524)
		rt2x00_init_adapter_capable_rf_set2(adapter);
	else if(adapter->chip_rf_type == RT2525)
		rt2x00_init_adapter_capable_rf_set3(adapter);
	else if(adapter->chip_rf_type == RT2525E)
		rt2x00_init_adapter_capable_rf_set4(adapter);
	else if(adapter->chip_rf_type == RT5222)
		rt2x00_init_adapter_capable_rf_set5(adapter);

	if(adapter->chip_adapter == RT2400){
		SET_FLAG(adapter, NETWORK_802_11_B);
		INFO("IEEE 802.11b capabilities loaded.\n");
	} else if(adapter->chip_adapter == RT2500 && adapter->chip_rf_type != RT5222){
		SET_FLAG(adapter, NETWORK_802_11_BG);
		INFO("IEEE 802.11b/g capabilities loaded.\n");
	} else {
		SET_FLAG(adapter, NETWORK_802_11_ABG);
		INFO("IEEE 802.11a/b/g capabilities loaded.\n");
	}

	write_unlock(&adapter->config_lock);

	return 0;
} 

char rt2x00_init_adapter_config(struct _rt2x00_adapter* adapter)
{
	write_lock(&adapter->config_lock);

	SET_FLAG(adapter, ACCEPT_UNICAST | ACCEPT_MULTICAST);

	adapter->config->config_mode = IW_MODE_AUTO;
	adapter->config->preamble = PREAMBLE_LONG_DURATION;
	adapter->config->tx_power = adapter->capability->size_tx_powers - 1;
	adapter->config->max_rate = adapter->capability->size_bit_rates - 1;
	adapter->config->cur_rate = adapter->capability->size_bit_rates - 1;
	adapter->config->short_retry = 4;
	adapter->config->long_retry = 7;
	adapter->config->channel = 0;
	adapter->config->sensitivity = SENSITIVITY_DEFAULT;
	adapter->config->fragmentation = MAX_FRAGMENTATION;
	adapter->config->rts_treshold = MAX_RTS_THRESHOLD;
	adapter->config->beacon = 100;
	adapter->config->plcp = 48;
	adapter->config->sifs = 10;
	adapter->config->slot_time = (TEST_FLAG(adapter, NETWORK_802_11_A)) ? 9 : 20;

	INFO("Configuration loaded.\n");

	write_unlock(&adapter->config_lock);

	return 0;
}

char rt2x00_init_adapter_register(struct _rt2x00_adapter* adapter)
{
	u8		bbp_ready = 0xff;
	u32		counter;

	write_lock(&adapter->config_lock);

	rt2x00_register_write(adapter, CSR9, TX_BUFFER_SIZE);	/* Max frame lenght. */

	if(adapter->chip_adapter == RT2400){
		/*
		 * RT2400 activate BBP registers: R36:RSSI + R32:SIGNAL + R3:TXPOWER.
		 */
		rt2x00_register_write(adapter, RXCSR3, 0x0083a4a0);
		/*
		 * Signal: 0x85 service: 0x86 low byte: 0x88 high byte: 0x87.
		 */
		rt2x00_register_write(adapter, ARCSR0, 0x87888685);
	} else {
		/*
		 * RT2500 activate BBP registers: R51:RSSI + R42:OFDM RATE + R47:SIGNAL.
		 */
		rt2x00_register_write(adapter, RXCSR3, 0x00aab3af);
		/*
		 * Signal: 0x8a service: 0x8b low byte: 0x8d high byte: 0x8c.
		 */
		rt2x00_register_write(adapter, ARCSR0, 0x8c8d8b8a);
	}

	/*
	 * Signal qualification count: RX:8 TX:63.
	 */
	rt2x00_register_write(adapter, CNT3, 0x3f080000);

	rt2x00_register_write(adapter, PWRCSR0, 0x3f3b3100);

	/*
	 * All delays to a different device state (sleep to deepsleep, etc.)
	 * will take 2048 us.
	 */
	rt2x00_register_write(adapter, PSCSR0, 0x00020002);
	rt2x00_register_write(adapter, PSCSR1, 0x00020002);
	rt2x00_register_write(adapter, PSCSR2, 0x00020002);
	rt2x00_register_write(adapter, PSCSR3, 0x00000002);

	/*
	 * Set BBP and RF registers to awake.
	 */
	rt2x00_register_write(adapter, PWRCSR1, 0x000001ff);

	/*
	 * Beacon expect window: 1/2 beacon interval.
	 */
	rt2x00_register_write(adapter, TIMECSR, 0x00003f21);

	/*
	 * Enable TX CRC and dribble. 1/4 system clock freq for BBP/RF register access.
	 */
	rt2x00_register_write(adapter, MACCSR0, 0x00213223);
	rt2x00_register_write(adapter, MACCSR1, 0x00235518);

	/*
	 * On reset it will write 0x11 to R26.
	 * This will enable the RX soft reset and the global reset.
	 */
	rt2x00_register_write(adapter, RALINKCSR, 0x9a009a11);

	/*
	 * RX_PE Low Width: 64 pci clock cycles.
	 * 64 * 33ns = 2us.
	 */
	rt2x00_register_write(adapter, MACCSR2, 0x00000040);

	
	/* Kick host before starting rx*/
	rt2x00_register_write(adapter, CSR1, 0x00000001);
	rt2x00_register_write(adapter, CSR1, 0x00000004);

	/*
	 * Before accessing the BBP registers, we have to make sure the register is ready.
	 * There is however a limit on the busy time of the bbp register.
	 */
	counter = MAX_REGISTER_BUSY_COUNT;
	rt2x00_bbp_regread(adapter, 0, &bbp_ready);
	while(((bbp_ready == 0xff) || (bbp_ready == 0x00)) && counter > 0){
		udelay(REGISTER_BUSY_DELAY);
		rt2x00_bbp_regread(adapter, 0, &bbp_ready);
		NOTICE("Waiting for BBP register.\n");
		counter--;
	}

	/*
	 * Did we time out before the BBP register was ready?
	 */
	if(counter == 0)
		goto exit_bbp_fail;

	if(adapter->chip_adapter == RT2400){
		rt2x00_bbp_regwrite(adapter, 10, 0x0f);	/* rx cca agc. */
		rt2x00_bbp_regwrite(adapter, 13, 0x08);	/* rx cca agc init. */
		rt2x00_bbp_regwrite(adapter, 15, 0x72);	/* agc lna on 40dbm. */
		rt2x00_bbp_regwrite(adapter, 16, 0x74);	/* agc lna off 0dbm. */
		rt2x00_bbp_regwrite(adapter, 17, 0x20);	/* agc lna off 40dbm. */
		rt2x00_bbp_regwrite(adapter, 18, 0x72);	/* cca antenna. */
		rt2x00_bbp_regwrite(adapter, 19, 0x0b);	/* cca course. */
		rt2x00_bbp_regwrite(adapter, 20, 0x00);	/* cca fine sq. */
		rt2x00_bbp_regwrite(adapter, 28, 0x11);	/* rx tracking. */
		rt2x00_bbp_regwrite(adapter, 29, 0x04);	/* rx adaptive. */
		rt2x00_bbp_regwrite(adapter, 30, 0x21);	/* timing control. */
		rt2x00_bbp_regwrite(adapter, 31, 0x00);	/* adaptive gear down. */

	} else {
		/*
		 * FIXME - Unfortunatly the BBP registers for RT2500 are undocumented.
		 * This means we have no idea what the following values actually mean.
		 * As soon as we receive further documentation this list has to be investigated.
		 */
		rt2x00_bbp_regwrite(adapter, 3, 0x02);
		rt2x00_bbp_regwrite(adapter, 4, 0x19);
		rt2x00_bbp_regwrite(adapter, 14, 0x1c);
		rt2x00_bbp_regwrite(adapter, 15, 0x30);
		rt2x00_bbp_regwrite(adapter, 16, 0xac);
		rt2x00_bbp_regwrite(adapter, 17, 0x48);
		rt2x00_bbp_regwrite(adapter, 18, 0x18);
		rt2x00_bbp_regwrite(adapter, 19, 0xff);
		rt2x00_bbp_regwrite(adapter, 20, 0x1e);
		rt2x00_bbp_regwrite(adapter, 21, 0x08);
		rt2x00_bbp_regwrite(adapter, 22, 0x08);
		rt2x00_bbp_regwrite(adapter, 23, 0x08);
		rt2x00_bbp_regwrite(adapter, 24, 0x70);
		rt2x00_bbp_regwrite(adapter, 25, 0x40);
		rt2x00_bbp_regwrite(adapter, 26, 0x08);
		rt2x00_bbp_regwrite(adapter, 27, 0x23);
		rt2x00_bbp_regwrite(adapter, 30, 0x10);
		rt2x00_bbp_regwrite(adapter, 31, 0x2b);
		rt2x00_bbp_regwrite(adapter, 32, 0xb9);
		rt2x00_bbp_regwrite(adapter, 34, 0x12);
		rt2x00_bbp_regwrite(adapter, 35, 0x50);
		rt2x00_bbp_regwrite(adapter, 39, 0xc4);
		rt2x00_bbp_regwrite(adapter, 40, 0x02);
		rt2x00_bbp_regwrite(adapter, 41, 0x60);
		rt2x00_bbp_regwrite(adapter, 53, 0x10);
		rt2x00_bbp_regwrite(adapter, 54, 0x18);
		rt2x00_bbp_regwrite(adapter, 56, 0x08);
		rt2x00_bbp_regwrite(adapter, 57, 0x10);
		rt2x00_bbp_regwrite(adapter, 58, 0x08);
		rt2x00_bbp_regwrite(adapter, 61, 0x6d);
		rt2x00_bbp_regwrite(adapter, 62, 0x10);
	}

	/*
	 * We must clear the FCS and FIFI error count.
	 * These registers are cleared on read, so we may pass a useless variable to store the value.
	 */
	rt2x00_register_read(adapter, CNT0, &counter);
	rt2x00_register_read(adapter, CNT4, &counter);

	write_unlock(&adapter->config_lock);

	INFO("Registers initialized.\n");

	return 0;

exit_bbp_fail:
	ERROR("Harware problem, BBP was never ready, aborting.\n");

	write_unlock(&adapter->config_lock);

	return -EBUSY;
}

char rt2x00_init_schedule_irq(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);

	tasklet_init(&adapter->tx_tasklet, rt2x00_delayed_start_tx, (unsigned long)adapter);
	tasklet_init(&adapter->prio_tasklet, rt2x00_delayed_start_tx_prio, (unsigned long)adapter);
	tasklet_init(&adapter->rx_tasklet, rt2x00_delayed_handle_rx, (unsigned long)adapter);
	
	/* Set the tasklets count to 1 so that the first enable will set it back to 0 (running) */
	tasklet_disable(&adapter->tx_tasklet);
	tasklet_disable(&adapter->prio_tasklet);
	tasklet_disable(&adapter->rx_tasklet);

	spin_unlock(&adapter->adapter_lock);

	INFO("Tasklets initialized.\n");

	return 0;
}

void rt2x00_init_rx_ring(struct _rt2x00_adapter* adapter)
{
	u8		index;
	void		*desc = adapter->dma_mem->rx_ring.desc_addr;
	void		*data = adapter->dma_mem->rx_ring.data_addr;
	dma_addr_t	data_dma = adapter->dma_mem->rx_data_dma;
	struct _rxd	rxd_desc;
	struct _rxd	*rxd;

	spin_lock(&adapter->rx_lock);

	memset(&rxd_desc, 0x00, sizeof(struct _rxd));

	/*
	 * These values are the same for all descriptors, so write these only once.
	 */
	if(adapter->chip_adapter == RT2400)
		rxd_desc.chip.rt2400.buffer_length = RX_BUFFER_SIZE;
	rxd_desc.owner = DESC_OWNED_BY_NIC;		/* Rx ring is held by nic. */

	/* Rx ring & DMA buffer */
	for (index = 0; index < RX_RING_SIZE; index++){
		adapter->dma->rx_ring[index].desc_addr = desc;
		adapter->dma->rx_ring[index].data_addr = data;

		rxd = (struct _rxd*) adapter->dma->rx_ring[index].desc_addr;
		rxd_desc.buffer_address = data_dma;
		/*
		 * Convert the descriptor to 32bit little endian and give it to the adapter.
		 */
		cpu_to_device(rxd, &rxd_desc, sizeof(rxd_desc));

		desc += RING_DESCRIPTOR_SIZE;
		data += RX_BUFFER_SIZE;
		data_dma += RX_BUFFER_SIZE;
	}

	/* Write RXCSR2 register */
	rt2x00_register_write(adapter, RX_RING_REGISTER, adapter->dma_mem->rx_desc_dma);

	skb_queue_head_init(&adapter->rx_packet_queue);
	skb_queue_head_init(&adapter->frag_packet_queue);

	spin_unlock(&adapter->rx_lock);

	INFO("Rx ring initialized.\n");
}

void rt2x00_init_tx_ring(struct _rt2x00_adapter* adapter)
{
	u8		index;
	void		*desc = adapter->dma_mem->tx_ring.desc_addr;
	void		*data = adapter->dma_mem->tx_ring.data_addr;
	dma_addr_t	data_dma = adapter->dma_mem->tx_data_dma;
	struct _txd	txd_desc;
	struct _txd	*txd;

	spin_lock(&adapter->tx_lock);

	memset(&txd_desc, 0x00, sizeof(struct _txd));

	/*
	 * These values are the same for all descriptors, so write these only once.
	 */
	if(adapter->chip_adapter == RT2400)
		txd_desc.chip.rt2400.buffer_length = TX_BUFFER_SIZE;
	txd_desc.owner = DESC_OWNED_BY_HOST;	/* tx ring is held by host. */

	/* Tx ring & DMA buffer */
	for (index = 0; index < TX_RING_SIZE; index++){
		adapter->dma->tx_ring[index].desc_addr = desc;
		adapter->dma->tx_ring[index].data_addr = data;

		txd = (struct _txd*) adapter->dma->tx_ring[index].desc_addr;
		txd_desc.buffer_address = data_dma;
		/*
		 * Convert the descriptor to 32bit little endian and give it to the adapter.
		 */
		cpu_to_device(txd, &txd_desc, sizeof(txd_desc));

		desc += RING_DESCRIPTOR_SIZE;
		data += TX_BUFFER_SIZE;
		data_dma += TX_BUFFER_SIZE;
	}

	/* Write TXCSR3 register */
	rt2x00_register_write(adapter, TX_RING_REGISTER, adapter->dma_mem->tx_desc_dma);

	skb_queue_head_init(&adapter->tx_packet_queue);

	spin_unlock(&adapter->tx_lock);

	INFO("Tx ring initialized.\n");
}

void rt2x00_init_atim_ring(struct _rt2x00_adapter* adapter)
{
	u8		index;
	void		*desc = adapter->dma_mem->atim_ring.desc_addr;
	void		*data = adapter->dma_mem->atim_ring.data_addr;
	dma_addr_t	data_dma = adapter->dma_mem->atim_data_dma;
	struct _txd	txd_desc;
	struct _txd	*txd;

	spin_lock(&adapter->atim_lock);

	memset(&txd_desc, 0x00, sizeof(struct _txd));

	/*
	 * These values are the same for all descriptors, so write these only once.
	 */
	if(adapter->chip_adapter == RT2400)
		txd_desc.chip.rt2400.buffer_length = ATIM_BUFFER_SIZE;
	txd_desc.owner = DESC_OWNED_BY_HOST;	/* tx ring is held by host. */

	/* Atim ring & DMA buffer */
	for (index = 0; index < ATIM_RING_SIZE; index++){
		adapter->dma->atim_ring[index].desc_addr = desc;
		adapter->dma->atim_ring[index].data_addr = data;

		txd = (struct _txd*) adapter->dma->atim_ring[index].desc_addr;
		txd_desc.buffer_address = data_dma;
		/*
		 * Convert the descriptor to 32bit little endian and give it to the adapter.
		 */
		cpu_to_device(txd, &txd_desc, sizeof(txd_desc));

		desc += RING_DESCRIPTOR_SIZE;
		data += ATIM_BUFFER_SIZE;
		data_dma += ATIM_BUFFER_SIZE;
	}

	/* Write TXCSR4 register */
	rt2x00_register_write(adapter, ATIM_RING_REGISTER, adapter->dma_mem->atim_desc_dma);

	spin_unlock(&adapter->atim_lock);

	INFO("Atim ring initialized.\n");
}

void rt2x00_init_prio_ring(struct _rt2x00_adapter* adapter)
{
	u8		index;
	void		*desc = adapter->dma_mem->prio_ring.desc_addr;
	void		*data = adapter->dma_mem->prio_ring.data_addr;
	dma_addr_t	data_dma = adapter->dma_mem->prio_data_dma;
	struct _txd	txd_desc;
	struct _txd	*txd;

	spin_lock(&adapter->prio_lock);

	memset(&txd_desc, 0x00, sizeof(struct _txd));

	/*
	 * These values are the same for all descriptors, so write these only once.
	 */
	if(adapter->chip_adapter == RT2400)
		txd_desc.chip.rt2400.buffer_length = PRIO_BUFFER_SIZE;
	txd_desc.owner = DESC_OWNED_BY_HOST;	/* tx ring is held by host. */

	/* Prio ring & DMA buffer */
	for (index = 0; index < PRIO_RING_SIZE; index++){
		adapter->dma->prio_ring[index].desc_addr = desc;
		adapter->dma->prio_ring[index].data_addr = data;

		txd = (struct _txd*) adapter->dma->prio_ring[index].desc_addr;
		txd_desc.buffer_address = data_dma;
		/*
		 * Convert the descriptor to 32bit little endian and give it to the adapter.
		 */
		cpu_to_device(txd, &txd_desc, sizeof(txd_desc));

		desc += RING_DESCRIPTOR_SIZE;
		data += PRIO_BUFFER_SIZE;
		data_dma += PRIO_BUFFER_SIZE;
	}

	/* Write TXCSR5 register */
	rt2x00_register_write(adapter, PRIO_RING_REGISTER, adapter->dma_mem->prio_desc_dma);

	skb_queue_head_init(&adapter->prio_packet_queue);

	spin_unlock(&adapter->prio_lock);

	INFO("Prio ring initialized.\n");
}

void rt2x00_init_beacon_ring(struct _rt2x00_adapter* adapter)
{
	struct _txd	txd_desc;
	struct _txd	*txd;

	spin_lock(&adapter->beacon_lock);

	memset(&txd_desc, 0x00, sizeof(struct _txd));

	/* Init Rx DMA buffer */
	adapter->dma->beacon_ring.desc_addr = adapter->dma_mem->beacon_ring.desc_addr;
	adapter->dma->beacon_ring.data_addr = adapter->dma_mem->beacon_ring.data_addr;

	/* Write RxD buffer address & allocated buffer length */
	txd = (struct _txd*) adapter->dma->beacon_ring.desc_addr;
	if(adapter->chip_adapter == RT2400)
		txd_desc.chip.rt2400.buffer_length = BEACON_BUFFER_SIZE;
	txd_desc.owner = DESC_OWNED_BY_HOST;	/* beacon ring is held by host. */
	txd_desc.buffer_address = adapter->dma_mem->beacon_data_dma;

	/*
	 * Convert the descriptor to 32bit little endian and give it to the adapter.
	 */
	cpu_to_device(txd, &txd_desc, sizeof(txd_desc));

	/* Write TXCSR6 register */
	rt2x00_register_write(adapter, BEACON_REGISTER, adapter->dma_mem->beacon_desc_dma);

	spin_unlock(&adapter->beacon_lock);

	INFO("Beacon ring initialized.\n");
}

void rt2x00_init_dma_register(struct _rt2x00_adapter* adapter)
{
	union _txcsr2	txcsr2;
	union _rxcsr1	rxcsr1;

	/* Write TXCSR2 register */
	txcsr2.field.txd_size = RING_DESCRIPTOR_SIZE;
	txcsr2.field.num_txd  = TX_RING_SIZE;
	txcsr2.field.num_atim  = ATIM_RING_SIZE;
	txcsr2.field.num_prio  = PRIO_RING_SIZE;	
	rt2x00_register_write(adapter, TXCSR2, txcsr2.word);

	/* Write RXCSR1 register */
	rxcsr1.field.rxd_size = RING_DESCRIPTOR_SIZE;
	rxcsr1.field.num_rxd  = RX_RING_SIZE;
	rt2x00_register_write(adapter, RXCSR1, rxcsr1.word);
	
	INFO("DMA registers initialized.\n");
}

char rt2x00_init_dma_ring(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	rt2x00_init_rx_ring(adapter);
	rt2x00_init_tx_ring(adapter);
	rt2x00_init_atim_ring(adapter);
	rt2x00_init_prio_ring(adapter);
	rt2x00_init_beacon_ring(adapter);

	rt2x00_init_dma_register(adapter);

	INFO("DMA memory allocated.\n");

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	return 0;
}

void rt2x00_cleanup_rx_ring(struct _rt2x00_adapter* adapter)
{
	struct _rxd	*rxd;
	u8		index;

	spin_lock(&adapter->rx_lock);

	for(index = 0; index < RX_RING_SIZE; index++){
		rxd = (struct _rxd*) adapter->dma->rx_ring[index].desc_addr;
	
		/* Re-initial Rx ring cell to owned by NIC. */
		rxd->owner = DESC_OWNED_BY_NIC;
	}

	adapter->dma->rx_ring_index = 0;
	adapter->dma->rx_ring_done_index = 0;

	skb_queue_purge(&adapter->rx_packet_queue);

	spin_unlock(&adapter->rx_lock);
}

void rt2x00_cleanup_tx_ring(struct _rt2x00_adapter* adapter)
{
	struct _txd	*txd;
	u8		index;

	spin_lock(&adapter->tx_lock);

	for(index = 0; index < TX_RING_SIZE; index++){
		txd  = (struct _txd *) adapter->dma->tx_ring[index].desc_addr;

		txd->owner = DESC_OWNED_BY_HOST;
		txd->valid = 0;
	}

	adapter->dma->tx_ring_index = 0;
	adapter->dma->tx_ring_done_index = 0;

	skb_queue_purge(&adapter->tx_packet_queue);

	spin_unlock(&adapter->tx_lock);
}

void rt2x00_cleanup_atim_ring(struct _rt2x00_adapter* adapter)
{
	struct _txd	*txd;
	u8		index;

	spin_lock(&adapter->atim_lock);

	for(index = 0; index < ATIM_RING_SIZE; index++){
		txd  = (struct _txd*) adapter->dma->atim_ring[index].desc_addr;
	
		/* We just re-claim these ring spaces. */
		txd->owner = DESC_OWNED_BY_HOST;
		txd->valid = 0;
	}

	adapter->dma->atim_ring_index = 0;
	adapter->dma->atim_ring_done_index = 0;

	spin_unlock(&adapter->atim_lock);
}

void rt2x00_cleanup_prio_ring(struct _rt2x00_adapter* adapter)
{
	struct _txd	*txd;
	u8		index;

	spin_lock(&adapter->prio_lock);

	for(index = 0; index < PRIO_RING_SIZE; index++){
		txd  = (struct _txd*) adapter->dma->prio_ring[index].desc_addr;
	
		/* We just re-claim these ring spaces. */
		txd->owner = DESC_OWNED_BY_HOST;
		txd->valid = 0;
	}

	adapter->dma->prio_ring_index = 0;
	adapter->dma->prio_ring_done_index = 0;

	skb_queue_purge(&adapter->prio_packet_queue);

	spin_unlock(&adapter->prio_lock);
}

void rt2x00_cleanup_beacon_ring(struct _rt2x00_adapter* adapter)
{
	struct _txd	*txd = adapter->dma->beacon_ring.desc_addr;

	spin_lock(&adapter->beacon_lock);
	
	/* We just re-claim these ring spaces. */
	txd->owner = DESC_OWNED_BY_HOST;
	txd->valid = 0;

	spin_unlock(&adapter->beacon_lock);
}

void rt2x00_cleanup_dma_ring(struct _rt2x00_adapter* adapter, u8 ring_type)
{
	if(ring_type & RX_RING)
		rt2x00_cleanup_rx_ring(adapter);

	if(ring_type & TX_RING)
		rt2x00_cleanup_tx_ring(adapter);

	if(ring_type & ATIM_RING)
		rt2x00_cleanup_atim_ring(adapter);

	if(ring_type & PRIO_RING)
		rt2x00_cleanup_prio_ring(adapter);

	if(ring_type & BEACON_RING)
		rt2x00_cleanup_beacon_ring(adapter);
}

/*
 * Some IOCTL calls, will enqueue their configuration change.
 * This because it would require the adapter to reset itself.
 * rt2x00_load_config() is called to load the pending configuration
 * into the adapter config structure.
 * This function should be called with the config_lock (write) held.
 */
static inline void rt2x00_load_config_mode(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_MODE))
		memcpy(&adapter->config->config_mode, &adapter->config->commits->mode, sizeof(u8));
}

static inline void rt2x00_load_config_network(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_NETWORK)){
		if(adapter->config->commits->network & NETWORK_A)
			SET_FLAG(adapter, NETWORK_802_11_A);
		else
			CLEAR_FLAG(adapter, NETWORK_802_11_A);
		if(adapter->config->commits->network & NETWORK_B)
			SET_FLAG(adapter, NETWORK_802_11_B);
		else
			CLEAR_FLAG(adapter, NETWORK_802_11_B);
		if(adapter->config->commits->network & NETWORK_G)
			SET_FLAG(adapter, NETWORK_802_11_G);
		else
			CLEAR_FLAG(adapter, NETWORK_802_11_G);
	}
}

static inline void rt2x00_load_config_channel(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_CHANNEL))
		memcpy(&adapter->config->channel, &adapter->config->commits->channel, sizeof(u8));
}

static inline void rt2x00_load_config_essid(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_ESSID)){
		memcpy(&adapter->config->essid, &adapter->config->commits->essid, IW_ESSID_MAX_SIZE);
		CLEAR_FLAG(adapter, FIX_BSSID);
	}
}

static inline void rt2x00_load_config_crypto(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_CRYPTO)){
		rt2x00_add_key(adapter, &adapter->config->commits->crypto);
		if(!TEST_FLAG(&adapter->config->commits->crypto, IW_ENCODE_DISABLED))
			SET_FLAG(adapter, CRYPTO_ENABLED);
		if(TEST_FLAG(&adapter->config->commits->crypto, IW_ENCODE_RESTRICTED))
			SET_FLAG(adapter, CRYPTO_RESTRICT);
		if(TEST_FLAG(&adapter->config->commits->crypto, IW_ENCODE_OPEN))
			CLEAR_FLAG(adapter, CRYPTO_RESTRICT);
		if(adapter->config->commits->crypto.type & CRYPTO_TYPE_WEP)
			SET_FLAG(adapter, CRYPTO_WEP);
		if(adapter->config->commits->crypto.type & CRYPTO_TYPE_WPA)
			SET_FLAG(adapter, CRYPTO_WPA);
	}
}

static inline void rt2x00_load_config_bssid(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_BSSID)){
		memcpy(&adapter->config->bssid, &adapter->config->commits->bssid, ETH_ALEN);
		SET_FLAG(adapter, FIX_BSSID);
	}
}

static inline void rt2x00_load_config_power_management(struct _rt2x00_adapter* adapter)
{
	if(TEST_FLAG(adapter, COMMIT_POWER)){
		if(adapter->config->commits->power.flags)
			SET_FLAG(adapter, ENABLED_POWER);
		else
			CLEAR_FLAG(adapter, ENABLED_POWER);

		if(TEST_FLAG(&adapter->config->commits->power, IW_POWER_UNICAST_R))
			SET_FLAG(adapter, ACCEPT_UNICAST);
		else
			CLEAR_FLAG(adapter, ACCEPT_UNICAST);

		if(TEST_FLAG(&adapter->config->commits->power, IW_POWER_MULTICAST_R))
			SET_FLAG(adapter, ACCEPT_MULTICAST);
		else
			CLEAR_FLAG(adapter, ACCEPT_MULTICAST);

		if(adapter->config->commits->power.period)
			adapter->config->cfp_period = adapter->config->commits->power.period;
		if(adapter->config->commits->power.timeout)
			adapter->config->atim_period = adapter->config->commits->power.timeout;
	}
}

void rt2x00_load_config(struct _rt2x00_adapter* adapter)
{
	/*
	 * Check if a commit is pending.
	 */
	if(!TEST_FLAG(adapter, COMMIT_PENDING))
		return;

	/*
	 * Altough the chance of this is very, very unlikely,
	 * it is better to check for the presence of adapter->config->commits.
	 */
	if(unlikely(!adapter->config->commits)){
		ERROR("Commits structure not available!\n");
		return;
	}

	INFO("loading configuration changes...\n");

	SET_FLAG(adapter, COMMIT_IN_PROGRESS);

	rt2x00_load_config_mode(adapter);
	rt2x00_load_config_network(adapter);
	rt2x00_load_config_channel(adapter);
	rt2x00_load_config_essid(adapter);
	rt2x00_load_config_crypto(adapter);
	rt2x00_load_config_bssid(adapter);
	rt2x00_load_config_power_management(adapter);

	/*
	 * Don't forget to free the commit structure and set
	   pointer to NULL or we have trouble on the next commit 
	 */
	kfree(adapter->config->commits);
	adapter->config->commits=NULL;

	/*
	 * Clear all commit flags.
	 */
	CLEAR_FLAG(adapter, COMMIT_IN_PROGRESS);
	CLEAR_FLAG(adapter, COMMIT_PENDING);
}

/*
 * After a configuration change rt2x00_update_adapter() should be called.
 * This will check if the configuration is valid.
 * Afterwards the adapter will be updated to load the configuration into the adapter.
 */
void rt2x00_update_config(struct _rt2x00_adapter* adapter)
{
	/*
	 * Altough TX power limits are managed by the capability structure,
	 * it is not depending on network type. This means that the txpower
	 * value cannot become invalid after a network type change.
	 */
	if(adapter->config->max_rate >= adapter->capability->size_bit_rates)
		adapter->config->max_rate = adapter->capability->size_bit_rates - 1;

	if(TEST_FLAG(adapter, NETWORK_802_11_B))
		adapter->capability->basic_rate = 2;
	else
		adapter->capability->basic_rate = 4;

	/*
	 * The first 4 values are the 802.11b rates.
	 * If network type does not have 802.11b we may not use any of the first 4 rates,
	 * instead we use the first not 802.11 specific rate.
	 * When we have a 802.11b only network, we may not use any other rate then the first 4.
	 */
	if(adapter->config->max_rate < 4 && !TEST_FLAG(adapter, NETWORK_802_11_B))
		adapter->config->max_rate = 4;
	else if(adapter->config->max_rate >= 4 && TEST_FLAG(adapter, NETWORK_802_11_B) && !TEST_FLAG(adapter, NETWORK_802_11_G))
		adapter->config->max_rate = 3;

	adapter->config->cur_rate = adapter->config->max_rate;

	/*
	 * 802.11b.g networks only have 14 channels.
	 * Channels higher then 14 are only allowed for 802.11a networks.
	 * size_channels contains the number of channels so we should not exceed that.
	 */

	if(adapter->config->channel > adapter->capability->size_channels)
		adapter->config->channel = adapter->capability->size_channels;

}

void rt2x00_update_txpreamble(struct _rt2x00_adapter* adapter)
{
	/* Short preamble is only available for rates 2, 5.5 and 11 Mbs. */
	if(adapter->config->max_rate == 0 || adapter->config->max_rate > 3){
		if(adapter->config->preamble == PREAMBLE_SHORT){
			INFO("Short preamble not supported with current rate.\n");
			adapter->config->preamble = PREAMBLE_LONG;
		}
	}

	if(adapter->config->preamble == PREAMBLE_SHORT){
		rt2x00_register_write(adapter, ARCSR2, 0x00700408);
		rt2x00_register_write(adapter, ARCSR3, 0x00380409);
		rt2x00_register_write(adapter, ARCSR4, 0x0015040a);
		rt2x00_register_write(adapter, ARCSR5, 0x000b840b);
	} else {
		rt2x00_register_write(adapter, ARCSR2, 0x00700400);
		rt2x00_register_write(adapter, ARCSR3, 0x00380401);
		rt2x00_register_write(adapter, ARCSR4, 0x00150402);
		rt2x00_register_write(adapter, ARCSR5, 0x000b8403);
	}
}

void rt2x00_update_txrate(struct _rt2x00_adapter* adapter)
{

	union _txcsr1	txcsr1;

	txcsr1.field.ack_timeout = adapter->config->sifs
				+ (2 * adapter->config->slot_time)
				+ rt2x00_packet_duration(adapter, adapter->config->max_rate, ACK_SIZE);

	txcsr1.field.ack_consume_time = adapter->config->sifs
					+ rt2x00_packet_duration(adapter, adapter->config->max_rate, ACK_SIZE);

	txcsr1.field.tsf_offset = 0x18;

	txcsr1.field.autoresponder = 1;

	rt2x00_register_write(adapter, TXCSR1, txcsr1.word);

	if(adapter->config->max_rate == 0)
		rt2x00_register_write(adapter, ARCSR1, 0x00000001);	/* Enable rate: 1 Mbps */
	else if(adapter->config->max_rate == 1)
		rt2x00_register_write(adapter, ARCSR1, 0x00000003);	/* Enable rate: 1 & 2 Mbps */
	else if(adapter->config->max_rate == 2)
		rt2x00_register_write(adapter, ARCSR1, 0x00000007);	/* Enable rate: 1 & 2 & 5.5 Mbps */
	else
		rt2x00_register_write(adapter, ARCSR1, 0x0000000f);	/* Enable rate: 1 & 2 & 5.5 & 11 Mbps */
}

void rt2x00_update_txpower(struct _rt2x00_adapter* adapter)
{
	if(adapter->chip_adapter == RT2400)
		rt2x00_bbp_regwrite(adapter, 3, TXPOWER(adapter));

	/*
	 * For RT2500 devices writing the tx_power is a bit different.
	 * The tx_power field in the registers is in the RF register for the channel.
	 * This means that the tx_power is written to the register at the same time as the channel.
	 */
}

void rt2x00_update_power_management(struct _rt2x00_adapter* adapter)
{
	union _csr13	csr13;
	union _csr14	csr14;

	rt2x00_register_read(adapter, CSR13, &csr13.word);
	rt2x00_register_read(adapter, CSR14, &csr14.word);

	/*
	 * Timeout and period units are in 1/16 TU (64 us).
	 */
	if(adapter->config->atim_period){
		csr13.field.atimw_duration = (u16)(adapter->config->atim_period / 16);
	}

	if(adapter->config->cfp_period){
		csr13.field.cfp_period = (u8)(adapter->config->cfp_period / 16);
		csr14.field.cfp_cnt_preload = adapter->config->cfp_count;
	}

	/*
	 * Enable or disable atim and cfp switching depending on value not equal to 0.
	 */
	csr14.field.tatimw = (adapter->config->atim_period) ? 1 : 0;
	csr14.field.tcfp = (adapter->config->cfp_period) ? 1 : 0;

	if(TEST_FLAG(adapter, ENABLED_POWER)){
		adapter->irq_mask &= ~0x06; /* Enable power management. */
		rt2x00_register_write(adapter, CSR13, csr13.word);
		rt2x00_register_write(adapter, CSR14, csr14.word);
		rt2x00_register_write(adapter, TXCSR7, 0x00000001);
	} else {
		adapter->irq_mask |= 0x06; /* Disable power management. */
		rt2x00_register_write(adapter, TXCSR7, 0x00000000);
	}

	/*
	 * When updating the irq_mask we must call enable_interrupt()
	 * to activate the new irq_mask. Make sure that enable_interrupt()
	 * is only called when interrupt is already up.
	 */
	if(TEST_FLAG(adapter, ADAPTER_INTERRUPT_ENABLED))
		rt2x00_enable_interrupt(adapter);
}

void rt2x00_update_antenna_diversity(struct _rt2x00_adapter* adapter)
{
	u8	antenna_value = 0;

	if(adapter->chip_adapter == RT2400)
		rt2x00_bbp_regread(adapter, 1, &antenna_value);
	else
		rt2x00_bbp_regread(adapter, 2, &antenna_value);

	if(adapter->config->antenna & ANTENNA_TX_DIV)
		antenna_value = (antenna_value & 0xfc) | 0x01;
	else if(adapter->config->antenna & ANTENNA_TX_A)
		antenna_value = (antenna_value & 0xfc) | 0x00;
	else if(adapter->config->antenna & ANTENNA_TX_B)
		antenna_value = (antenna_value & 0xfc) | 0x02;

	if(adapter->chip_adapter == RT2400)
		rt2x00_bbp_regwrite(adapter, 1, antenna_value);
	else
		rt2x00_bbp_regwrite(adapter, 2, antenna_value);

	if(adapter->chip_adapter == RT2400)
		rt2x00_bbp_regread(adapter, 4, &antenna_value);
	else
		rt2x00_bbp_regread(adapter, 14, &antenna_value);

	if(adapter->config->antenna & ANTENNA_RX_DIV)
		antenna_value = (antenna_value & 0xf9) | 0x02;
	else if(adapter->config->antenna & ANTENNA_RX_A)
		antenna_value = (antenna_value & 0xf9) | 0x00;
	else if(adapter->config->antenna & ANTENNA_RX_B)
		antenna_value = (antenna_value & 0xf9) | 0x04;

	if(adapter->chip_adapter == RT2400)
		rt2x00_bbp_regwrite(adapter, 4, antenna_value);
	else
		rt2x00_bbp_regwrite(adapter, 14, antenna_value);
}

void rt2x00_update_bssid(struct _rt2x00_adapter* adapter)
{
	u32	set_bssid;

	memset(&set_bssid, 0x00, sizeof(u32));
	set_bssid =	(u32) (adapter->config->bssid[0]) |
			(u32) (adapter->config->bssid[1] << 8) |
			(u32) (adapter->config->bssid[2] << 16) |
			(u32) (adapter->config->bssid[3] << 24);

	rt2x00_register_write(adapter, CSR5, set_bssid);

	memset(&set_bssid, 0x00, sizeof(u32));
	set_bssid =	(u32) (adapter->config->bssid[4]) |
			(u32) (adapter->config->bssid[5] << 8);

	rt2x00_register_write(adapter, CSR6, set_bssid);
}

void rt2x00_get_rf_value(struct _rt2x00_adapter *adapter, u32 *rf1, u32 *rf2, u32 *rf3, u32 *rf4)
{
	/* RF2 is chipset independent. */
	*rf2 = CHANNELRF(adapter);

	if(adapter->chip_adapter == RT2400){
		*rf1 = 0x94022058;
		*rf3 = (adapter->chip_rf_type == RT2420) ? 0x94000111 : 0x94000101;
		*rf4 = 0x00000000;
		return;
	}

	/* RT2500 chipsets */
	if(adapter->chip_rf_type == RT2522){
		*rf1 = 0x94002050;
		*rf3 = 0x94000101;
		*rf4 = 0x00000000;
	} else if(adapter->chip_rf_type == RT2523){
		*rf1 = 0x94022010;
		*rf3 = 0x940e0111;
		*rf4 = (CHANNEL(adapter) != 14) ? 0x94000a1b : 0x94000a03;
	} else if(adapter->chip_rf_type == RT2524){
		*rf1 = 0x94032020;
		*rf3 = 0x94000101;
		*rf4 = (CHANNEL(adapter) != 14) ? 0x94000a1b : 0x94000a03;
	} else if(adapter->chip_rf_type == RT2525){
		*rf1 = 0x94022020;
		*rf3 = 0x94060111;
		*rf4 = (CHANNEL(adapter) != 14) ? 0x94000a1b : 0x94000a03;
	} else if(adapter->chip_rf_type == RT2525E){
		*rf1 = 0x94022020;
		*rf3 = 0x94060111;
		*rf4 = (CHANNEL(adapter) != 14) ? 0x94000a0b : 0x94000a1b;
	} else if(adapter->chip_rf_type == RT5222){
		*rf3 = 0x94000101;

		if(CHANNEL(adapter) >= 149 && CHANNEL(adapter) <= 161){
			*rf1 = 0x94022020;
			*rf4 = 0x94000a07;
		} else if(CHANNEL(adapter) >= 36 && CHANNEL(adapter) <= 140){
			*rf1 = 0x94022010;
			*rf4 = (CHANNEL(adapter) >= 100) ? 0x94000a0f : 0x94000a1f;
		} else if(CHANNEL(adapter) <= 14){
			*rf1 = 0x94022020;
			*rf4 = (CHANNEL(adapter) != 14) ? 0x94000a0b : 0x94000a1b;
		} else
			BUG();	/* We shouldn't be here. */
	} else
		BUG();		/* We shouldn't be here. */
}

void rt2x00_update_channel(struct _rt2x00_adapter* adapter)
{
	u32		r1;
	u32		r2;
	u32		r3;
	u32		r4;

	rt2x00_get_rf_value(adapter, &r1, &r2, &r3, &r4);

	/* For RT2500 devices set TX_power. */
	if(adapter->chip_adapter == RT2500)
		r3 |= TXPOWER(adapter) << 9;

	rt2x00_rf_regwrite(adapter, r1);
	rt2x00_rf_regwrite(adapter, r2);
	rt2x00_rf_regwrite(adapter, r3);

	/*
	 * If the r4 value is 0 the device does not support the RF4 register.
	 */
	if(r4)
		rt2x00_rf_regwrite(adapter, r4);

	/*
	 * We have set the frequency to the register, now we need to lock it down.
	 */
	if(adapter->chip_adapter == RT2400){
		if(adapter->chip_rf_type == RT2420)
			return;
		/*
		 * For the RT2421 chipsets we need to write an invalid
		 * reference clock rate to activate auto_tune.
		 */
		rt2x00_rf_regwrite(adapter, r1);
		rt2x00_rf_regwrite(adapter, 0x940c2a32);
		rt2x00_rf_regwrite(adapter, r3);
	} else {
		/*
		 * Channel 14 requires the Japan filter bit to be set.
		 */
		if(adapter->config->channel == 14)
			rt2x00_bbp_regwrite(adapter, 70, 0x4e);
		else
			rt2x00_bbp_regwrite(adapter, 70, 0x46);
	}

	/*
	 * We have to wait 1 ms to let the adapter do its work.
	 */
	mdelay(1);

	/*
	 * RT2421 requires the rf registers to be set again to the correct channel.
	 */
	if(adapter->chip_adapter == RT2400 && adapter->chip_rf_type == RT2421){
		rt2x00_rf_regwrite(adapter, r1);
		rt2x00_rf_regwrite(adapter, r2);
		rt2x00_rf_regwrite(adapter, r3);
		/*
		 * And wait again to let the auto tuning do its work.
		 */
		mdelay(1);
	}

	/*
	 * Switch off tuning bits.
	 */
	r1 &= 0xfffdffff;
	r3 &= 0xfffffeff;

	/*
	 * For RT2523 devices we do not need to update the R1 register.
	 */
	if(adapter->chip_rf_type != RT2523)
		rt2x00_rf_regwrite(adapter, r1);

	rt2x00_rf_regwrite(adapter, r3);

	/* 
	 * Clear false CRC during channel switch.
	 */
	rt2x00_register_write(adapter, CNT0, r4);
}

void rt2x00_update_duration(struct _rt2x00_adapter* adapter)
{
	union _csr11	csr11;
	union _csr18	csr18;
	union _csr19	csr19;

	/*
	 * In 802.11a networks we make use of a short slot time.
	 */
	adapter->config->slot_time = (TEST_FLAG(adapter, NETWORK_802_11_A)) ? 9 : 20;

	rt2x00_register_read(adapter, CSR11, &csr11.word);
	csr11.field.cwmin = 5;	/* 2^5 = 32 */
	csr11.field.cwmax = 10;	/* 2^10 = 1024 */
	csr11.field.slot_time = adapter->config->slot_time;
	csr11.field.long_retry = adapter->config->long_retry;
	csr11.field.short_retry = adapter->config->short_retry;
	rt2x00_register_write(adapter, CSR11, csr11.word);

	rt2x00_register_read(adapter, CSR18, &csr18.word);
	csr18.field.sifs = adapter->config->sifs;
	csr18.field.pifs = csr18.field.sifs + adapter->config->slot_time;
	rt2x00_register_write(adapter, CSR18, csr18.word);

	rt2x00_register_read(adapter, CSR19, &csr19.word);
	csr19.field.difs = csr18.field.pifs + adapter->config->slot_time;
	/* EIFS is the time that an ACK packet is send in a basic rate. */
	csr19.field.eifs = csr18.field.sifs + DURATION((IEEE802_11_HLEN + ACK_SIZE), BASIC_RATE(adapter));
	rt2x00_register_write(adapter, CSR19, csr19.word);
}

void rt2x00_update_packet_filter(struct _rt2x00_adapter* adapter)
{
	union _rxcsr0	rxcsr0;

	rxcsr0.word = 0;

	rxcsr0.field.disable_rx = 0;

	if(LINK_CFG_TYPE(adapter, IW_MODE_ADHOC)
	|| LINK_CFG_TYPE(adapter, IW_MODE_INFRA)
	|| LINK_CFG_TYPE(adapter, IW_MODE_AUTO))
		rxcsr0.field.drop_tods = 1;

	if(!TEST_FLAG(adapter, ACCEPT_PROMISC))
		rxcsr0.field.drop_not_to_me = 1;

	if(!LINK_CFG_TYPE(adapter, IW_MODE_MONITOR)){
		rxcsr0.field.drop_crc = 1;
		rxcsr0.field.drop_physical = 1;
		rxcsr0.field.drop_control = 1;
		rxcsr0.field.drop_versionerr = 1;
		adapter->net_dev->type = ARPHRD_ETHER;
	} else {
		adapter->net_dev->type = ARPHRD_IEEE80211;	/* 802.11 raw mode */
		rxcsr0.field.drop_crc = 1;
	}

	if(adapter->chip_adapter == RT2500 && !TEST_FLAG(adapter, ACCEPT_MULTICAST)){
		rxcsr0.field.drop_mcast = 1;
		rxcsr0.field.drop_bcast = 1;
	}

	rxcsr0.field.pass_crc = 0;
	rxcsr0.field.pass_plcp = 0;
	rxcsr0.field.enable_qos = 0;

	rt2x00_register_write(adapter, RXCSR0, rxcsr0.word);
}

void rt2x00_update_adapter(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	rt2x00_update_config(adapter);
	rt2x00_update_txpreamble(adapter);
	rt2x00_update_txrate(adapter);
	rt2x00_update_txpower(adapter);
	rt2x00_update_power_management(adapter);
	rt2x00_update_antenna_diversity(adapter);
	rt2x00_update_bssid(adapter);
	rt2x00_update_channel(adapter);
	rt2x00_update_duration(adapter);
	rt2x00_update_packet_filter(adapter);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
}

void rt2x00_radio_on(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, ADAPTER_RADIO_ON)){
		write_unlock(&adapter->config_lock);
		spin_unlock(&adapter->adapter_lock);
		return;
	}

	/* Load pending configuration */
	rt2x00_load_config(adapter);

	/* Turn on radio */
	rt2x00_register_write(adapter, PWRCSR0, 0x3f3b3100);
	/* Abort Tx */
	rt2x00_register_write(adapter, TXCSR0, 0x00000008);
	/* Disable Rx */
	rt2x00_register_write(adapter, RXCSR0, 0x00000001);

	LED_ON(adapter);

	/*
	 * Allow the tasklets to run again.
	 */
	tasklet_enable(&adapter->tx_tasklet);
	tasklet_enable(&adapter->prio_tasklet);
	tasklet_enable(&adapter->rx_tasklet);

	rt2x00_enable_interrupt(adapter);

	SET_FLAG(adapter, ADAPTER_RADIO_ON);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	INFO("Radio on.\n");

	rt2x00_link_up(adapter);
}

void rt2x00_radio_off(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(!TEST_FLAG(adapter, ADAPTER_RADIO_ON)){
		spin_unlock(&adapter->adapter_lock);
		write_unlock(&adapter->config_lock);
		return;
	}

	rt2x00_disable_interrupt(adapter);

	/*
	 * Prevent the tasklets from running.
	 */
	tasklet_kill(&adapter->tx_tasklet);
	tasklet_kill(&adapter->prio_tasklet);
	tasklet_kill(&adapter->rx_tasklet);

	tasklet_disable(&adapter->tx_tasklet);
	tasklet_disable(&adapter->prio_tasklet);
	tasklet_disable(&adapter->rx_tasklet);

	/* Turn off radio */
	rt2x00_register_write(adapter, PWRCSR0, 0x00000000);
	/* Abort Tx */
	rt2x00_register_write(adapter, TXCSR0, 0x00000008);
	/* Disable Rx */
	rt2x00_register_write(adapter, RXCSR0, 0x00000001);

	LED_OFF(adapter);

	CLEAR_FLAG(adapter, ADAPTER_RADIO_ON);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	INFO("Radio off.\n");

	if(LINK_IS_UP(adapter))
		rt2x00_link_down(adapter);
}
