/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_connection.c

	Abstract: All routines to start and halt a connection.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/netdevice.h>
#include <linux/timer.h>

#include <net/iw_handler.h>

#include "rt2x00_adapter.h"

/*
 * Function declaration for periodic execution.
 */
void rt2x00_link_periodic(unsigned long data);

/*
 * return value for link* functions to notify the main handler that the periodic execution is required.
 */
#define ENABLE_PERIODIC_EXEC	1

/*
 * link_set() and link_unset() must be called with adapter->config_lock held;
 */
char rt2x00_link_set(struct _rt2x00_adapter* adapter, struct _scan_entry *node, u8 mode)
{
	if(unlikely(adapter->connection != NULL)){
		ERROR("Attempting to set link, while connection is already up!\n");
		return -EBUSY;
	}

	adapter->config->active_mode = mode;

	/*
	 * When we are not connecting to a specific node we are all done.
	 */
	if(!node)
		return 0;

	adapter->connection = kmalloc(sizeof(struct _scan_entry), GFP_KERNEL);
	if(!adapter->connection){
		ERROR("Not enough memory to start connection.\n");
		return -ENOMEM;
	}

	memcpy(adapter->connection, node, sizeof(struct _scan_entry));

	return 0;
}

void rt2x00_link_unset(struct _rt2x00_adapter* adapter)
{
	if(adapter->connection){
		kfree(adapter->connection);
	}
	adapter->connection = NULL;
	adapter->config->active_mode = 0;
}

void rt2x00_link_up_config(struct _rt2x00_adapter* adapter)
{
	u8	counter;
	u8	rate;

	/*
	 * When we are not connected to a node, we have nothing to do.
	 */
	if(!adapter->connection)
		return;

	memcpy(&adapter->config->essid, &adapter->connection->essid, IW_ESSID_MAX_SIZE);
	memcpy(&adapter->config->bssid, &adapter->connection->bssid, ETH_ALEN);

	/*
	 * Before we can set the bit rate we have to search the rate array
	 * to find if the bit rate we want is supported.
	 */
	rate = adapter->config->max_rate;

	/*
	 * Check if we have set a non 802.11b rate.
	 * Otherwise we will have to loop through the array to search for our rate.
	 */
	if(adapter->connection->rates[4] == 0){
		rate = (rate > 3) ? 3 : rate;
	} else {
		for(counter = 0; counter < MAX_BIT_RATES; counter++){
			if(_BITRATE(adapter, rate) == adapter->connection->rates[counter])
				break;
		}

		rate = counter;
	}

	/*
	 * Don't forget to update our current rate too.
	 */
	adapter->config->max_rate = rate;
	adapter->config->cur_rate = rate;

	adapter->config->channel = adapter->connection->channel;

	adapter->config->atim_count = adapter->connection->atim_count;
	adapter->config->cfp_count = adapter->connection->cfp_count;
	adapter->config->atim_period = adapter->connection->atim_period;
	adapter->config->cfp_period = adapter->connection->cfp_period;
}

void rt2x00_link_down_config(struct _rt2x00_adapter* adapter)
{
	/*
	 * If user had set the BSSID we must clear the ESSID, otherwise we must clear the BSSID.
	 */
	if(TEST_FLAG(adapter, FIX_BSSID))
		memset(&adapter->config->essid, 0x00, IW_ESSID_MAX_SIZE);
	else
		memset(&adapter->config->bssid, 0x00, ETH_ALEN);
}

void rt2x00_link_periodic_config(struct _rt2x00_adapter* adapter)
{
	u8	status;

	/*
	 * Periodically we check the quality of the network, and if allowed optimize it.
	 * Step 1: Check the quality of the node we are connected to. And change the rate accordingly.
	 */
	if(!TEST_FLAG(adapter, FIX_RATE)){
		if(LINK_QUALITY_BAD(adapter)){
			if(adapter->config->cur_rate != 0);
				adapter->config->cur_rate--;
		} else if(LINK_QUALITY_GOOD(adapter)){
			if(adapter->config->cur_rate != adapter->capability->size_bit_rates);
				adapter->config->cur_rate++;
		}
	}

	/*
	 * Step 2: Check for background noise, and decide if we need to increase the tx power.
	 */
	if(!TEST_FLAG(adapter, FIX_TXPOWER)){
		/* TODO: Check background noise, and change tx power accordingly. */
	}

	/*
	 * Step 3: Check for the statistics, and decide if we need to change the fragmentation or RTS.
	 * For this we calculate the percentage of packets that have been retried.
	 */
	status = ((adapter->stats.collisions + adapter->iw_stats.discard.retries) / adapter->stats.tx_packets) * 100;

	if(!TEST_FLAG(adapter, FIX_FRAG)){
		if(status > HIGH_RETRY_ERROR_FRAG){
			adapter->config->fragmentation -= 128;
			if(adapter->config->fragmentation < MIN_FRAGMENTATION)
				adapter->config->fragmentation = MIN_FRAGMENTATION;
		} else if(status < LOW_RETRY_ERROR_FRAG){
			adapter->config->fragmentation += 128;
			if(adapter->config->fragmentation > MAX_FRAGMENTATION)
				adapter->config->fragmentation = MAX_FRAGMENTATION;
		}
	}

	if(!TEST_FLAG(adapter, FIX_RTS)){
		if(status > HIGH_RETRY_ERROR_RTS){
			adapter->config->rts_treshold -= 64;
			/*
			 * MIN_RTS_THRESHOLD is 0. This means that when we go lower then this limit,
			 * The value in rts_treshold will be very high. Make the correct check. 
			 */
			if(adapter->config->rts_treshold > MAX_RTS_THRESHOLD)
				adapter->config->rts_treshold = MIN_RTS_THRESHOLD;
		} else if(status < LOW_RETRY_ERROR_RTS){
			adapter->config->rts_treshold += 64;
			if(adapter->config->rts_treshold > MAX_RTS_THRESHOLD)
				adapter->config->rts_treshold = MAX_RTS_THRESHOLD;
		}
	}
}

char rt2x00_link_up_adhoc(struct _rt2x00_adapter* adapter, struct _scan_entry* node)
{
	struct _scan_entry	*connect = node;

	/*
	 * Start searching for a node we can connect to.
	 * When we are configured in auto mode, this searched has already occured,
	 * so not point in repeating the search.
	 */
	if(!connect && !LINK_CFG_TYPE(adapter, IW_MODE_AUTO))
		connect = rt2x00_search_scan_entry(adapter, IW_MODE_ADHOC);

	/*
	 * When no node has been found, check if the last scan has expired,
	 * if that is the case we must start a scan to search for a node.
	 */
	if(!connect && SCAN_IS_EXPIRED(adapter)){
		rt2x00_start_scan(adapter, IW_SCAN_ALL_RATE);
		return -EINPROGRESS;
	}

	rt2x00_start_beacon(adapter, IW_MODE_ADHOC);

	if(rt2x00_link_set(adapter, connect, IW_MODE_ADHOC))
		return -ENOMEM;

	INFO("Link up in adhoc mode.\n");
	return ENABLE_PERIODIC_EXEC;
}

void rt2x00_link_down_adhoc(struct _rt2x00_adapter* adapter)
{
	rt2x00_halt_beacon(adapter);

	rt2x00_link_unset(adapter);

	INFO("Link down from adhoc mode.\n");
}

char rt2x00_link_periodic_adhoc(struct _rt2x00_adapter* adapter)
{
	char	link_fail = 0;

	if(!LINK_IS_UP(adapter))
		goto connect;

	/*
	 * Link is up, but we are not connected to a node.
	 * This means we are the host of this network.
	 */
	if(!adapter->connection)
		return 0;

	/*
	 * If the node we are connected to expired, we must reconnect to the network.
	 */
	if(BEACON_IS_LOST(adapter->connection)){
		rt2x00_link_down_adhoc(adapter);
		goto connect;
	}

	return 0;

connect:
	link_fail = rt2x00_link_up_adhoc(adapter, NULL);
	if(!link_fail)
		rt2x00_link_up_config(adapter);
	return link_fail;
}

char rt2x00_link_up_infra(struct _rt2x00_adapter* adapter, struct _scan_entry* node)
{
	struct _scan_entry	*connect = node;

	if(!connect){
		connect = rt2x00_search_scan_entry(adapter, IW_MODE_INFRA);
		/*
		 * If no node has been found we must rescan.
		 * It does not matter if the previous scan has expired or not.
		 */
		if(!connect){
			rt2x00_start_scan(adapter, IW_SCAN_ALL_RATE);
			return ENABLE_PERIODIC_EXEC;
		}
	}

	rt2x00_start_beacon(adapter, IW_MODE_INFRA);

	if(rt2x00_link_set(adapter, connect, IW_MODE_INFRA))
		return -ENOMEM;

	INFO("Link up in managed mode.\n");
	return ENABLE_PERIODIC_EXEC;
}

void rt2x00_link_down_infra(struct _rt2x00_adapter* adapter)
{
	rt2x00_halt_beacon(adapter);

	rt2x00_link_unset(adapter);

	INFO("Link down from managed mode.\n");
}

char rt2x00_link_periodic_infra(struct _rt2x00_adapter* adapter)
{
	char	link_fail = 0;

	if(!LINK_IS_UP(adapter))
		goto connect;

	/*
	 * Link is up, but we are not connected to a node.
	 * This means we are the host of this network.
	 */
	if(!adapter->connection)
		return 0;

	/*
	 * If the node we are connected to expired,
	 * we need to scan for a new node or continue this network on ourself.
	 */
	if(BEACON_IS_LOST(adapter->connection)){
		rt2x00_link_down_infra(adapter);
		goto connect;
	}

	return 0;
connect:
	link_fail = rt2x00_link_up_infra(adapter, NULL);
	if(!link_fail)
		rt2x00_link_up_config(adapter);
	return link_fail;
}

char rt2x00_link_up_auto(struct _rt2x00_adapter* adapter)
{
	u8			link_fail = 0;
	struct _scan_entry	*connect = NULL;

	INFO("Link up auto\n");

	if(rt2x00_link_set(adapter, connect, IW_MODE_AUTO))
		return -ENOMEM;
	
	connect = rt2x00_search_scan_entry(adapter, IW_MODE_AUTO);
    	
	/*
	 * When nothing is found, we must start a new scan.
	 * And try again when the scan has finished.
	 */
	if(!connect && SCAN_IS_EXPIRED(adapter)){
		rt2x00_start_scan(adapter, IW_SCAN_ALL_FREQ | IW_SCAN_ALL_RATE | IW_SCAN_ALL_MODE | IW_SCAN_ALL_ESSID );
		INFO("Enable periodic exec\n");
		return ENABLE_PERIODIC_EXEC;
	}

	if(!connect || connect->mode == IW_MODE_ADHOC)
		link_fail = rt2x00_link_up_adhoc(adapter, connect);
	else if(connect->mode == IW_MODE_MASTER)
		link_fail = rt2x00_link_up_infra(adapter, connect);
	else {
		WARNING("Invalid connect mode.\n");
		link_fail = -EINVAL;
	}

	return link_fail;
}

void rt2x00_link_down_auto(struct _rt2x00_adapter* adapter)
{
	if(LINK_TYPE(adapter, IW_MODE_ADHOC)){
		rt2x00_link_down_adhoc(adapter);
	} else if(LINK_TYPE(adapter, IW_MODE_INFRA)){
		rt2x00_link_down_infra(adapter);
	} else {
		WARNING("Invalid connect mode.\n");
	}
}

char rt2x00_link_periodic_auto(struct _rt2x00_adapter* adapter)
{
	struct _scan_entry	*connect = NULL;
	char			link_fail = -EINVAL;

	connect = rt2x00_search_scan_entry(adapter, IW_MODE_AUTO);

	/*
	 * Search finished, if we found a managed network, connect to that,
	 * Else we either join or start an ad-hoc network.
	 */
	if(!connect || connect->mode == IW_MODE_ADHOC){
		link_fail = rt2x00_link_up_adhoc(adapter, connect);
	} else if(connect->mode == IW_MODE_MASTER){
		link_fail = rt2x00_link_up_infra(adapter, connect);
	} else {
		WARNING("Invalid connect mode.\n");
	}

	return link_fail;
}

char rt2x00_link_up_master(struct _rt2x00_adapter* adapter)
{
	rt2x00_start_beacon(adapter, IW_MODE_MASTER);

	CLEAR_FLAG(adapter, ENABLED_POWER);

	if(rt2x00_link_set(adapter, NULL, IW_MODE_MASTER))
		return -ENOMEM;

	INFO("Link up in master mode.\n");
	return 0;
}

void rt2x00_link_down_master(struct _rt2x00_adapter* adapter)
{
	rt2x00_halt_beacon(adapter);

	rt2x00_link_unset(adapter);

	INFO("Link down from master mode.\n");
}

char rt2x00_link_up_repeat(struct _rt2x00_adapter* adapter)
{
	CLEAR_FLAG(adapter, ENABLED_POWER);

	rt2x00_link_set(adapter, NULL, IW_MODE_REPEAT);

	INFO("Link up in repeater mode.\n");
	return 0;
}

void rt2x00_link_down_repeat(struct _rt2x00_adapter* adapter)
{
	rt2x00_link_unset(adapter);

	INFO("Link down from repeater mode.\n");
}

char rt2x00_link_up_second(struct _rt2x00_adapter* adapter)
{
	struct _scan_entry	*connect = NULL;

	connect = rt2x00_search_scan_entry(adapter, IW_MODE_MASTER);

	if(!connect && SCAN_IS_EXPIRED(adapter)){
		rt2x00_start_scan(adapter, 0);
		return ENABLE_PERIODIC_EXEC;
	}

	rt2x00_start_beacon(adapter, IW_MODE_MASTER);

	CLEAR_FLAG(adapter, ENABLED_POWER);

	if(rt2x00_link_set(adapter, connect, IW_MODE_SECOND))
		return -ENOMEM;

	INFO("Link up in secondary mode.\n");
	return ENABLE_PERIODIC_EXEC;
}

void rt2x00_link_down_second(struct _rt2x00_adapter* adapter)
{
	rt2x00_halt_beacon(adapter);

	rt2x00_link_unset(adapter);

	INFO("Link down from secondary mode.\n");
}

char rt2x00_link_periodic_second(struct _rt2x00_adapter* adapter)
{
	struct _scan_entry	*connect = NULL;

	if(LINK_TYPE(adapter, IW_MODE_SECOND)){
		/*
		 * If the node we are connected to expired,
		 * we must start up as master outself.
		 */
		if(BEACON_IS_LOST(adapter->connection)){
			rt2x00_link_down_second(adapter);
			return rt2x00_link_up_master(adapter);
		}
		return 0;
	}

	if(LINK_TYPE(adapter, IW_MODE_MASTER)){
		/*
		 * We are currently acting as backup, check if the AP is back.
		 */
		connect = rt2x00_search_scan_entry(adapter, IW_MODE_MASTER);
		/*
		 * Still no AP, we remain AP.
		 */
		if(!connect)
			return 0;

		rt2x00_link_down_master(adapter);

		rt2x00_start_beacon(adapter, IW_MODE_MASTER);

		CLEAR_FLAG(adapter, ENABLED_POWER);
    
		
		if(rt2x00_link_set(adapter, connect, IW_MODE_SECOND))
			return -ENOMEM;

		INFO("Link up in secondary mode.\n");
	}

	return 0;
}

char rt2x00_link_up_monitor(struct _rt2x00_adapter* adapter)
{
	CLEAR_FLAG(adapter, ENABLED_POWER);

	/*
	 * Disable TX interrupts.
	 */
	adapter->irq_mask |= 0x39;
	if(TEST_FLAG(adapter, ADAPTER_INTERRUPT_ENABLED))
		rt2x00_enable_interrupt(adapter);

	if(rt2x00_link_set(adapter, NULL, IW_MODE_MONITOR))
		return -ENOMEM;

	INFO("Link up in Monitor mode.\n");
	return 0;
}

void rt2x00_link_down_monitor(struct _rt2x00_adapter* adapter)
{
	rt2x00_link_unset(adapter);

	/*
	 * Enable TX interrupts.
	 */
	adapter->irq_mask &= ~0x39;
	if(TEST_FLAG(adapter, ADAPTER_INTERRUPT_ENABLED))
		rt2x00_enable_interrupt(adapter);

	INFO("Link down from Monitor mode.\n");
}

void rt2x00_link_up(struct _rt2x00_adapter* adapter)
{
	char			link_fail = 0;

	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(LINK_IS_UP(adapter)){
		ERROR("link already up.\n");
		goto link_up_failure;
	}

	/*
	 * When starting a connection we no longer require any packet that remains in the varous rings.
	 */
	rt2x00_cleanup_dma_ring(adapter, RX_RING | TX_RING | ATIM_RING | PRIO_RING | BEACON_RING);

	if(LINK_CFG_TYPE(adapter, IW_MODE_AUTO)){
		link_fail = rt2x00_link_up_auto(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_ADHOC)){
		link_fail = rt2x00_link_up_adhoc(adapter, NULL);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_INFRA)){
		link_fail = rt2x00_link_up_infra(adapter, NULL);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_MASTER)){
		link_fail = rt2x00_link_up_master(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_REPEAT)){
		link_fail = rt2x00_link_up_repeat(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_SECOND)){
		link_fail = rt2x00_link_up_second(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_MONITOR)){
		link_fail = rt2x00_link_up_monitor(adapter);
	} else {
		ERROR("Invalid adapter connection mode.\n");
	}

	/*
	 * Did we fail to establish a link
	 */
	if(link_fail < 0)
		goto link_up_failure;

	/*
	 * We succeeded to establish a link, but we need to periodically try to optimize our connection.
	 */
	if(link_fail == ENABLE_PERIODIC_EXEC){
		NOTICE("Activating periodic connection tuning.\n");
		adapter->periodic.expires = jiffies + PERIODIC_EXEC;
		adapter->periodic.data = (unsigned long) adapter;
		adapter->periodic.function = &rt2x00_link_periodic;
		add_timer(&adapter->periodic);
	}

	/*
	 * Configure adapter with current connection.
	 */
	rt2x00_link_up_config(adapter);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	/*
	 * The configuration data has most likely been changed.
	 * To activate these changes we need to update the adapter.
	 */
	rt2x00_update_adapter(adapter);

	netif_wake_queue(adapter->net_dev);
	return;

link_up_failure:
	WARNING("Failure to establish link.\n");
	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
}

void rt2x00_link_down(struct _rt2x00_adapter* adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(LINK_CFG_TYPE(adapter, IW_MODE_AUTO)){
		rt2x00_link_down_auto(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_ADHOC)){
		rt2x00_link_down_adhoc(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_INFRA)){
		rt2x00_link_down_infra(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_MASTER)){
		rt2x00_link_down_master(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_REPEAT)){
		rt2x00_link_down_repeat(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_SECOND)){
		rt2x00_link_down_second(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_MONITOR)){
		rt2x00_link_down_monitor(adapter);
	} else {
		ERROR("Link already down.\n");
	}

	del_timer_sync(&adapter->periodic);
	rt2x00_link_down_config(adapter);

	/*
	 * Since the link is down we no longer require the packets in the queue.
	 */
	rt2x00_cleanup_dma_ring(adapter, RX_RING | TX_RING | ATIM_RING | PRIO_RING | BEACON_RING);

	netif_stop_queue(adapter->net_dev);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
}

void rt2x00_link_periodic(unsigned long data)
{
	struct _rt2x00_adapter	*adapter = (struct _rt2x00_adapter*)data;
	char			link_fail = -EAGAIN;

	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	/*
	 * No periodic work when we are scanning.
	 */
	if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS))
		goto periodic;

	/*
	 * First do all Link specific periodic work.
	 */
	if(LINK_CFG_TYPE(adapter, IW_MODE_AUTO)){
		link_fail = rt2x00_link_periodic_auto(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_ADHOC)){
		link_fail = rt2x00_link_periodic_adhoc(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_INFRA)){
		link_fail = rt2x00_link_periodic_infra(adapter);
	} else if(LINK_CFG_TYPE(adapter, IW_MODE_SECOND)){
		link_fail = rt2x00_link_periodic_second(adapter);
	} else {
		ERROR("Invalid periodic connection mode.\n");
	}

	if(link_fail)
		goto periodic;

	/*
	 * Tune configuration to optimize the connection quality/strength.
	 */
	rt2x00_link_periodic_config(adapter);

periodic:

	mod_timer(&adapter->periodic, jiffies + PERIODIC_EXEC);
	
	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
	
	if(!link_fail)
		rt2x00_update_adapter(adapter);
}
