/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_eeprom.c

	Abstract: eeprom routines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/delay.h>

#include <net/iw_handler.h>

#include    "rt2x00_adapter.h"

void rt2x00_eeprom_pulse_high(struct _rt2x00_adapter *adapter, u32 *x)
{
	/*
	 * Set data clock.
	 */
	*x = *x | EEPROM_DATA_CLOCK;
	rt2x00_register_write(adapter, CSR21, *x);
	udelay(1);
}

void rt2x00_eeprom_pulse_low(struct _rt2x00_adapter *adapter, u32 *x)
{
	/*
	 * Clear data clock.
	 */
	*x = *x & ~EEPROM_DATA_CLOCK;
	rt2x00_register_write(adapter, CSR21, *x);
	udelay(1);
}

void rt2x00_eeprom_shift_out_bits(struct _rt2x00_adapter *adapter, u16 data, u16 count)
{
	u32	x;
	u32	mask;

	/*
	 * Set start position.
	 */
	mask = 0x01 << (count - 1);

	rt2x00_register_read(adapter, CSR21, &x);

	/*
	 * Clear data flags.
	 */
	x &= ~(EEPROM_DATA_OUT | EEPROM_DATA_IN);

	/*
	 * Start writing all bits. 
	 */
	do{
		/*
		 * Only set the data_in flag when weare at the correct bit.
		 */
		x &= ~EEPROM_DATA_IN;
		if(data & mask)
			x |= EEPROM_DATA_IN;

		rt2x00_register_write(adapter, CSR21, x);

		rt2x00_eeprom_pulse_high(adapter, &x);
		rt2x00_eeprom_pulse_low(adapter, &x);
	
		mask = mask >> 1;
	} while(mask);

	x &= ~EEPROM_DATA_IN;
	rt2x00_register_write(adapter, CSR21, x);
}

u16 rt2x00_eeprom_shift_in_bits(struct _rt2x00_adapter *adapter)
{
	u32	x;
	u16	data = 0;
	u8	counter;

	rt2x00_register_read(adapter, CSR21, &x);

	/*
	 * Clear data flags.
	 */
	x &= ~( EEPROM_DATA_OUT | EEPROM_DATA_IN);

	/*
	 * Start reading all 16 bits.
	 */
	for(counter = 0; counter < 16; counter++){
		/*
		 * Shift to the next bit.
		 */
		data = data << 1;

		rt2x00_eeprom_pulse_high(adapter, &x);

		rt2x00_register_read(adapter, CSR21, &x);

		/*
		 * Clear data_in flag and set the data bit to 1 when the data_out flag is set.
		 */
		x &= ~(EEPROM_DATA_IN);
		if(x & EEPROM_DATA_OUT)
			data |= 1;
	
		rt2x00_eeprom_pulse_low(adapter, &x);
	}

	return data;
}


void rt2x00_eeprom_cleanup(struct _rt2x00_adapter *adapter)
{
	u32	x;

	/*
	 * Clear chip_select and data_in flags.
	 */
	rt2x00_register_read(adapter, CSR21, &x);
	x &= ~(EEPROM_CHIP_SELECT | EEPROM_DATA_IN);
	rt2x00_register_write(adapter, CSR21, x);

	/*
	 * kick a pulse.
	 */
	rt2x00_eeprom_pulse_high(adapter, &x);
	rt2x00_eeprom_pulse_low(adapter, &x);
}

u16 rt2x00_eeprom_read16(struct _rt2x00_adapter *adapter, u8 offset, u8 bus_width)
{
	u32	x;
	u16	data;

	offset /= 2;

	/*
	 * Clear all flags, and enable chip select.
	 */
	rt2x00_register_read(adapter, CSR21, &x);
	x &= ~(EEPROM_DATA_IN | EEPROM_DATA_OUT | EEPROM_DATA_CLOCK);
	x |= EEPROM_CHIP_SELECT;
	rt2x00_register_write(adapter, CSR21, x);

	/*
	 * kick a pulse.
	 */
	rt2x00_eeprom_pulse_high(adapter, &x);
	rt2x00_eeprom_pulse_low(adapter, &x);

	/*
	 * Select the read opcode and bus_width.
	 */
	rt2x00_eeprom_shift_out_bits(adapter, EEPROM_READ_OPCODE, 3);
	rt2x00_eeprom_shift_out_bits(adapter, offset, bus_width);

	data = rt2x00_eeprom_shift_in_bits(adapter);

	rt2x00_eeprom_cleanup(adapter);

	return data;
}
