/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_ioctl.c

	Abstract: IOCTL related subroutines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_arp.h>
#include <linux/delay.h>

#include <asm/uaccess.h>

#include <net/iw_handler.h>

#include "rt2x00_mgmt.h"
#include "rt2x00_adapter.h"

int rt2x00_commit_init(struct _rt2x00_adapter *adapter)
{
	struct _config_commit *commits;

	commits = kmalloc(sizeof(struct _config_commit), GFP_KERNEL);
	if(!commits){
		ERROR("Not enough memory to allocate commit structure.\n");
		return -ENOMEM;
	}

	memset(commits, 0x00, sizeof(struct _config_commit));

	write_lock(&adapter->config_lock);
	adapter->config->commits = commits;
	write_unlock(&adapter->config_lock);

	return 0;
}

int rt2x00_commit(struct _rt2x00_adapter *adapter, void *data, u32 type)
{
	if(!adapter->config->commits && rt2x00_commit_init(adapter))
		return -ENOMEM;

	write_lock(&adapter->config_lock);

	switch(type){
		case COMMIT_ESSID:
			memcpy(&adapter->config->commits->essid, data, IW_ESSID_MAX_SIZE);
		break;
		case COMMIT_NETWORK:
			memcpy(&adapter->config->commits->network, data, sizeof(u8));
		break;
		case COMMIT_MODE:
			memcpy(&adapter->config->commits->mode, data, sizeof(u8));
		break;
		case COMMIT_CRYPTO:
			memcpy(&adapter->config->commits->crypto, data, sizeof(struct _crypto));
		break;
		case COMMIT_BSSID:
			memcpy(&adapter->config->commits->bssid, data, ETH_ALEN);
		break;
		case COMMIT_POWER:
			memcpy(&adapter->config->commits->power, data, sizeof(struct _power));
		break;
		case COMMIT_CHANNEL:
			memcpy(&adapter->config->commits->channel, data, sizeof(u8));
		break;
	}

	SET_FLAG(adapter, type);

	write_unlock(&adapter->config_lock);

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_commit(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u8			status;

	read_lock(&adapter->config_lock);

	/*
	 * We cannot reset the connection while busy scanning.
	 * so kill the timers first and clear the flags.
	 */
	 
	if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
		del_timer_sync(&adapter->scan_timer);
		CLEAR_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS);
		del_timer_sync(&adapter->periodic);
	}

	/*
	 * When no data is pending we only need to update the adapter.
	 */
	if(!TEST_FLAG(adapter, COMMIT_PENDING)){
		rt2x00_update_adapter(adapter);
		read_unlock(&adapter->config_lock);
		return 0;
	}

	/*
	 * Check if the adapter is active.
	 */
	status = TEST_FLAG(adapter, ADAPTER_ENABLED);

	read_unlock(&adapter->config_lock);

	/*
	 * There is data pending that has to be activated.
	 * There are 2 possibilities to activate this configuration.
	 *   If the interface is up, we reset the connection by switching the radio off and on again.
	 *   If the interface is not up, just calling rt2x00_load_config() is sufficient.
	 */
	if(status){
		INFO("SIOCSIWCOMMIT: Connection will be reset.\n");
		rt2x00_radio_off(adapter);
		rt2x00_radio_on(adapter);
	} else {
		INFO("SIOCSIWCOMMIT: Pending configuration will be loaded.\n");
		write_lock(&adapter->config_lock);
		rt2x00_load_config(adapter);
		write_unlock(&adapter->config_lock);
	}

	return 0;
}

static int rt2x00_ioctl_get_name(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, NETWORK_802_11_A) && TEST_FLAG(adapter, NETWORK_802_11_G))
		strncpy(wrqu->name, "IEEE 802.11a/b/g", sizeof(wrqu->name));
	else if(TEST_FLAG(adapter, NETWORK_802_11_G))
		strncpy(wrqu->name, "IEEE 802.11b/g", sizeof(wrqu->name));
	else if(TEST_FLAG(adapter, NETWORK_802_11_B))
		strncpy(wrqu->name, "IEEE 802.11b", sizeof(wrqu->name));
	else if(TEST_FLAG(adapter, NETWORK_802_11_A))
		strncpy(wrqu->name, "IEEE 802.11a", sizeof(wrqu->name));
	else
		strncpy(wrqu->name, "Unknown", sizeof(wrqu->name));

	read_unlock(&adapter->config_lock);

	return 0;  
}

static int rt2x00_ioctl_set_frequency(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter		*adapter = netdev_priv(net_dev);
	u8				channel;
	u8				counter;

	read_lock(&adapter->config_lock);

	if ((wrqu->freq.m > 0) && (wrqu->freq.m <= 1000)) {
		/* Set channel by channel number. */
		channel = wrqu->freq.m;
		for(counter = 0; counter < adapter->capability->size_channels; counter++){
			if(channel == _CHANNEL(adapter, counter))
				break;
		}
	} else {
		/* Set channel by frequency. */
		channel = wrqu->freq.m / 100000;
		for(counter = 0; counter < adapter->capability->size_channels; counter++){
			if(channel == _CHANNELFREQ(adapter, counter))
				break;
		}
	}

	channel = counter;

	if(channel > adapter->capability->size_channels){
		read_unlock(&adapter->config_lock);
		return -EINVAL;
	}

	INFO("SIOCSIWFREQ: Channel = %d.\n", _CHANNEL(adapter, channel));

	read_unlock(&adapter->config_lock);

	return rt2x00_commit(adapter, &channel, COMMIT_CHANNEL);
}

static int rt2x00_ioctl_get_frequency(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->freq.m = CHANNEL(adapter);
	wrqu->freq.e = 0;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_operation_mode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	return rt2x00_commit(adapter, &wrqu->mode, COMMIT_MODE);
}

static int rt2x00_ioctl_get_operation_mode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(LINK_IS_UP(adapter))
		wrqu->mode = (u32)adapter->config->active_mode;
	else
		wrqu->mode = (u32)adapter->config->config_mode;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_sens(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	write_lock(&adapter->config_lock);

	if(wrqu->sens.disabled){
		adapter->config->sensitivity = SENSITIVITY_DEFAULT;
		CLEAR_FLAG(adapter, ENABLED_SENS);
	} else {
		adapter->config->sensitivity = wrqu->sens.value;
		SET_FLAG(adapter, ENABLED_SENS);
	}

	write_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_sens(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->sens.value = adapter->config->sensitivity;
	wrqu->sens.disabled = !GET_FLAG(adapter, ENABLED_SENS);
	wrqu->sens.fixed = 1;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_range(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter		*adapter = netdev_priv(net_dev);
	struct _rt2x00_capability	*capability = adapter->capability;
	struct iw_range			*range = (struct iw_range *) extra;
	int				counter;

	memset(range, 0x00, sizeof(struct iw_range));

	range->throughput = _BITRATE(adapter, capability->size_bit_rates - 1) * 500000;

	range->min_nwid = 0;
	range->max_nwid = 0;

	range->sensitivity = SENSITIVITY_DEFAULT;

	range->max_qual.qual = 70;
	range->max_qual.level = 0;
	range->max_qual.noise = 0;

	range->avg_qual.qual = 20;
	range->avg_qual.level = -60;
	range->avg_qual.noise = -95;

	range->num_bitrates = capability->size_bit_rates;
	for(counter = 0; counter < range->num_bitrates; counter++)
		range->bitrate[counter] = _BITRATE(adapter, counter) * 500000;

	range->min_rts = MIN_RTS_THRESHOLD;
	range->max_rts = MAX_RTS_THRESHOLD;

	range->min_frag = MIN_FRAGMENTATION;
	range->max_frag = MAX_FRAGMENTATION;

	range->min_pmp = MIN_POWER_PERIOD * 1024;
	range->max_pmp = MAX_POWER_PERIOD * 1024;
	range->min_pmt = MIN_POWER_TIMEOUT * 1024;
	range->max_pmt = MAX_POWER_TIMEOUT * 1024;
	range->pmp_flags = IW_POWER_PERIOD;
	range->pmt_flags = IW_POWER_TIMEOUT;
	range->pm_capa = IW_POWER_PERIOD | IW_POWER_TIMEOUT | IW_POWER_ALL_R | IW_POWER_UNICAST_R | IW_POWER_MULTICAST_R;

	range->encoding_size[0] = CRYPTO_SIZE_WEP_SMALL;
	range->encoding_size[1] = CRYPTO_SIZE_WEP_LARGE;
	range->encoding_size[2] = CRYPTO_SIZE_WPA;
	range->num_encoding_sizes = 3;
	range->max_encoding_tokens = 4;
	range->encoding_login_index = 0;

	range->txpower_capa = IW_TXPOW_MWATT;
	range->num_txpower = capability->size_tx_powers;
	for(counter = 0; counter < range->num_txpower; counter++)
		range->txpower[counter] = _TXPOWER(adapter, counter);

	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = 16;

	range->retry_capa = IW_RETRY_LIMIT;
	range->retry_flags = IW_RETRY_LIMIT;
	range->min_retry = 0;
	range->max_retry = 255;

	/*
	 * Only write the first 14 frequencies to the range structure.
	 * When we are 802.11a capable we will have over 32 frequencies available,
	 * and we are not allowed to write them all to the range structure.
	 */
	range->num_channels = 14;
	range->num_frequency = 14;

	for(counter = 0; counter < 14; counter++) {
		range->freq[counter].i = counter + 1;
		range->freq[counter].m = _CHANNELFREQ(adapter, counter) * 100000;
		range->freq[counter].e = 1;
	}

	wrqu->data.length = sizeof(struct iw_range);

	return 0;
}

static int rt2x00_ioctl_set_wap(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter		*adapter = netdev_priv(net_dev);
	/*
	 * If the wap address in invalid we are going to commit it as an 'any' essid.
	 * This means that the wap[] array must have the size of an essid instead of a mac address.
	 */
	char				wap[IW_ESSID_MAX_SIZE];

	if(wrqu->ap_addr.sa_family != ARPHRD_ETHER)
		return -EINVAL;

	memset(wap, 0x00, IW_ESSID_MAX_SIZE);
	memcpy(wap, wrqu->ap_addr.sa_data, ETH_ALEN);

	/* 
	 * When the address is not valid, because it is, for example, a broadcast address,
	 * we must set the essid to 'any' instead.
	 */

	if(!is_valid_ether_addr(wap)){
		INFO("SIOCSIWAP: Any Essid.\n");

		memset(wap, 0x00, IW_ESSID_MAX_SIZE);
	
		return rt2x00_commit(adapter, wap, COMMIT_ESSID);
	} else {
		INFO("SIOCSIWAP: Bssid = " PRINT_ADDR_STRING ".\n", PRINT_ADDR(wap));

		return rt2x00_commit(adapter, wap, COMMIT_BSSID);
	}
}

static int rt2x00_ioctl_get_wap(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(!LINK_IS_UP(adapter)){
		read_unlock(&adapter->config_lock);
		return -ENOTCONN;
	}

	wrqu->ap_addr.sa_family = ARPHRD_ETHER;

	/*
	 * Only adhoc and managed mode are connected to a access point, 
	 * so only return a valid value for these connection types.
	 */
	if(LINK_TYPE(adapter, IW_MODE_ADHOC) || LINK_TYPE(adapter, IW_MODE_INFRA))
		memcpy(wrqu->ap_addr.sa_data, adapter->config->bssid, ETH_ALEN);
	else
		memset(wrqu->ap_addr.sa_data, 0x00, ETH_ALEN);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_scan(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);
		
	/*
	 * Only start a new scan, when we are not already busy with one.
	 */
	if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
		read_unlock(&adapter->config_lock);
		return 0;
	}

	read_unlock(&adapter->config_lock);

	if(wrqu->param.flags)
	    rt2x00_start_scan(adapter,wrqu->param.flags);    
	else
	    rt2x00_start_scan(adapter, IW_SCAN_ALL_FREQ | IW_SCAN_ALL_RATE | IW_SCAN_ALL_MODE | IW_SCAN_ALL_ESSID);

	return 0;
}

static int rt2x00_ioctl_get_scan(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter		*adapter = netdev_priv(net_dev);
	struct _rt2x00_scan		*scan = &adapter->scan; 
	struct iw_event			scan_result;
	char				*curr_buff = extra;
	char				*end_buff = extra + IW_SCAN_MAX_DATA;
	char				*rate_buff;
	struct _scan_entry		*entry;
	u8 				counter;

	read_lock(&adapter->config_lock);

	/* We should block */

	counter=0;

	while(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
	    mdelay(SCAN_TIME);
	    if(counter>(adapter->capability->size_channels+2)){
		ERROR("Timeout waiting for adapter scan results\n");
		break;
	    }
	    counter++;
	}
	
	/*
	 * There aren't any scan results.
	 */
	if(list_empty(&scan->result)){
		read_unlock(&adapter->config_lock);
		return -ENODATA;
	}

	list_for_each_entry(entry, &scan->result, list){
		/*
		 * make sure we aren't exceeding the limit.
		 */
		if(curr_buff >= end_buff)
			break;

		scan_result.cmd = SIOCGIWAP;
		scan_result.u.ap_addr.sa_family = ARPHRD_ETHER;
		memcpy(scan_result.u.ap_addr.sa_data, entry->bssid, ETH_ALEN);
		curr_buff = iwe_stream_add_event(curr_buff, end_buff, &scan_result, IW_EV_ADDR_LEN);

		scan_result.cmd = SIOCGIWESSID;
		scan_result.u.data.flags = 1;
		scan_result.u.data.length = strnlen(entry->essid, IW_ESSID_MAX_SIZE);
		curr_buff = iwe_stream_add_point(curr_buff, end_buff, &scan_result, entry->essid);

		scan_result.cmd = SIOCGIWMODE;
		scan_result.u.mode = entry->mode;
		curr_buff = iwe_stream_add_event(curr_buff, end_buff, &scan_result, IW_EV_UINT_LEN);

		scan_result.cmd = SIOCGIWFREQ;
		scan_result.u.freq.m = entry->channel;
		scan_result.u.freq.e = 0;
		scan_result.u.freq.i = 0;
		curr_buff = iwe_stream_add_event(curr_buff, end_buff, &scan_result, IW_EV_FREQ_LEN);

		scan_result.cmd = IWEVQUAL;
		scan_result.u.qual.level = entry->rssi;
		scan_result.u.qual.noise = 0;
		scan_result.u.qual.qual = 0;
		scan_result.u.qual.updated = 1;
		curr_buff = iwe_stream_add_event(curr_buff, end_buff, &scan_result, IW_EV_QUAL_LEN);

		scan_result.cmd = SIOCGIWENCODE;
		if (entry->capability & CAPABLE_PRIVACY)
			scan_result.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
		else
			scan_result.u.data.flags = IW_ENCODE_DISABLED;
		scan_result.u.data.length = 0;
		curr_buff = iwe_stream_add_point(curr_buff, end_buff, &scan_result, entry->essid);

    		scan_result.cmd = SIOCGIWRATE;
		rate_buff = curr_buff + IW_EV_LCP_LEN;
		scan_result.u.bitrate.fixed = 0;
		scan_result.u.bitrate.disabled = 0;
		for(counter = 0 ; counter < 8 ; counter++) {
			scan_result.u.bitrate.value = 500000*(entry->rates[counter]&0x7f);
			if(entry->rates[counter]==0)
			    break;
			rate_buff = iwe_stream_add_value(curr_buff, rate_buff, end_buff, &scan_result, IW_EV_PARAM_LEN);
		}
	


		/* Check if we added any event */
		if((rate_buff - curr_buff) > IW_EV_LCP_LEN)
			curr_buff = rate_buff;
	}

	wrqu->data.length = curr_buff - extra;
	CLEAR_FLAGS(&wrqu->data);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_essid(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	char			*essid = "";

	if (wrqu->essid.flags) {
		if (wrqu->essid.length > IW_ESSID_MAX_SIZE)
			return -E2BIG;

		essid = extra;
	}

	INFO("SIOCSIWESSID: Essid = %s length = %d.\n", essid, strnlen(essid, IW_ESSID_MAX_SIZE));

	return rt2x00_commit(adapter, essid, COMMIT_ESSID);
}

static int rt2x00_ioctl_get_essid(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->essid.length = strnlen(adapter->config->essid, IW_ESSID_MAX_SIZE);
	/*
	 * If the 'any' essid was set, the length and flags should both be 0.
	 */
	wrqu->essid.flags = (wrqu->essid.length) ? 1 : 0;

	memcpy(extra, adapter->config->essid, wrqu->essid.length);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_nickname(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	if (wrqu->data.length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	write_lock(&adapter->config_lock);

	memset(adapter->config->nickn, 0x00, IW_ESSID_MAX_SIZE);

	memcpy(adapter->config->nickn, extra, wrqu->data.length);

	write_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_nickname(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->data.length = strnlen(adapter->config->nickn, IW_ESSID_MAX_SIZE);
	memcpy(extra, adapter->config->nickn, wrqu->data.length);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_bit_rate(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter		*adapter = netdev_priv(net_dev);
	u8				rate = 0;
	u8				counter;

	write_lock(&adapter->config_lock);

	if(wrqu->bitrate.value == -1){
		adapter->config->max_rate = adapter->capability->size_bit_rates - 1;
		adapter->config->cur_rate = adapter->config->max_rate;
		goto exit;
	}

	rate = wrqu->bitrate.value / 500000;

	for(counter = 0; counter < adapter->capability->size_bit_rates; counter++){
		if(rate == _BITRATE(adapter, counter)){
			adapter->config->max_rate = counter;
			adapter->config->cur_rate = adapter->config->max_rate;
			goto exit;
		}
	}

	write_unlock(&adapter->config_lock);
	return -EINVAL;

exit:
	if(wrqu->bitrate.fixed)
		SET_FLAG(adapter, FIX_RATE);
	else
		CLEAR_FLAG(adapter, FIX_RATE);

	write_unlock(&adapter->config_lock);

	INFO("SIOCSIWRATE: Bitrate = %d.\n", BITRATE(adapter) * 500000);

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_get_bit_rate(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->bitrate.value = BITRATE(adapter) * 500000;
	wrqu->bitrate.fixed = TEST_FLAG(adapter, FIX_RATE);
	wrqu->bitrate.disabled = 0;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_rts_treshold(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u16			rts_treshold = wrqu->rts.value;

	write_lock(&adapter->config_lock);

	if(wrqu->rts.disabled)
		rts_treshold = MAX_RTS_THRESHOLD;
	if(rts_treshold > MAX_RTS_THRESHOLD){
		write_unlock(&adapter->config_lock);
		return -EINVAL;
	}

	adapter->config->rts_treshold = rts_treshold;

	if(wrqu->rts.fixed)
		SET_FLAG(adapter, FIX_RTS);
	else
		CLEAR_FLAG(adapter, FIX_RTS);

	write_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_rts_treshold(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->rts.value = adapter->config->rts_treshold;
	wrqu->rts.disabled = (wrqu->rts.value >= MAX_RTS_THRESHOLD);
	wrqu->rts.fixed = TEST_FLAG(adapter, FIX_RTS);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_fragmentation(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u16			fragmentation = wrqu->frag.value;

	write_lock(&adapter->config_lock);

	if(wrqu->frag.disabled)
		fragmentation = MAX_FRAGMENTATION;

	if((fragmentation < MIN_FRAGMENTATION) || (fragmentation > MAX_FRAGMENTATION)){
		write_unlock(&adapter->config_lock);
		return -EINVAL;
	}

	adapter->config->fragmentation = fragmentation;

	if(wrqu->frag.fixed)
		SET_FLAG(adapter, FIX_FRAG);
	else
		CLEAR_FLAG(adapter, FIX_FRAG);

	write_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_fragmentation(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->frag.value = adapter->config->fragmentation;
	wrqu->frag.disabled = (wrqu->frag.value >= MAX_FRAGMENTATION);
	wrqu->frag.fixed = TEST_FLAG(adapter, FIX_FRAG);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_txpower(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u16			tx_power = wrqu->txpower.value;
	int 			counter;

	/*
	 * First check if the radio should be disabled or enabled.
	 */
	if(wrqu->txpower.disabled || tx_power == 0){
		rt2x00_radio_off(adapter);
		INFO("SIOCSIWTXPOW: Radio off.\n");
	} else if(!TEST_FLAG(adapter, ADAPTER_RADIO_ON)){
		rt2x00_radio_on(adapter);
		INFO("SIOCSIWTXPOW: Radio on.\n");
	} 

	write_lock(&adapter->config_lock);

	/*
	 * We don not support DBM values.
	 */
	if(TEST_FLAG(&wrqu->txpower, IW_TXPOW_DBM)){
		write_unlock(&adapter->config_lock);
		return -ENOTSUPP;
	}

	adapter->config->tx_power = 0;

	for(counter = 0; counter < adapter->capability->size_tx_powers; counter++){
		if(_TXPOWER(adapter, counter) >= tx_power)
			break;
		adapter->config->tx_power = counter;
	}

	if(wrqu->txpower.fixed)
		SET_FLAG(adapter, FIX_TXPOWER);
	else
		CLEAR_FLAG(adapter, FIX_TXPOWER);

	write_unlock(&adapter->config_lock);

	INFO("SIOCSIWTXPOW: Tx power = %d.\n", _TXPOWER(adapter, counter));

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_get_txpower(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->txpower.value = TXPOWER(adapter);
	wrqu->txpower.disabled = !TEST_FLAG(adapter, ADAPTER_RADIO_ON);
	wrqu->txpower.fixed = TEST_FLAG(adapter, FIX_TXPOWER);
	SET_FLAG(&wrqu->txpower, IW_TXPOW_MWATT);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_retry(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	/*
	 * We do not support retry lifetime, or any other type then retry limit.
	 */
	if(TEST_FLAG(&wrqu->retry, IW_RETRY_LIFETIME) || !TEST_FLAG(&wrqu->retry, IW_RETRY_LIMIT))
		return -ENOTSUPP;

	write_lock(&adapter->config_lock);

	if(TEST_FLAG(&wrqu->retry, IW_RETRY_MAX))
		adapter->config->long_retry = wrqu->retry.value;
	else if(TEST_FLAG(&wrqu->retry, IW_RETRY_MIN))
		adapter->config->short_retry = wrqu->retry.value;
	else{
		adapter->config->short_retry = wrqu->retry.value;
		adapter->config->long_retry = wrqu->retry.value;
	}

	write_unlock(&adapter->config_lock);

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_get_retry(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	if(TEST_FLAG(&wrqu->retry, IW_RETRY_LIFETIME) || !TEST_FLAG(&wrqu->retry, IW_RETRY_LIMIT))
		return -ENOTSUPP;

	read_lock(&adapter->config_lock);

	if(TEST_FLAG(&wrqu->retry, IW_RETRY_MAX))
		wrqu->retry.value = adapter->config->long_retry;
	else if(TEST_FLAG(&wrqu->retry, IW_RETRY_MIN))
		wrqu->retry.value = adapter->config->short_retry;
	else {
		wrqu->retry.value = adapter->config->long_retry;
		if(adapter->config->short_retry != adapter->config->long_retry)
			SET_FLAG(&wrqu->retry, IW_RETRY_MAX);
	}

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_encode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _crypto		crypto;

	memset(&crypto, 0x00, sizeof(struct _crypto));

	if(wrqu->encoding.length != CRYPTO_SIZE_WEP_SMALL || wrqu->encoding.length != CRYPTO_SIZE_WEP_LARGE)
		return -EINVAL;

	crypto.flags = wrqu->encoding.flags;
	crypto.type = CRYPTO_TYPE_WEP;
	crypto.length = wrqu->encoding.length;

	memcpy(crypto.key, extra, crypto.length);

	return rt2x00_commit(adapter, &crypto, COMMIT_CRYPTO);
}

static int rt2x00_ioctl_get_encode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _crypto		crypto;

	if(!rt2x00_get_key(adapter, &crypto))
		return -ENODATA;

	if(!crypto.key || !(crypto.type & CRYPTO_TYPE_WEP))
		return -ENODATA;

	read_lock(&adapter->config_lock);

	wrqu->encoding.flags = crypto.flags;
	wrqu->encoding.length = crypto.length;
	memcpy(extra, crypto.key, crypto.length);

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_power(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _power		power;

	memset(&power, 0x00, sizeof(struct _power));

	if(wrqu->power.disabled){
		power.flags = 0; /* Disable power management. */
		goto exit;
	} else {
		SET_FLAG(&power, GET_FLAG(&wrqu->power, IW_POWER_MODE));

		if(TEST_FLAG(&power, IW_POWER_FORCE_S) || TEST_FLAG(&power, IW_POWER_REPEATER))
			return -EINVAL;

		SET_FLAG(&power, ENABLED_POWER);
	}

	if(GET_FLAG(&wrqu->power, IW_POWER_TYPE) == IW_POWER_PERIOD){
		power.period = wrqu->power.value / 1024;

		power.period = (power.period < MIN_POWER_PERIOD) ? MIN_POWER_PERIOD : power.period;

		if(GET_FLAG(&wrqu->power, IW_POWER_MODIFIER) == IW_POWER_MIN)
			SET_FLAG(&power, MIN_POWER_PERIOD);
		else
			SET_FLAG(&power, MAX_POWER_PERIOD);
	} else if(GET_FLAG(&wrqu->power, IW_POWER_TYPE) == IW_POWER_TIMEOUT){
		power.timeout = wrqu->power.value / 1024;

		power.period = (power.period > MAX_POWER_TIMEOUT) ? MAX_POWER_TIMEOUT : power.period;
		power.period = (power.period < MIN_POWER_TIMEOUT) ? MIN_POWER_TIMEOUT : power.period;

		if(GET_FLAG(&wrqu->power, IW_POWER_MODIFIER) == IW_POWER_MIN)
			SET_FLAG(&power, MIN_POWER_TIMEOUT);
		else
			SET_FLAG(&power, MAX_POWER_TIMEOUT);
	}

exit:
	return rt2x00_commit(adapter, &power, COMMIT_POWER);
}

static int rt2x00_ioctl_get_power(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	wrqu->power.disabled = !TEST_FLAG(adapter, ENABLED_POWER);
	wrqu->power.flags = GET_FLAG(adapter, IW_POWER_MODE);
	if(TEST_FLAG(adapter, ACCEPT_UNICAST))
		SET_FLAG(&wrqu->power, IW_POWER_UNICAST_R);
	if(TEST_FLAG(adapter, ACCEPT_MULTICAST))
		SET_FLAG(&wrqu->power, IW_POWER_MULTICAST_R);

	if(GET_FLAG(&wrqu->power, IW_POWER_TYPE) == IW_POWER_PERIOD){
		SET_FLAG(&wrqu->power, IW_POWER_PERIOD);
		wrqu->power.value = adapter->config->cfp_period * 1024;
	} else if(GET_FLAG(&wrqu->power, IW_POWER_TYPE) == IW_POWER_TIMEOUT){
		SET_FLAG(&wrqu->power, IW_POWER_TIMEOUT);
		wrqu->power.value = adapter->config->atim_period * 1024;
	} 

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_network(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u8			network = *wrqu;

	/*
	 * network should be set according to the following rules:
	 *
	 * NETWORK_A and NETWORK_G may only be set when we have the supporting chips.
	 * It should be at least NETWORK_A, NETWORK_B or NETWORK_G
	 * In all other cases we can correct any mistakes automaticly.
	 * NETWORK_A + NETWORK_B => NETWORK_G will be set.
	 * NETWORK_A + NETWORK_G => NETWORK_B will be set.
	 * NETWORK_G => NETWORK_B will be set.
	 */

	read_lock(&adapter->config_lock);

	/*
	 * RT2400 only supports NETWORK_B, so don't bother validating it.
	 */
	if(adapter->chip_adapter == RT2400)
		network = NETWORK_B;

	/*
	 * Only RT5222 support NETWORK_A, so it can be removed with any other chip.
	 */
	if(adapter->chip_rf_type != RT5222)
		network &= ~NETWORK_A;

	if((network & NETWORK_A) && adapter->chip_rf_type != RT5222){
		NOTICE("Adapter does not have 802.11a capabilities.\n");
		read_unlock(&adapter->config_lock);
		return -EINVAL;
	}

	read_unlock(&adapter->config_lock);

	if(!(network & NETWORK_A) || !(network & NETWORK_B) || !(network & NETWORK_G))
		return -EINVAL;

	if(network & NETWORK_A && network & NETWORK_B && !(network & NETWORK_G))
		network |= NETWORK_G;

	if(network & NETWORK_G && !(network & NETWORK_B))
		network |= NETWORK_B;

	return rt2x00_commit(adapter, &network, COMMIT_POWER);
}

static int rt2x00_ioctl_get_network(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, NETWORK_802_11_ABG))
		*wrqu = NETWORK_ABG;
	else if(TEST_FLAG(adapter, NETWORK_802_11_BG))
		*wrqu = NETWORK_BG;
	else if(TEST_FLAG(adapter, NETWORK_802_11_B))
		*wrqu = NETWORK_B;
	else if(TEST_FLAG(adapter, NETWORK_802_11_A))
		*wrqu = NETWORK_A;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_preamble(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	write_lock(&adapter->config_lock);

	/*
	 * We do not need to test if short preamble is available with the current rate.
	 * This is being checked when the adapter is updated. And the value is written to the register.
	 */
	adapter->config->preamble = (*wrqu == PREAMBLE_SHORT) ? PREAMBLE_SHORT_DURATION : PREAMBLE_LONG_DURATION;

	write_unlock(&adapter->config_lock);

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_get_preamble(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	*wrqu = adapter->config->preamble;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_antenna(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	u8			antenna = *wrqu;

	if(!(antenna & ANTENNA_RX_A) || !(antenna & ANTENNA_RX_B) || !(antenna & ANTENNA_RX_DIV)){
		DEBUG("Invalid rx-preamble value, switching rx_antenna to diversity.\n");
		antenna = ANTENNA_RX_DIV;
	}
	if(!(antenna & ANTENNA_TX_A) || !(antenna & ANTENNA_TX_B) || !(antenna & ANTENNA_TX_DIV)){
		DEBUG("Invalid tx-preamble value, switching tx_antenna to diversity.\n");
		antenna = ANTENNA_TX_DIV;
	}

	write_lock(&adapter->config_lock);

	adapter->config->antenna = antenna;

	write_unlock(&adapter->config_lock);

	return -EIWCOMMIT;
}

static int rt2x00_ioctl_get_antenna(
	struct net_device *net_dev,
	struct iw_request_info *info,
	char *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	*wrqu = adapter->config->antenna;

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_wpa(
	struct net_device *net_dev,
	struct iw_request_info *info,
	u32 *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _crypto		crypto;

	memset(&crypto, 0x00, sizeof(struct _crypto));

	crypto.type = CRYPTO_TYPE_WPA;

	if(copy_from_user(crypto.key, wrqu, CRYPTO_SIZE_WPA))
		return -EFAULT;

	crypto.length = CRYPTO_SIZE_WPA;

	return rt2x00_commit(adapter, &crypto, COMMIT_CRYPTO);
}

static int rt2x00_ioctl_get_wpa(
	struct net_device *net_dev,
	struct iw_request_info *info,
	u32 *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _crypto		crypto;

	if(!rt2x00_get_key(adapter, &crypto))
		return -ENODATA;

	if(!crypto.key || !(crypto.type & CRYPTO_TYPE_WPA))
		return -ENODATA;

	if(unlikely(crypto.length != CRYPTO_SIZE_WPA))
		return -EINVAL;

	read_lock(&adapter->config_lock);

	if(copy_to_user(wrqu, crypto.key, crypto.length)){
		read_unlock(&adapter->config_lock);
		return -EFAULT;
	}

	read_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_set_mac(
	struct net_device *net_dev,
	struct iw_request_info *info,
	struct sockaddr *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _mac_filter	*entry;
	struct _mac_filter	*next;
	u8			mac[ETH_ALEN];

	if(wrqu->sa_family != ARPHRD_ETHER)
		return -EINVAL;

	memset(&mac, 0x00, ETH_ALEN);
	memcpy(&mac, &wrqu->sa_data, ETH_ALEN);

	write_lock(&adapter->config_lock);

	if(!is_valid_ether_addr(mac))
		goto purge;

	list_for_each_entry(entry, adapter->mac_filter, list){
		if(memcmp(&entry->addr, &mac, ETH_ALEN) == 0)
			goto remove;
	}

	/* Entry not found, add address. */
	entry = kmalloc(sizeof(struct _mac_filter), GFP_KERNEL);
	if(!entry){
		write_unlock(&adapter->config_lock);
		return -ENOMEM;
	}

	memcpy(&entry->addr, &mac, ETH_ALEN);

	list_add_tail(&entry->list, adapter->mac_filter);

	if(list_empty(adapter->mac_filter)){
		NOTICE("Enabling MAC filtering.\n");
	}

	NOTICE("MAC address: " PRINT_ADDR_STRING " allowed to network.\n", PRINT_ADDR(entry->addr));

	write_unlock(&adapter->config_lock);

	return 0;

remove:
	NOTICE("MAC address: " PRINT_ADDR_STRING " blocked from network.\n", PRINT_ADDR(entry->addr));

	list_del(&entry->list);

	kfree(entry);

	write_unlock(&adapter->config_lock);

	return 0;

purge:
	list_for_each_entry_safe(entry, next, adapter->mac_filter, list) {
		list_del(&entry->list);
		kfree(entry);
	}

	NOTICE("Disabling MAC filtering.\n");

	write_unlock(&adapter->config_lock);

	return 0;
}

static int rt2x00_ioctl_get_mac(
	struct net_device *net_dev,
	struct iw_request_info *info,
	struct iw_point *wrqu,
	char *extra)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct _mac_filter	*entry;
	struct sockaddr		*dest = (struct sockaddr*)extra;

	read_lock(&adapter->config_lock);

	wrqu->length = 0;

	list_for_each_entry(entry, adapter->mac_filter, list){
		memcpy(dest->sa_data, entry->addr, ETH_ALEN);
		dest->sa_family = ARPHRD_ETHER;
		wrqu->length++;
		dest++;
		if(wrqu->length >= IW_MAX_AP)
			break;
	}

	read_unlock(&adapter->config_lock);

	return 0;
}

static const iw_handler rt2x00_ioctl_handler[] = {
	(iw_handler) rt2x00_ioctl_commit,		/* SIOCSIWCOMMIT */
	(iw_handler) rt2x00_ioctl_get_name,		/* SIOCGIWNAME */
	(iw_handler) NULL,				/* SIOCSIWNWID */
	(iw_handler) NULL,				/* SIOCGIWNWID */
	(iw_handler) rt2x00_ioctl_set_frequency,	/* SIOCSIWFREQ */
	(iw_handler) rt2x00_ioctl_get_frequency,	/* SIOCGIWFREQ */
	(iw_handler) rt2x00_ioctl_set_operation_mode,	/* SIOCSIWMODE */
	(iw_handler) rt2x00_ioctl_get_operation_mode,	/* SIOCGIWMODE */
	(iw_handler) rt2x00_ioctl_set_sens,		/* SIOCSIWSENS */
	(iw_handler) rt2x00_ioctl_get_sens,		/* SIOCGIWSENS */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWRANGE */
	(iw_handler) rt2x00_ioctl_get_range,		/* SIOCGIWRANGE */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWPRIV */
	(iw_handler) NULL, /* Provided by kernel */	/* SIOCGIWPRIV */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWSTATS */
	(iw_handler) NULL, /* Provided by kernel */	/* SIOCGIWSTATS */
	iw_handler_set_spy,				/* SIOCSIWSPY */
	iw_handler_get_spy,				/* SIOCGIWSPY */
	iw_handler_set_thrspy,				/* SIOCSIWTHRSPY */
	iw_handler_get_thrspy,				/* SIOCGIWTHRSPY */
	(iw_handler) rt2x00_ioctl_set_wap,		/* SIOCSIWAP */
	(iw_handler) rt2x00_ioctl_get_wap,		/* SIOCGIWAP */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) NULL, /* Deprecated */		/* SIOCGIWAPLIST */
	(iw_handler) rt2x00_ioctl_set_scan,		/* SIOCSIWSCAN */
	(iw_handler) rt2x00_ioctl_get_scan,		/* SIOCGIWSCAN */
	(iw_handler) rt2x00_ioctl_set_essid,		/* SIOCSIWESSID */
	(iw_handler) rt2x00_ioctl_get_essid,		/* SIOCGIWESSID */
	(iw_handler) rt2x00_ioctl_set_nickname,		/* SIOCSIWNICKN */
	(iw_handler) rt2x00_ioctl_get_nickname,		/* SIOCGIWNICKN */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) rt2x00_ioctl_set_bit_rate,		/* SIOCSIWRATE */
	(iw_handler) rt2x00_ioctl_get_bit_rate,		/* SIOCGIWRATE */
	(iw_handler) rt2x00_ioctl_set_rts_treshold,	/* SIOCSIWRTS */
	(iw_handler) rt2x00_ioctl_get_rts_treshold,	/* SIOCGIWRTS */
	(iw_handler) rt2x00_ioctl_set_fragmentation,	/* SIOCSIWFRAG */
	(iw_handler) rt2x00_ioctl_get_fragmentation,	/* SIOCGIWFRAG */
	(iw_handler) rt2x00_ioctl_set_txpower,		/* SIOCSIWTXPOW */
	(iw_handler) rt2x00_ioctl_get_txpower,		/* SIOCGIWTXPOW */
	(iw_handler) rt2x00_ioctl_set_retry,		/* SIOCSIWRETRY */
	(iw_handler) rt2x00_ioctl_get_retry,		/* SIOCGIWRETRY */
	(iw_handler) rt2x00_ioctl_set_encode,		/* SIOCSIWENCODE */
	(iw_handler) rt2x00_ioctl_get_encode,		/* SIOCGIWENCODE */
	(iw_handler) rt2x00_ioctl_set_power,		/* SIOCSIWPOWER */
	(iw_handler) rt2x00_ioctl_get_power,		/* SIOCGIWPOWER */
};

static const struct iw_priv_args rt2x00_ioctl_priv_args[] = {
	{ SIOCSIWPRIVNETWORK, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_network" },
	{ SIOCGIWPRIVNETWORK, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_network" },
	{ SIOCSIWPRIVPREAMBLE, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_preamble" },
	{ SIOCGIWPRIVPREAMBLE, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_preamble" },
	{ SIOCSIWPRIVANTENNA, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_antenna" },
	{ SIOCGIWPRIVANTENNA, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_antenna" },
	{ SIOCSIWPRIVWPA, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "set_wpa"},
	{ SIOCGIWPRIVWPA, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "get_wpa"},
	{ SIOCSIWPRIVMAC, IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0, "set_mac"},
	{ SIOCGIWPRIVMAC, 0, IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | IW_MAX_AP, "get_mac"},
};

static const iw_handler rt2x00_ioctl_priv_handler[] = {
	(iw_handler) rt2x00_ioctl_set_network,		/* SIOCSIWPRIVNETWORK */
	(iw_handler) rt2x00_ioctl_get_network,		/* SIOCGIWPRIVNETWORK */
	(iw_handler) rt2x00_ioctl_set_preamble,		/* SIOCSIWPRIVPREAMBLE */
	(iw_handler) rt2x00_ioctl_get_preamble,		/* SIOCGIWPRIVPREAMBLE */
	(iw_handler) rt2x00_ioctl_set_antenna,		/* SIOCSIWPRIVANTENNA */
	(iw_handler) rt2x00_ioctl_get_antenna,		/* SIOCGIWPRIVANTENNA */
	(iw_handler) rt2x00_ioctl_set_wpa,		/* SIOCSIWPRIVWPA */
	(iw_handler) rt2x00_ioctl_get_wpa,		/* SIOCGIWPRIVWPA */
	(iw_handler) rt2x00_ioctl_set_mac,		/* SIOCSIWPRIVMAC */
	(iw_handler) rt2x00_ioctl_get_mac,		/* SIOCGIWPRIVMAC */
};

const struct iw_handler_def rt2x00_ioctl_handler_def = {
	.num_standard		= ARRAY_SIZE(rt2x00_ioctl_handler),
	.num_private		= ARRAY_SIZE(rt2x00_ioctl_priv_handler),
	.num_private_args	= ARRAY_SIZE(rt2x00_ioctl_priv_args),
	.standard		= (iw_handler*)rt2x00_ioctl_handler,
	.private		= (iw_handler*)rt2x00_ioctl_priv_handler,
	.private_args		= (struct iw_priv_args*)rt2x00_ioctl_priv_args,
	.spy_offset		= offsetof(struct _rt2x00_adapter, spy_data),
};
