/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_main.c

	Abstract: non device specific routines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>

#include <net/iw_handler.h>

#include "rt2x00_adapter.h"

#ifndef CONFIG_NET_RADIO
#error Kernel has not been compiled with wireless extensions!
#endif /* CONFIG_NET_RADIO */

#if WIRELESS_EXT < 16
#error Kernel has been compiled with an unsupported wireless extensions version!
#endif /* WIRELESS_EXT < 16 */

int rt2x00_open(struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, ADAPTER_ENABLED)){
		read_unlock(&adapter->config_lock);
		return 0;
	}
	read_unlock(&adapter->config_lock);

	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(!try_module_get(THIS_MODULE)){
		write_unlock(&adapter->config_lock);
		spin_unlock(&adapter->adapter_lock);
		ERROR("Failure to acquire module.\n");
		return -ENODEV;
	}

	/* Start net interface tx /rx */
	netif_start_queue(net_dev);
	netif_carrier_on(net_dev);

	/* Interrupts.
	 * Before we can active the interrupts we have to set the IRQ mask we are going to work with.
	 * RT2500 has a lot more interrupts available then RT2400 but we do not intend to use them.
	 * To block those unused interrupts we will mask all but the first 7 interrupts.
	 * Before we load the irq_mask, we have to clear all garbage interrupts first.
	 */
	adapter->irq_mask = 0xffffff80;
	rt2x00_register_write(adapter, CSR7, 0xffffffff);
	rt2x00_enable_interrupt(adapter);

	SET_FLAG(adapter, ADAPTER_ENABLED);

	NOTICE("Adapter started.\n");

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	rt2x00_radio_on(adapter);

	return 0;
}

int rt2x00_stop(struct net_device *net_dev)
{
	struct _rt2x00_adapter *adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(!TEST_FLAG(adapter, ADAPTER_ENABLED)){
		read_unlock(&adapter->config_lock);
		return 0;
	}
	read_unlock(&adapter->config_lock);

	rt2x00_radio_off(adapter);

	/* del_timer_sync could race with the spinlocks so delete timers before 
	spin locking */
	
	del_timer_sync(&adapter->scan_timer);
	
	CLEAR_FLAG(adapter,ADAPTER_SCAN_IN_PROGRESS);
	del_timer_sync(&adapter->periodic);

	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	rt2x00_disable_interrupt(adapter);

	netif_stop_queue(net_dev);
	netif_carrier_off(net_dev);

	/* Issue reset and clear from reset state */
	rt2x00_register_write(adapter, CSR1, 0x01);
	rt2x00_register_write(adapter, CSR1, 0x00);

	CLEAR_FLAG(adapter, ADAPTER_ENABLED);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	module_put(THIS_MODULE);

	NOTICE("Adapter halted.\n");

	return 0;
}

int rt2x00_xmit(struct sk_buff *skb, struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	read_lock(&adapter->config_lock);

	if(!LINK_IS_UP(adapter) || LINK_TYPE(adapter, IW_MODE_MONITOR) || LINK_TYPE(adapter, IW_MODE_REPEAT)){
		read_unlock(&adapter->config_lock);
		return NET_XMIT_DROP;
	}

	read_unlock(&adapter->config_lock);

	/* system did not allocate packet correctly. */
	if(skb->data == NULL)
		return NET_XMIT_DROP;

	/* Invalid packet lenght. */
	if(skb->len < MIN_PACKET_SIZE || skb->len > MAX_PACKET_SIZE){
		adapter->stats.tx_dropped++;
		return NET_XMIT_DROP;
	}

	if(rt2x00_enqueue_packet(adapter, skb)){
		adapter->stats.tx_dropped++;
		return NET_XMIT_DROP;
	}

	net_dev->trans_start = jiffies;

	/* Schedule packet transmission. */
	tasklet_schedule(&adapter->tx_tasklet);

	dev_kfree_skb_any(skb);

	return NET_XMIT_SUCCESS;
}

struct net_device_stats* rt2x00_get_stats(struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	return &adapter->stats;
}

struct iw_statistics* rt2x00_get_wireless_stats(struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	return &adapter->iw_stats;
}

void rt2x00_set_multicast(struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	write_lock(&adapter->config_lock);

	/*
	 * Monitor and repeater mode must be capable to receive all packets.
	 */
	if(LINK_CFG_TYPE(adapter, IW_MODE_REPEAT) || LINK_CFG_TYPE(adapter, IW_MODE_MONITOR)){
		NOTICE("Promisc && All Multicast mode forced on!\n");
		SET_FLAG(net_dev, IFF_PROMISC);
		SET_FLAG(net_dev, IFF_ALLMULTI);
	}

	/*
	 * Master and secondary mode may only receive packets that are intended for them.
	 */
	if(LINK_CFG_TYPE(adapter, IW_MODE_MASTER) || LINK_CFG_TYPE(adapter, IW_MODE_SECOND)){
		NOTICE("Promisc mode forced off!\n");
		CLEAR_FLAG(net_dev, IFF_PROMISC);
	}

	if(TEST_FLAG(net_dev, IFF_PROMISC)){
		SET_FLAG(adapter, ACCEPT_PROMISC);
		NOTICE("Promisc mode enabled.\n");
	} else if(TEST_FLAG(adapter, ACCEPT_PROMISC)){
		CLEAR_FLAG(adapter, ACCEPT_PROMISC);
		NOTICE("Promisc mode disabled.\n");	
	}

	if(TEST_FLAG(net_dev, IFF_ALLMULTI)){
		SET_FLAG(adapter, ACCEPT_ALLMULTICAST);
		NOTICE("Accept all multicast packages.\n");
	} else if(TEST_FLAG(adapter, ACCEPT_ALLMULTICAST)){
		CLEAR_FLAG(adapter, ACCEPT_ALLMULTICAST);
		NOTICE("Do not accept all multicast packages.\n");
	}

	write_unlock(&adapter->config_lock);
}

int rt2x00_set_mac_address(struct net_device *net_dev, void *addr)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	struct sockaddr		*mac = (struct sockaddr*) addr;
	u32			set_mac;

	if(netif_running(net_dev))
		return -EBUSY;

	BUG_ON(net_dev->addr_len != ETH_ALEN);

	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	memcpy(net_dev->dev_addr, mac->sa_data, ETH_ALEN);

	memset(&set_mac, 0x00, sizeof(u32));
	set_mac =	(net_dev->dev_addr[0]) |
			(net_dev->dev_addr[1] << 8) |
			(net_dev->dev_addr[2] << 16) |
			(net_dev->dev_addr[3] << 24);

	rt2x00_register_write(adapter, CSR3, set_mac);

	memset(&set_mac, 0x00, sizeof(u32));
	set_mac =	(net_dev->dev_addr[4]) |
			(net_dev->dev_addr[5] << 8);

	rt2x00_register_write(adapter, CSR4, set_mac);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);

	INFO("Mac address changed to: " PRINT_ADDR_STRING ".\n", PRINT_ADDR(net_dev->dev_addr));

	return 0;
}

static int rt2x00_change_mtu(struct net_device *net_dev, int new_mtu)
{
	if((new_mtu < MIN_PACKET_SIZE) || (new_mtu > MAX_PACKET_SIZE))
		return -EINVAL;

	net_dev->mtu = new_mtu;

	return 0;
}

void rt2x00_tx_timeout(struct net_device *net_dev)
{
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	WARNING("TX timed out.\n");

	spin_lock_bh(&adapter->adapter_lock);

	rt2x00_disable_interrupt(adapter);

	/*
	 * We only have to cleanup the transmit packets.
	 */
	rt2x00_cleanup_dma_ring(adapter, TX_RING | PRIO_RING);

	rt2x00_enable_interrupt(adapter);

	netif_wake_queue(net_dev);

	adapter->stats.tx_errors++;

	spin_unlock_bh(&adapter->adapter_lock);
}

struct net_device* rt2x00_probe(u16 chip_type, void __iomem *csr_addr, struct _device_mem *dma)
{
	struct net_device	*net_dev = NULL;
	struct _rt2x00_adapter	*adapter = NULL;
	union _csr3		csr3;
	union _csr4		csr4;

	/*
	 * Allocate all required memory.
	 */

	net_dev = alloc_etherdev(sizeof(struct _rt2x00_adapter));
	if(!net_dev)
		goto exit_fail;
	adapter = netdev_priv(net_dev);
	memset(adapter, 0x00, sizeof(struct _rt2x00_adapter));

	adapter->capability = kmalloc(sizeof(struct _rt2x00_capability), GFP_KERNEL);
	if(!adapter->capability)
		goto exit_fail;
	memset(adapter->capability, 0x00, sizeof(struct _rt2x00_capability));

	adapter->config = kmalloc(sizeof(struct _rt2x00_config), GFP_KERNEL);
	if(!adapter->config)
		goto exit_fail;
	memset(adapter->config, 0x00, sizeof(struct _rt2x00_config));

	adapter->dma = kmalloc(sizeof(struct _rt2x00_dma), GFP_KERNEL);
	if(!adapter->dma)
		goto exit_fail;
	memset(adapter->dma, 0x00, sizeof(struct _rt2x00_dma));

	/*
	 * Initialize all spinlocks.
	 */
	spin_lock_init(&adapter->adapter_lock);
	rwlock_init(&adapter->config_lock);
	spin_lock_init(&adapter->rx_lock);
	spin_lock_init(&adapter->tx_lock);
	spin_lock_init(&adapter->atim_lock);
	spin_lock_init(&adapter->prio_lock);
	spin_lock_init(&adapter->beacon_lock);

	/*
	 * Save CSR virtual address and irq to device structure.
	 */
	adapter->register_base_address = csr_addr;

	adapter->net_dev = net_dev;

	adapter->chip_adapter = chip_type;

	adapter->dma_mem = dma;

	INIT_LIST_HEAD(&adapter->scan.result);

	init_timer(&adapter->scan_timer);
	init_timer(&adapter->periodic);

	if(rt2x00_init_adapter_eeprom(adapter))
		goto exit_fail;
	if(rt2x00_init_adapter_capable(adapter))
		goto exit_fail;
	if(rt2x00_init_adapter_config(adapter))
		goto exit_fail;
	if(rt2x00_init_schedule_irq(adapter))
		goto exit_fail;
	if(rt2x00_init_dma_ring(adapter))
		goto exit_fail;
	if(rt2x00_init_adapter_register(adapter))
		goto exit_fail;

	rt2x00_register_read(adapter, CSR3, &csr3.word);
	rt2x00_register_read(adapter, CSR4, &csr4.word);
	net_dev->dev_addr[0] = csr3.field.byte0;
	net_dev->dev_addr[1] = csr3.field.byte1;
	net_dev->dev_addr[2] = csr3.field.byte2;
	net_dev->dev_addr[3] = csr3.field.byte3;
	net_dev->dev_addr[4] = csr4.field.byte4;
	net_dev->dev_addr[5] = csr4.field.byte5;

	net_dev->addr_len = 6;

	net_dev->open = rt2x00_open;
	net_dev->stop = rt2x00_stop;
	net_dev->hard_start_xmit = rt2x00_xmit;
	net_dev->get_stats = rt2x00_get_stats;
	net_dev->get_wireless_stats = rt2x00_get_wireless_stats;
	net_dev->wireless_handlers = (struct iw_handler_def *) &rt2x00_ioctl_handler_def;
	net_dev->set_multicast_list = rt2x00_set_multicast;
	net_dev->set_mac_address = rt2x00_set_mac_address;
	net_dev->change_mtu = rt2x00_change_mtu;
	net_dev->tx_timeout = rt2x00_tx_timeout;
	net_dev->watchdog_timeo = TX_EXPIRE;

	return net_dev;

exit_fail:
	if(adapter->dma)
		kfree(adapter->dma);

	if(adapter->config)
		kfree(adapter->config);

	if(adapter->capability)
		kfree(adapter->capability);

	return NULL;
}

void rt2x00_remove(struct net_device *net_dev)
{
	struct _rt2x00_adapter *adapter = netdev_priv(net_dev);

	del_timer_sync(&adapter->scan_timer);
	del_timer_sync(&adapter->periodic);

	unregister_netdev(net_dev);

	kfree(adapter->dma);
	kfree(adapter->config);
	kfree(adapter->capability);

	free_netdev(net_dev);
}

irqreturn_t rt2x00_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct _rt2x00_adapter	*adapter = netdev_priv((struct net_device*)dev_instance);
	union _csr7		csr7;
	unsigned long		irqflag;

	spin_lock_irqsave(&adapter->adapter_lock, irqflag);
	write_lock(&adapter->config_lock);

	/* 1. Disable interrupt */
	rt2x00_disable_interrupt(adapter);

	/* 2. Get the interrupt sources & saved to local variable */
	rt2x00_register_read(adapter, CSR7, &csr7.word);
	/* Write register value back to clear pending interrupt. */
	rt2x00_register_write(adapter, CSR7, csr7.word);

	/* 3. Handle interrupt, walk through all bits and run the tasks. */
	if(csr7.field.tbcn_expire)/* Beacon timer expired interrupt. */
		rt2x00_interrupt_tcbn_expire(adapter);
	if(csr7.field.twake_expire)/* Wakeup timer expired interrupt. */
		rt2x00_interrupt_twake_expire(adapter);
	if(csr7.field.tatimw_expire)/* Atim window timer expired interrupt. */
		rt2x00_interrupt_tatimw_expire(adapter);
	if(csr7.field.txdone_txring)/* Tx ring transmit done interrupt. */
		rt2x00_interrupt_txdone_txring(adapter);
	if(csr7.field.txdone_atimring)/* Atim ring transmit done interrupt. */
		rt2x00_interrupt_txdone_atimring(adapter);
	if(csr7.field.txdone_prioring)/* Priority ring transmit done interrupt. */
		rt2x00_interrupt_txdone_prioring(adapter);
	if(csr7.field.rxdone)/* Rx ring done interrupt. */
		rt2x00_interrupt_rxdone(adapter);

	/* 4. Enable interrupt */
	rt2x00_enable_interrupt(adapter);

	write_unlock(&adapter->config_lock);
	spin_unlock_irqrestore(&adapter->adapter_lock, irqflag);
	return IRQ_HANDLED;
}

int rt2x00_suspend(struct _rt2x00_adapter *adapter, u32 state)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(!TEST_FLAG(adapter, ADAPTER_ENABLED)){
		write_unlock(&adapter->config_lock);
		spin_unlock(&adapter->adapter_lock);
		return 0;
	}

	rt2x00_disable_interrupt(adapter);

	NOTICE("Going to sleep.\n");

	/* Set BBP & RF desire_state to sleep mode. After this bring device to sleep. */
	rt2x00_register_write(adapter, PWRCSR1, 0x0000020a);

	netif_device_detach(adapter->net_dev);

	CLEAR_FLAG(adapter, ADAPTER_ENABLED);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
	return 0;
}

int rt2x00_resume(struct _rt2x00_adapter *adapter)
{
	spin_lock(&adapter->adapter_lock);
	write_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, ADAPTER_ENABLED)){
		write_unlock(&adapter->config_lock);
		spin_unlock(&adapter->adapter_lock);
		return 0;
	}

	rt2x00_enable_interrupt(adapter);

	NOTICE("Waking up.\n");

	/* Set BBP & RF desire_state to awake mode. After this wake up device. */
	rt2x00_register_write(adapter, PWRCSR1, 0x0000001f);

	netif_device_attach(adapter->net_dev);

	SET_FLAG(adapter, ADAPTER_ENABLED);

	write_unlock(&adapter->config_lock);
	spin_unlock(&adapter->adapter_lock);
	return 0;
}

static char version[] = DRV_NAME " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT "\n";

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("Ralink RT2400/RT2500 Wireless LAN driver. Released: " DRV_RELDATE);
MODULE_SUPPORTED_DEVICE("Ralink RT2400 & RT2500 chipset based cards");
MODULE_LICENSE("GPL");

/*
 * Module options.
 */
char enable_debug = 0;
MODULE_PARM(enable_debug, "s");
MODULE_PARM_DESC(enable_debug, "Set this value to 1 to enable debug output.");

/*
 * Symbols to be exported to the device specific modules.
 */
EXPORT_SYMBOL(rt2x00_probe);
EXPORT_SYMBOL(rt2x00_remove);
EXPORT_SYMBOL(rt2x00_interrupt);
EXPORT_SYMBOL(rt2x00_suspend);
EXPORT_SYMBOL(rt2x00_resume);
EXPORT_SYMBOL(enable_debug);

static int rt2x00_init(void)
{
	printk(KERN_INFO "Loading module: %s", version);
	return 0;
}

static void rt2x00_exit(void)
{
	printk(KERN_INFO "Unloading module: %s", version);
}

module_init(rt2x00_init);
module_exit(rt2x00_exit);
