/*-
 * Copyright (c) 2005
 *	Damien Bergamini <damien.bergamini@free.fr>
 * Copyright (c) 2005
 *	Jeremy Bobbio <jeremy.bobbio@etu.upmc.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#define RAL_DEBUG 1

__FBSDID("$Id: if_ral.c,v 1.5 2005/04/02 12:14:48 damien Exp $");

/*-
 * Ralink Technology RT2500USB chipset driver
 * http://www.ralinktech.com/
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/etherdevice.h>
#include <linux/usb.h>
#include <linux/types.h>
#include <linux/ethtool.h>
#include <linux/sysctl.h>
#include <linux/if_arp.h>
#include <linux/wireless.h>
#include <net/iw_handler.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/byteorder.h>

#include <net80211/if_ethersubr.h>
#include <net80211/if_media.h>
#include <net80211/if_llc.h>
#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_radiotap.h>

#include "if_ralreg.h"
#include "if_ralvar.h"

#ifdef RAL_DEBUG
#define DPRINTF(x)	do { if (ural_debug > 0) printf x; } while (0)
#define DPRINTFN(n, x)	do { if (ural_debug >= (n)) printf x; } while (0)
int ural_debug = 2;
#else
#define DPRINTF(x)
#define DPRINTFN(n, x)
#endif

/* Version Information */
#define DRIVER_NAME "ural"
#define DRIVER_VERSION "v0.8.1"
#define DRIVER_AUTHOR "Jeremy Bobbio <jeremy.bobbio@etu.upmc.fr>"
#define DRIVER_DESC "Ralink Technology RT2500USB chipset driver"

#define RAL_TX_TIMEOUT	5000
#define SCAN_JIFFIES	(HZ / 5)

/* various supported device vendors/products */
static const struct usb_device_id ural_table[] = {
	{USB_DEVICE(VENDOR_ID_ASUS, PRODUCT_ID_ASUS_RT2570)},
	{USB_DEVICE(VENDOR_ID_ASUS, PRODUCT_ID_RALINK_RT2570)},
	{USB_DEVICE(VENDOR_ID_CONCEPTRONIC, PRODUCT_ID_CONCEPTRONIC_C54U)},
	{USB_DEVICE(VENDOR_ID_DLINK, PRODUCT_ID_DLINK_RT2570)},
        {USB_DEVICE(VENDOR_ID_CISCOLINKSYS, PRODUCT_ID_CISCOLINKSYS_WUSB54G)},
        {USB_DEVICE(VENDOR_ID_CISCOLINKSYS, PRODUCT_ID_CISCOLINKSYS_WUSB54GP)},
	{USB_DEVICE(VENDOR_ID_MELCO, PRODUCT_ID_MELCO_KG54) },
	{USB_DEVICE(VENDOR_ID_RALINK, PRODUCT_ID_RALINK_RT2570)},
	{USB_DEVICE(VENDOR_ID_SMC, PRODUCT_ID_SMC_RT2570)},
	{}
};

MODULE_DEVICE_TABLE(usb, ural_table);

static void		ural_fill_skb_pool(struct ural_softc *);
static void		ural_free_skb_pool(struct ural_softc *);
static struct sk_buff * ural_pull_skb(struct ural_softc *);
static void		ural_rx_fixup(unsigned long data);
static int		ural_media_change(struct net_device *);
static void		ural_next_scan(void *);
static void		ural_work(void *);
static int		ural_newstate(struct ieee80211com *,
			    enum ieee80211_state, int);
static void		ural_txeof(struct urb *, struct pt_regs *);
static void		ural_rxeof(struct urb *, struct pt_regs *);
static int		ural_ack_rate(int);
static u_int16_t	ural_txtime(int, int, u_int32_t);
static u_int8_t		ural_plcp_signal(int);
static void		ural_setup_tx_desc(struct ural_softc *,
			    struct ural_tx_desc *, u_int32_t, int, int);
static int		ural_tx_bcn(struct ural_softc *, struct sk_buff *,
			    struct ieee80211_node *);
static int		ural_tx_mgt(struct ieee80211com *ic,
			    struct sk_buff *skb0);
static int		ural_tx_data(struct ural_softc *, struct sk_buff *,
			    struct ieee80211_node *);
static int		ural_start_xmit(struct sk_buff *skb,
			    struct net_device *dev);
static void		ural_tx_timeout(struct net_device *);
static void		ural_rx_capture(struct net_device *,
			    struct ural_rx_desc *, struct sk_buff *);
static int		ural_reset(struct net_device *);
static int		ural_ioctl(struct net_device *, struct ifreq *, int);
static struct net_device_stats *
			ural_getstats(struct net_device *);
static struct iw_statistics *
			ural_iw_getstats(struct net_device *dev);
static void		ural_eeprom_read(struct ural_softc *, u_int16_t, void *,
			    int);
static u_int16_t	ural_read(struct ural_softc *, u_int16_t);
static void		ural_read_multi(struct ural_softc *, u_int16_t, void *,
			    int);
static void		ural_write(struct ural_softc *, u_int16_t, u_int16_t);
static void		ural_write_multi(struct ural_softc *, u_int16_t,
			    u_int8_t *, u_int16_t);
static void		ural_bbp_write(struct ural_softc *, u_int8_t, u_int8_t);
static u_int8_t		ural_bbp_read(struct ural_softc *, u_int8_t);
static void		ural_rf_write(struct ural_softc *, u_int8_t, u_int32_t);
static void		ural_set_chan(struct ural_softc *,
			    struct ieee80211_channel *);
#if 0
static void		ural_disable_rf_tune(struct ural_softc *);
#endif
static void		ural_enable_tsf_sync(struct ural_softc *);
static void		ural_set_bssid(struct ural_softc *, u_int8_t *);
static int		ural_set_macaddr(struct net_device *, void *);
static int		ural_set_macaddr_locked(struct net_device *, void *);
static void		ural_set_multicast_list(struct net_device *);
static void		ural_mode_init(void *);
static const char *	ural_get_rf(int);
static void		ural_read_eeprom(struct ural_softc *);
static int		ural_bbp_init(struct ural_softc *);
static void		ural_set_txantenna(struct ural_softc *, int);
static void		ural_set_rxantenna(struct ural_softc *, int);
static int 		ural_find_endpoints(struct ural_softc *,
			    struct usb_interface *);
static int		ural_probe(struct usb_interface *,
			    const struct usb_device_id *);
static void		ural_disconnect(struct usb_interface *);
static void		ural_init_wq(void *);
static int		ural_init(struct net_device *);
static int		ural_init_locked(struct net_device *);
static int		ural_stop(struct net_device *);
static int		ural_stop_locked(struct net_device *);

/*
 * Supported rates for 802.11a/b/g modes (in 500Kbps unit).
 */
static const struct ieee80211_rateset ural_rateset_11a =
	{ 8, { 12, 18, 24, 36, 48, 72, 96, 108 } };

static const struct ieee80211_rateset ural_rateset_11b =
	{ 4, { 2, 4, 11, 22 } };

static const struct ieee80211_rateset ural_rateset_11g =
	{ 12, { 2, 4, 11, 22, 12, 18, 24, 36, 48, 72, 96, 108 } };

/*
 * Default values for MAC registers; values taken from the reference driver.
 */
static const struct {
	u_int16_t	reg;
	u_int16_t	val;
} ural_def_mac[] = {
	{ RAL_TXRX_CSR5,  0x8c8d },
	{ RAL_TXRX_CSR6,  0x8b8a },
	{ RAL_TXRX_CSR7,  0x8687 },
	{ RAL_TXRX_CSR8,  0x0085 },
	{ RAL_MAC_CSR13,  0x1111 },
	{ RAL_MAC_CSR14,  0x1e11 },
	{ RAL_TXRX_CSR21, 0xe78f },
	{ RAL_MAC_CSR9,   0xff1d },
	{ RAL_MAC_CSR11,  0x0002 },
	{ RAL_MAC_CSR22,  0x0053 },
	{ RAL_MAC_CSR15,  0x0000 },
	{ RAL_MAC_CSR8,   0x0780 },
	{ RAL_TXRX_CSR19, 0x0000 },
	{ RAL_TXRX_CSR18, 0x005a },
	{ RAL_PHY_CSR2,   0x0000 },
	{ RAL_TXRX_CSR0,  0x1ec0 },
	{ RAL_PHY_CSR4,   0x000f }
};

/*
 * Default values for BBP registers; values taken from the reference driver.
 */
static const struct {
	u_int8_t	reg;
	u_int8_t	val;
} ural_def_bbp[] = {
	{  3, 0x02 },
	{  4, 0x19 },
	{ 14, 0x1c },
	{ 15, 0x30 },
	{ 16, 0xac },
	{ 17, 0x48 },
	{ 18, 0x18 },
	{ 19, 0xff },
	{ 20, 0x1e },
	{ 21, 0x08 },
	{ 22, 0x08 },
	{ 23, 0x08 },
	{ 24, 0x80 },
	{ 25, 0x50 },
	{ 26, 0x08 },
	{ 27, 0x23 },
	{ 30, 0x10 },
	{ 31, 0x2b },
	{ 32, 0xb9 },
	{ 34, 0x12 },
	{ 35, 0x50 },
	{ 39, 0xc4 },
	{ 40, 0x02 },
	{ 41, 0x60 },
	{ 53, 0x10 },
	{ 54, 0x18 },
	{ 56, 0x08 },
	{ 57, 0x10 },
	{ 58, 0x08 },
	{ 61, 0x60 },
	{ 62, 0x10 },
	{ 75, 0xff }
};

/*
 * Default values for RF register R2 indexed by channel numbers.
 */
static const u_int32_t ural_rf2522_r2[] = {
	0x307f6, 0x307fb, 0x30800, 0x30805, 0x3080a, 0x3080f, 0x30814,
	0x30819, 0x3081e, 0x30823, 0x30828, 0x3082d, 0x30832, 0x3083e
};

static const u_int32_t ural_rf2523_r2[] = {
	0x00327, 0x00328, 0x00329, 0x0032a, 0x0032b, 0x0032c, 0x0032d,
	0x0032e, 0x0032f, 0x00340, 0x00341, 0x00342, 0x00343, 0x00346
};

static const u_int32_t ural_rf2524_r2[] = {
	0x00327, 0x00328, 0x00329, 0x0032a, 0x0032b, 0x0032c, 0x0032d,
	0x0032e, 0x0032f, 0x00340, 0x00341, 0x00342, 0x00343, 0x00346
};

static const u_int32_t ural_rf2525_r2[] = {
	0x20327, 0x20328, 0x20329, 0x2032a, 0x2032b, 0x2032c, 0x2032d,
	0x2032e, 0x2032f, 0x20340, 0x20341, 0x20342, 0x20343, 0x20346
};

static const u_int32_t ural_rf2525_hi_r2[] = {
	0x2032f, 0x20340, 0x20341, 0x20342, 0x20343, 0x20344, 0x20345,
	0x20346, 0x20347, 0x20348, 0x20349, 0x2034a, 0x2034b, 0x2034e
};

static const u_int32_t ural_rf2525e_r2[] = {
	0x2044d, 0x2044e, 0x2044f, 0x20460, 0x20461, 0x20462, 0x20463,
	0x20464, 0x20465, 0x20466, 0x20467, 0x20468, 0x20469, 0x2046b
};

static const u_int32_t ural_rf2526_hi_r2[] = {
	0x0022a, 0x0022b, 0x0022b, 0x0022c, 0x0022c, 0x0022d, 0x0022d,
	0x0022e, 0x0022e, 0x0022f, 0x0022d, 0x00240, 0x00240, 0x00241
};

static const u_int32_t ural_rf2526_r2[] = {
	0x00226, 0x00227, 0x00227, 0x00228, 0x00228, 0x00229, 0x00229,
	0x0022a, 0x0022a, 0x0022b, 0x0022b, 0x0022c, 0x0022c, 0x0022d
};

/*
 * For dual-band RF, RF registers R1 and R4 also depend on channel number;
 * values taken from the reference driver.
 */
static const struct {
	u_int8_t	chan;
	u_int32_t	r1;
	u_int32_t	r2;
	u_int32_t	r4;
} ural_rf5222[] = {
	/* channels in the 2.4GHz band */
	{   1, 0x08808, 0x0044d, 0x00282 },
	{   2, 0x08808, 0x0044e, 0x00282 },
	{   3, 0x08808, 0x0044f, 0x00282 },
	{   4, 0x08808, 0x00460, 0x00282 },
	{   5, 0x08808, 0x00461, 0x00282 },
	{   6, 0x08808, 0x00462, 0x00282 },
	{   7, 0x08808, 0x00463, 0x00282 },
	{   8, 0x08808, 0x00464, 0x00282 },
	{   9, 0x08808, 0x00465, 0x00282 },
	{  10, 0x08808, 0x00466, 0x00282 },
	{  11, 0x08808, 0x00467, 0x00282 },
	{  12, 0x08808, 0x00468, 0x00282 },
	{  13, 0x08808, 0x00469, 0x00282 },
	{  14, 0x08808, 0x0046b, 0x00286 },

	/* channels in the 5.2GHz band */
	{  36, 0x08804, 0x06225, 0x00287 },
	{  40, 0x08804, 0x06226, 0x00287 },
	{  44, 0x08804, 0x06227, 0x00287 },
	{  48, 0x08804, 0x06228, 0x00287 },
	{  52, 0x08804, 0x06229, 0x00287 },
	{  56, 0x08804, 0x0622a, 0x00287 },
	{  60, 0x08804, 0x0622b, 0x00287 },
	{  64, 0x08804, 0x0622c, 0x00287 },

	{ 100, 0x08804, 0x02200, 0x00283 },
	{ 104, 0x08804, 0x02201, 0x00283 },
	{ 108, 0x08804, 0x02202, 0x00283 },
	{ 112, 0x08804, 0x02203, 0x00283 },
	{ 116, 0x08804, 0x02204, 0x00283 },
	{ 120, 0x08804, 0x02205, 0x00283 },
	{ 124, 0x08804, 0x02206, 0x00283 },
	{ 128, 0x08804, 0x02207, 0x00283 },
	{ 132, 0x08804, 0x02208, 0x00283 },
	{ 136, 0x08804, 0x02209, 0x00283 },
	{ 140, 0x08804, 0x0220a, 0x00283 },

	{ 149, 0x08808, 0x02429, 0x00281 },
	{ 153, 0x08808, 0x0242b, 0x00281 },
	{ 157, 0x08808, 0x0242d, 0x00281 },
	{ 161, 0x08808, 0x0242f, 0x00281 }
};

#define RAL_CHAR_BOUNCE(name)                                           \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 char *erq, char *extra)                                \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_POINT_BOUNCE(name)                                          \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 struct iw_point *erq, char *extra)                     \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_PARAM_BOUNCE(name)                                          \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 struct iw_param *erq, char *extra)                     \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_SOCKADDR_BOUNCE(name)                                       \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 struct sockaddr *erq, char *extra)                     \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_FREQ_BOUNCE(name)                                           \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 struct iw_freq *erq, char *extra)                      \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_U32_BOUNCE(name)                                            \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 __u32 *erq, char *extra)                               \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
        return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}
#define RAL_VOID_BOUNCE(name)                                           \
static int                                                              \
ural_ioctl_##name(struct net_device *dev, struct iw_request_info *info,  \
		 void *erq, char *extra)                                \
{                                                                       \
	struct ural_softc *sc = netdev_priv(dev);                        \
	return ieee80211_ioctl_##name(&sc->sc_ic, info, erq, extra);    \
}

RAL_CHAR_BOUNCE(giwname)
RAL_POINT_BOUNCE(siwencode)
RAL_POINT_BOUNCE(giwencode)
RAL_PARAM_BOUNCE(siwrate)
RAL_PARAM_BOUNCE(giwrate)
RAL_PARAM_BOUNCE(siwsens)
RAL_PARAM_BOUNCE(giwsens)
RAL_PARAM_BOUNCE(siwrts)
RAL_PARAM_BOUNCE(giwrts)
RAL_PARAM_BOUNCE(siwfrag)
RAL_PARAM_BOUNCE(giwfrag)
RAL_SOCKADDR_BOUNCE(siwap)
RAL_SOCKADDR_BOUNCE(giwap)
RAL_POINT_BOUNCE(siwnickn)
RAL_POINT_BOUNCE(giwnickn)
RAL_FREQ_BOUNCE(siwfreq)
RAL_FREQ_BOUNCE(giwfreq)
RAL_POINT_BOUNCE(siwessid)
RAL_POINT_BOUNCE(giwessid)
RAL_POINT_BOUNCE(giwrange)
RAL_U32_BOUNCE(siwmode)
RAL_U32_BOUNCE(giwmode)
RAL_PARAM_BOUNCE(siwpower)
RAL_PARAM_BOUNCE(giwpower)
RAL_PARAM_BOUNCE(siwretry)
RAL_PARAM_BOUNCE(giwretry)
RAL_PARAM_BOUNCE(siwtxpow)
RAL_PARAM_BOUNCE(giwtxpow)
RAL_POINT_BOUNCE(iwaplist)
#ifdef SIOCGIWSCAN
RAL_POINT_BOUNCE(siwscan)
RAL_POINT_BOUNCE(giwscan)
#endif
RAL_VOID_BOUNCE(setparam)
RAL_VOID_BOUNCE(getparam)
RAL_VOID_BOUNCE(setkey)
RAL_VOID_BOUNCE(delkey)
RAL_VOID_BOUNCE(setmlme)
RAL_VOID_BOUNCE(setoptie)
RAL_VOID_BOUNCE(getoptie)
RAL_VOID_BOUNCE(addmac)
RAL_VOID_BOUNCE(delmac)
RAL_VOID_BOUNCE(chanlist)

/* Structures to export the Wireless Handlers */
static const iw_handler ural_handlers[] = {
	(iw_handler) NULL,                              /* SIOCSIWCOMMIT */
	(iw_handler) ural_ioctl_giwname,                 /* SIOCGIWNAME */
	(iw_handler) NULL,                              /* SIOCSIWNWID */
	(iw_handler) NULL,                              /* SIOCGIWNWID */
	(iw_handler) ural_ioctl_siwfreq,                 /* SIOCSIWFREQ */
	(iw_handler) ural_ioctl_giwfreq,                 /* SIOCGIWFREQ */
	(iw_handler) ural_ioctl_siwmode,                 /* SIOCSIWMODE */
	(iw_handler) ural_ioctl_giwmode,                 /* SIOCGIWMODE */
	(iw_handler) ural_ioctl_siwsens,                 /* SIOCSIWSENS */
	(iw_handler) ural_ioctl_giwsens,                 /* SIOCGIWSENS */
	(iw_handler) NULL /* not used */,               /* SIOCSIWRANGE */
	(iw_handler) ural_ioctl_giwrange,                /* SIOCGIWRANGE */
	(iw_handler) NULL /* not used */,               /* SIOCSIWPRIV */
	(iw_handler) NULL /* kernel code */,            /* SIOCGIWPRIV */
	(iw_handler) NULL /* not used */,               /* SIOCSIWSTATS */
	(iw_handler) NULL /* kernel code */,            /* SIOCGIWSTATS */
	(iw_handler) NULL,                              /* SIOCSIWSPY */
	(iw_handler) NULL,                              /* SIOCGIWSPY */
	(iw_handler) NULL,                              /* -- hole -- */
	(iw_handler) NULL,                              /* -- hole -- */
	(iw_handler) ural_ioctl_siwap,                   /* SIOCSIWAP */
	(iw_handler) ural_ioctl_giwap,                   /* SIOCGIWAP */
	(iw_handler) NULL,                              /* -- hole -- */
	(iw_handler) ural_ioctl_iwaplist,                /* SIOCGIWAPLIST */
#ifdef SIOCGIWSCAN
	(iw_handler) ural_ioctl_siwscan,                 /* SIOCSIWSCAN */
	(iw_handler) ural_ioctl_giwscan,                 /* SIOCGIWSCAN */
#else
	(iw_handler) NULL,                              /* SIOCSIWSCAN */
	(iw_handler) NULL,                              /* SIOCGIWSCAN */
#endif /* SIOCGIWSCAN */
	(iw_handler) ural_ioctl_siwessid,                /* SIOCSIWESSID */
	(iw_handler) ural_ioctl_giwessid,                /* SIOCGIWESSID */
	(iw_handler) ural_ioctl_siwnickn,                /* SIOCSIWNICKN */
	(iw_handler) ural_ioctl_giwnickn,                /* SIOCGIWNICKN */
	(iw_handler) NULL,                              /* -- hole -- */
	(iw_handler) NULL,                              /* -- hole -- */
	(iw_handler) ural_ioctl_siwrate,                 /* SIOCSIWRATE */
	(iw_handler) ural_ioctl_giwrate,                 /* SIOCGIWRATE */
	(iw_handler) ural_ioctl_siwrts,                  /* SIOCSIWRTS */
	(iw_handler) ural_ioctl_giwrts,                  /* SIOCGIWRTS */
	(iw_handler) ural_ioctl_siwfrag,                 /* SIOCSIWFRAG */
	(iw_handler) ural_ioctl_giwfrag,                 /* SIOCGIWFRAG */
	(iw_handler) ural_ioctl_siwtxpow,                /* SIOCSIWTXPOW */
	(iw_handler) ural_ioctl_giwtxpow,                /* SIOCGIWTXPOW */
	(iw_handler) ural_ioctl_siwretry,                /* SIOCSIWRETRY */
	(iw_handler) ural_ioctl_giwretry,                /* SIOCGIWRETRY */
	(iw_handler) ural_ioctl_siwencode,               /* SIOCSIWENCODE */
	(iw_handler) ural_ioctl_giwencode,               /* SIOCGIWENCODE */
	(iw_handler) ural_ioctl_siwpower,                /* SIOCSIWPOWER */
	(iw_handler) ural_ioctl_giwpower,                /* SIOCGIWPOWER */
};
static const iw_handler ural_priv_handlers[] = {
	(iw_handler) ural_ioctl_setparam,                /* SIOCWFIRSTPRIV+0 */
	(iw_handler) ural_ioctl_getparam,                /* SIOCWFIRSTPRIV+1 */
	(iw_handler) ural_ioctl_setkey,                  /* SIOCWFIRSTPRIV+2 */
	(iw_handler) NULL,                              /* SIOCWFIRSTPRIV+3 */
	(iw_handler) ural_ioctl_delkey,                  /* SIOCWFIRSTPRIV+4 */
	(iw_handler) NULL,                              /* SIOCWFIRSTPRIV+5 */
	(iw_handler) ural_ioctl_setmlme,                 /* SIOCWFIRSTPRIV+6 */
	(iw_handler) NULL,                              /* SIOCWFIRSTPRIV+7 */
	(iw_handler) ural_ioctl_setoptie,                /* SIOCWFIRSTPRIV+8 */
	(iw_handler) ural_ioctl_getoptie,                /* SIOCWFIRSTPRIV+9 */
	(iw_handler) ural_ioctl_addmac,                  /* SIOCWFIRSTPRIV+10 */
	(iw_handler) NULL,                              /* SIOCWFIRSTPRIV+11 */
	(iw_handler) ural_ioctl_delmac,                  /* SIOCWFIRSTPRIV+12 */
	(iw_handler) NULL,                              /* SIOCWFIRSTPRIV+13 */
	(iw_handler) ural_ioctl_chanlist,                /* SIOCWFIRSTPRIV+14 */
};

static struct iw_handler_def ural_iw_handler_def = {     
#define N(a)    (sizeof (a) / sizeof (a[0]))            
	.standard               = (iw_handler *) ural_handlers,
	.num_standard           = N(ural_handlers),
	.private                = (iw_handler *) ural_priv_handlers,
	.num_private            = N(ural_priv_handlers),
#undef N
};

static int ural_find_endpoints(struct ural_softc *sc,
		               struct usb_interface *intf)
{
	int i;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *ed;

	iface_desc = intf->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		ed = &iface_desc->endpoint[i].desc;
		if (!(ed->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK) {
			continue;
		}
		if (!sc->sc_rx_no && (ed->bEndpointAddress & USB_DIR_IN)) {
			sc->sc_rx_no = ed->bEndpointAddress &
					USB_ENDPOINT_NUMBER_MASK;
		}
		if (!sc->sc_tx_no) {
			sc->sc_tx_no = ed->bEndpointAddress &
					USB_ENDPOINT_NUMBER_MASK;
		}
	}
	if (!sc->sc_rx_no || !sc->sc_tx_no) {
		err("Could not find bulk endpoints - rx %u tx %u\n",
		    sc->sc_rx_no, sc->sc_tx_no);
		return -EIO;
	}
	return 0;
}

static int ural_probe(struct usb_interface *intf,
		      const struct usb_device_id *id)
{
#define N(a)	(sizeof (a) / sizeof ((a)[0]))
	struct ural_softc *sc = NULL;
	struct net_device *netdev;
	struct ieee80211com *ic;
	int i;
	int retval = -ENOMEM;

	if (!(netdev = alloc_etherdev(sizeof (struct ural_softc)))) {
		err("can't alloc etherdev");
		goto error;
	}
	sc = netdev_priv(netdev);
	memset(sc, 0, sizeof (struct ural_softc));
	sc->udev = usb_get_dev(interface_to_usbdev(intf));
	sc->netdev = netdev;
	SET_MODULE_OWNER(netdev);
	ic = &sc->sc_ic;

	sc->sc_rx_pipe = usb_rcvbulkpipe(sc->udev, sc->sc_rx_no);
	sc->sc_tx_pipe = usb_sndbulkpipe(sc->udev, sc->sc_tx_no);

	/* save our data pointer in this interface device */
	usb_set_intfdata(intf, sc);

	if (0 != (retval = ural_find_endpoints(sc, intf))) {
		goto error;
	}

	RAL_LOCK_INIT(sc);
	INIT_WORK(&sc->sc_work, ural_work, sc);
	INIT_WORK(&sc->sc_scan_ch, ural_next_scan, sc);
	INIT_WORK(&sc->sc_init, ural_init_wq, netdev);
	INIT_WORK(&sc->sc_mode_init, ural_mode_init, sc);

	tasklet_init(&sc->rx_fixup_task, ural_rx_fixup, (unsigned long) sc);
	spin_lock_init(&sc->rx_pool_lock);

	/* retrieve RT2570 rev. no */
	sc->asic_rev = ural_read(sc, RAL_MAC_CSR0);

	/* retrieve MAC address and various other things from EEPROM */
	ural_read_eeprom(sc);
	IEEE80211_ADDR_COPY(netdev->dev_addr, ic->ic_myaddr);

	netdev->flags = IFF_BROADCAST | IFF_MULTICAST;
	netdev->open = ural_init;
	netdev->stop = ural_stop;
	netdev->hard_start_xmit = ural_start_xmit;
	netdev->tx_timeout = ural_tx_timeout;
	netdev->watchdog_timeo = 5 * HZ;
	netdev->set_multicast_list = ural_set_multicast_list;
	netdev->do_ioctl = ural_ioctl;
	netdev->get_stats = ural_getstats;
	netdev->set_mac_address = ural_set_macaddr;
	netdev->get_wireless_stats = ural_iw_getstats;
	ieee80211_ioctl_iwsetup(&ural_iw_handler_def);
	netdev->wireless_handlers = &ural_iw_handler_def;

	ic->ic_dev = netdev;
	ic->ic_phytype = IEEE80211_T_OFDM; /* not only, but not used */
	ic->ic_opmode = IEEE80211_M_STA; /* default to BSS mode */
	ic->ic_state = IEEE80211_S_INIT;

	/* set device capabilities */
	ic->ic_caps = IEEE80211_C_MONITOR | IEEE80211_C_IBSS |
	    IEEE80211_C_HOSTAP | IEEE80211_C_SHPREAMBLE | IEEE80211_C_PMGT |
	    IEEE80211_C_TXPMGT | IEEE80211_C_WEP;

	if (sc->rf_rev == RAL_RF_5222) {
		/* set supported .11a rates */
		ic->ic_sup_rates[IEEE80211_MODE_11A] = ural_rateset_11a;

		/* set supported .11a channels */
		for (i = 36; i <= 64; i += 4) {
			ic->ic_channels[i].ic_freq =
			    ieee80211_ieee2mhz(i, IEEE80211_CHAN_5GHZ);
			ic->ic_channels[i].ic_flags = IEEE80211_CHAN_A;
		}
		for (i = 100; i <= 140; i += 4) {
			ic->ic_channels[i].ic_freq =
			    ieee80211_ieee2mhz(i, IEEE80211_CHAN_5GHZ);
			ic->ic_channels[i].ic_flags = IEEE80211_CHAN_A;
		}
		for (i = 149; i <= 161; i += 4) {
			ic->ic_channels[i].ic_freq =
			    ieee80211_ieee2mhz(i, IEEE80211_CHAN_5GHZ);
			ic->ic_channels[i].ic_flags = IEEE80211_CHAN_A;
		}
	}

	/* set supported .11b and .11g rates */
	ic->ic_sup_rates[IEEE80211_MODE_11B] = ural_rateset_11b;
	ic->ic_sup_rates[IEEE80211_MODE_11G] = ural_rateset_11g;

	/* set supported .11b and .11g channels (1 through 14) */
	for (i = 1; i <= 14; i++) {
		ic->ic_channels[i].ic_freq =
		    ieee80211_ieee2mhz(i, IEEE80211_CHAN_2GHZ);
		ic->ic_channels[i].ic_flags =
		    IEEE80211_CHAN_CCK | IEEE80211_CHAN_OFDM |
		    IEEE80211_CHAN_DYN | IEEE80211_CHAN_2GHZ;
	}

	ieee80211_ifattach(ic);
	ic->ic_devstats = &sc->sc_devstats;
	ic->ic_mgtstart = ural_tx_mgt;
	ic->ic_init = ural_init;
	ic->ic_reset = ural_reset;

	/* override state transition machine */
	sc->sc_newstate = ic->ic_newstate;
	ic->ic_newstate = ural_newstate;
	ieee80211_media_init(ic, ural_media_change, ieee80211_media_status);

	if (register_netdev(netdev)) {
		err("%s: unable to register device\n", netdev->name);
		retval = -EIO;
		goto error;
	}

	ieee80211_announce(ic);
	printk("%s: RT2570 (rev 0x%02x), RF %s, ether %s at usb-%s-%s\n",
	    sc->netdev->name, sc->asic_rev, ural_get_rf(sc->rf_rev),
	    ether_sprintf(sc->netdev->dev_addr), sc->udev->bus->bus_name,
	    sc->udev->devpath);

	return retval;

error:
	if (sc) {
		ural_disconnect(intf);
	}
	return retval;
#undef N
}

static void ural_disconnect(struct usb_interface *intf)
{
	struct ural_softc *sc;
	struct ieee80211com *ic;
	struct net_device *netdev;

	sc = usb_get_intfdata(intf);
	usb_set_intfdata(intf, NULL);
	if (!sc) {
		return;
	}

	ic = &sc->sc_ic;
	netdev = sc->netdev;

	sc->sc_flags |= RAL_UNPLUG;

	tasklet_kill(&sc->rx_fixup_task);
	
	cancel_delayed_work(&sc->sc_scan_ch);
	flush_scheduled_work();
	cancel_delayed_work(&sc->sc_work);

	ieee80211_ifdetach(ic);

	if (sc->sc_tx_ni != NULL && sc->sc_tx_ni != sc->sc_ic.ic_bss) {
		ieee80211_free_node(&sc->sc_ic, sc->sc_tx_ni);
		sc->sc_tx_ni = NULL;
	}

	if (sc->sc_rx_urb != NULL) {
		usb_kill_urb(sc->sc_rx_urb);
		usb_free_urb(sc->sc_rx_urb);
		sc->sc_rx_urb = NULL;
	}
	spin_lock_irq(&sc->rx_pool_lock);
	ural_free_skb_pool(sc);
	spin_unlock_irq(&sc->rx_pool_lock);

	if (sc->sc_tx_urb != NULL) {
		usb_kill_urb(sc->sc_tx_urb);
		usb_free_urb(sc->sc_tx_urb);
		sc->sc_tx_urb = NULL;
	}
	if (sc->sc_tx_skb != NULL) {
		dev_kfree_skb(sc->sc_tx_skb);
		sc->sc_tx_skb = NULL;
	}

	RAL_LOCK_DESTROY(sc);
	
	unregister_netdev(netdev);

	usb_put_dev(sc->udev);
	if (NULL != sc->netdev) {
		free_netdev(sc->netdev);
	}
	kfree(sc);

	return;
}

static void
ural_rx_fixup(unsigned long data)
{
	struct ural_softc *sc = (struct ural_softc *) data;
	struct sk_buff *skb;

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return;
	}
	
	spin_lock_irq(&sc->rx_pool_lock);
	ural_fill_skb_pool(sc);
	spin_unlock_irq(&sc->rx_pool_lock);

	if (RAL_RX_URB_FAILED == (sc->sc_flags & RAL_RX_URB_FAILED) &&
	    NULL != sc->sc_rx_skb) {
		goto try_again;
	}

	spin_lock_irq(&sc->rx_pool_lock);
	skb = ural_pull_skb(sc);
	spin_unlock_irq(&sc->rx_pool_lock);

	if (NULL == skb) {
		goto resched;
	}
	sc->sc_rx_skb = skb;
	usb_fill_bulk_urb(sc->sc_rx_urb, sc->udev, sc->sc_rx_pipe,
			  sc->sc_rx_skb->data, sc->sc_rx_buf_size,
			  ural_rxeof, sc);

try_again:
	if (0 != usb_submit_urb(sc->sc_rx_urb, GFP_ATOMIC)) {
		sc->sc_flags |= RAL_RX_URB_FAILED;
		goto resched;
	}
	sc->sc_flags &= ~RAL_RX_URB_FAILED;
	return;

resched:
	tasklet_schedule(&sc->rx_fixup_task);	
}

static int
ural_media_change(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	int error;

	error = ieee80211_media_change(netdev);
	if (error != ENETRESET) {
		DPRINTFN(5, ("%s: error %d", __func__, error));
		return error;
	}
	if ((netdev->flags & (IFF_UP | IFF_RUNNING)) ==
	    (IFF_UP | IFF_RUNNING)) {
		DPRINTFN(5, ("%s: reinit", __func__));
		usb_kill_urb(sc->sc_tx_urb);
		netif_stop_queue(netdev);
		schedule_work(&sc->sc_init);
	}
	return 0;
}

/*
 * This function is called periodically (every 200ms) during scanning to
 * switch from one channel to another.
 */
static void
ural_next_scan(void *arg)
{
	struct ural_softc *sc = arg;
	struct ieee80211com *ic = &sc->sc_ic;

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return;
	}
	if (ic->ic_state == IEEE80211_S_SCAN)
		DPRINTFN(10, ("next scan\n"));
		ieee80211_next_scan(ic);
}

static void
ural_work(void *arg)
{
	struct ural_softc *sc = arg;
	struct ieee80211com *ic = &sc->sc_ic;
	enum ieee80211_state ostate;
	struct sk_buff *skb;

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return;
	}
	
	ostate = ic->ic_state;

	DPRINTFN(10, ("ural_work\n"));
	switch (sc->sc_state) {
	case IEEE80211_S_INIT:
		if (ostate == IEEE80211_S_RUN) {
			/* abort TSF synchronization */
			ural_write(sc, RAL_TXRX_CSR19, 0);

			/* force tx led to stop blinking */
			ural_write(sc, RAL_MAC_CSR20, 0);
		}
		break;

	case IEEE80211_S_SCAN:
		ural_set_chan(sc, ic->ic_bss->ni_chan);
		schedule_delayed_work(&sc->sc_scan_ch, SCAN_JIFFIES);
		break;

	case IEEE80211_S_AUTH:
		ural_set_chan(sc, ic->ic_bss->ni_chan);
		break;

	case IEEE80211_S_ASSOC:
		ural_set_chan(sc, ic->ic_bss->ni_chan);
		break;

	case IEEE80211_S_RUN:
		ural_set_chan(sc, ic->ic_bss->ni_chan);

		if (ic->ic_opmode != IEEE80211_M_MONITOR)
			ural_set_bssid(sc, ic->ic_bss->ni_bssid);

		if (ic->ic_opmode == IEEE80211_M_HOSTAP ||
		    ic->ic_opmode == IEEE80211_M_IBSS) {
			skb = ieee80211_beacon_alloc(ic, ic->ic_bss,
						     &sc->sc_bo);
			if (skb == NULL) {
				err("could not allocate beacon\n");
				RAL_UNLOCK(sc);
				return;
			}

			if (ural_tx_bcn(sc, skb, ic->ic_bss) != 0) {
				err("could not send beacon\n");
				RAL_UNLOCK(sc);
				return;
			}
			/* XXX: freeing beacon skb? */
		}

		/* make tx led blink on tx (controlled by ASIC) */
		ural_write(sc, RAL_MAC_CSR20, 1);

		if (ic->ic_opmode != IEEE80211_M_MONITOR)
			ural_enable_tsf_sync(sc);
		break;
	}

	sc->sc_newstate(ic, sc->sc_state, -1);
}

static int
ural_newstate(struct ieee80211com *ic, enum ieee80211_state nstate, int arg)
{
	struct ural_softc *sc = netdev_priv(ic->ic_dev);

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return -ENETRESET;
	}
	
	flush_scheduled_work();
	cancel_delayed_work(&sc->sc_scan_ch);
	DPRINTFN(10, ("canceled delayed work\n"));
	sc->sc_state = nstate;
	schedule_work(&sc->sc_work);
	return 0;
}

/* quickly determine if a given rate is CCK or OFDM */
#define RAL_RATE_IS_OFDM(rate) ((rate) >= 12 && (rate) != 22)

#define RAL_ACK_SIZE	14	/* 10 + 4(FCS) */
#define RAL_CTS_SIZE	14	/* 10 + 4(FCS) */
#define RAL_SIFS	10

static void
ural_txeof(struct urb *urb, struct pt_regs *regs)
{
	struct ural_softc *sc = urb->context;
	struct net_device *netdev = sc->netdev;

	if (0 != urb->status) {
		switch (urb->status) {
			case -EPIPE:
				break;
			case -EINPROGRESS:
				return;
			default:
				break;
		}
		info("could not transmit buffer: %d\n", urb->status);
		sc->sc_devstats.tx_errors++;
		return;
	}

	sc->sc_devstats.tx_packets++;
	sc->sc_devstats.tx_bytes += sc->sc_tx_skb->len;

	dev_kfree_skb_irq(sc->sc_tx_skb);
	sc->sc_tx_skb = NULL;
	if (sc->sc_tx_ni != NULL && sc->sc_tx_ni != sc->sc_ic.ic_bss) {
		ieee80211_free_node(&sc->sc_ic, sc->sc_tx_ni);
		sc->sc_tx_ni = NULL;
	}
	DPRINTFN(10, ("tx done\n"));

	netdev->trans_start = jiffies;
	netif_wake_queue(netdev);
}

static void
ural_rxeof(struct urb *urb, struct pt_regs *regs)
{
#define IS_CTL(wh) \
        ((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) == IEEE80211_FC0_TYPE_CTL)
	struct ural_softc *sc;
	struct ieee80211com *ic;
	struct net_device *netdev;
	struct sk_buff *skb;
	struct ural_rx_desc *desc;
	struct ieee80211_frame *wh;
	struct ieee80211_node *ni;
	int len;
	int pkt_len;

	if (NULL == urb) {
		return;
	}
	sc = urb->context;
	if (NULL == sc) {
		return;
	}
	ic = &sc->sc_ic; 
	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return;
	}
	netdev = ic->ic_dev;
	if (!netif_device_present(netdev)) {
		return;
	}

	if (0 != urb->status) {
		switch (urb->status) {
			case 0:
				break;
			case -ENOENT:
			case -EPIPE:
				/* the urb is in unlink state */
				return;
			case -EPROTO:
			case -ETIMEDOUT:
			case -EILSEQ:
				goto skip;
			default:
				DPRINTFN(5, ("%s: Rx status %d", __func__,
					     urb->status));
				return;
		}
	}

	len = urb->actual_length;

	if (len < RAL_RX_DESC_SIZE) {
		err("xfer too short %d\n", len);
		sc->sc_devstats.rx_errors++;
		goto skip;
	}

	if (NULL == sc->sc_rx_skb) {
		DPRINTFN(10, ("%s: no skb! rescheduling...\n", __func__));
		goto resched;
	}

        skb = sc->sc_rx_skb;

	/* rx descriptor is located at the end */
	desc = (struct ural_rx_desc *) (skb->data + len - RAL_RX_DESC_SIZE);

	if (le32_to_cpu(desc->flags) & (RAL_RX_PHY_ERROR | RAL_RX_CRC_ERROR)) {
		/*
		 * This should not happen since we did not request to receive
		 * those frames when we filled RAL_TXRX_CSR2.
		 */
		DPRINTFN(5, ("PHY or CRC error\n"));
		sc->sc_devstats.rx_errors++;
		dev_kfree_skb(skb);
		goto next_rx;
	}

	pkt_len = (le32_to_cpu(desc->flags) >> 16) & 0xfff;
	DPRINTFN(5, ("%s: actual_length %u - pkt_len %d\n", __func__,
		     len, pkt_len));

	/* XXX: We should not need this */
	if (pkt_len > len) {
		DPRINTFN(5, ("BAD: pkt_len > actual_length!\n"));
		sc->sc_devstats.rx_errors++;
		dev_kfree_skb(skb);
		goto next_rx;
	}

	if (IEEE80211_M_MONITOR == ic->ic_opmode) {
		/* ieee80211_dump_pkt(skb->data, pkt_len, 4, desc->rssi); */
		/* XXX: That's for debugging ! */
		if (skb->truesize !=
			(SKB_DATA_ALIGN(sc->sc_rx_buf_size +
				       sizeof (wlan_ng_prism2_header)) +
			 sizeof (struct sk_buff))) {
			DPRINTFN(5, ("%s: Wrong skb size %u for monitor mode !\n", __func__, skb->truesize));
			sc->sc_devstats.rx_errors++;
			dev_kfree_skb(skb);
			goto next_rx;
		}
		if (IEEE80211_ACK_LEN > pkt_len) {
			DPRINTFN(10, ("%s: runt packet %d\n",
				      __func__, pkt_len));
			sc->sc_devstats.rx_errors++;
			dev_kfree_skb(skb);
			goto next_rx;
		}
		if (IEEE80211_MAX_LEN < pkt_len) {
			DPRINTFN(10, ("%s: packet to long %d\n",
				      __func__, pkt_len));
			sc->sc_devstats.rx_errors++;
			dev_kfree_skb(skb);
			goto next_rx;
		}
		ural_rx_capture(sc->netdev, desc, skb);
		goto next_rx;
	}

	if (IEEE80211_MIN_LEN > pkt_len) {
		DPRINTFN(5, ("%s: packet too short %d\n", __func__, pkt_len));
		sc->sc_devstats.tx_errors++;
		dev_kfree_skb(skb);
		goto next_rx;
	}

	skb_put(skb, pkt_len);
	skb->protocol = ETH_P_CONTROL;
	skb->dev = netdev;

        /* hardware appends FCS */
	skb_trim(skb, skb->len - IEEE80211_CRC_LEN);
	
	/* ieee80211_dump_pkt(skb->data, pkt_len, 4, desc->rssi); */
	wh = (struct ieee80211_frame *) skb->data;

	/*
	 * Locate the node for sender, track state, and then
	 * pass the (referenced) node up to the 802.11 layer
	 * for its use.  We are required to pass some node so
	 * we fall back to ic_bss when this frame is from an
	 * unknown sender.  The 802.11 layer knows this means the
	 * sender wasn't in the node table and acts accordingly.
	 * Note also that by convention we do not reference
	 * count ic_bss, only other nodes (ic_bss is never free'd).
	 */
	if (ic->ic_opmode != IEEE80211_M_STA) {
		if (IS_CTL(wh)) {
			ni = ieee80211_find_node(ic,
				((struct ieee80211_frame_min *) wh)->i_addr1);
		} else {
			ni = ieee80211_find_node(ic,
				((struct ieee80211_frame_min *) wh)->i_addr2);
		}
		if (ni == NULL) {
			ni = ic->ic_bss;
		}
	} else {
		ni = ic->ic_bss;
	}

	/* send the frame to the 802.11 layer */
	ieee80211_input(ic, skb, ni, desc->rssi, 0);

	if (NULL != ni && ni != ic->ic_bss) {
		/* node is no longer needed */
		ieee80211_free_node(ic, ni);
	}

next_rx:
	spin_lock(&sc->rx_pool_lock);
	sc->sc_rx_skb = ural_pull_skb(sc);
	spin_unlock(&sc->rx_pool_lock);
skip:
	/* setup a new transfer */
	if (NULL == sc->sc_rx_skb) {
		goto resched;
	}
	usb_fill_bulk_urb(sc->sc_rx_urb, sc->udev, sc->sc_rx_pipe,
			  sc->sc_rx_skb->data, sc->sc_rx_buf_size,
			  ural_rxeof, sc);
	if (0 != usb_submit_urb(sc->sc_rx_urb, GFP_ATOMIC)) {
		sc->sc_flags |= RAL_RX_URB_FAILED;
		goto resched;
	}
	sc->sc_flags &= ~RAL_RX_URB_FAILED;
	return;

resched:
	tasklet_schedule(&sc->rx_fixup_task);
#undef IS_CTL
}

/*
 * Return the expected ack rate for a frame transmitted at rate `rate'.
 * XXX: this should depend on the destination node basic rate set.
 */
static int
ural_ack_rate(int rate)
{
	switch (rate) {
	/* CCK rates */
	case 2:
		return 2;
	case 4:
	case 11:
	case 22:
		return 4;

	/* OFDM rates */
	case 12:
	case 18:
		return 12;
	case 24:
	case 36:
		return 24;
	case 48:
	case 72:
	case 96:
	case 108:
		return 48;
	}

	/* default to 1Mbps */
	return 2;
}

/*
 * Compute the duration (in us) needed to transmit `len' bytes at rate `rate'.
 * The function automatically determines the operating mode depending on the
 * given rate. `flags' indicates whether short preamble is in use or not.
 */
static u_int16_t
ural_txtime(int len, int rate, u_int32_t flags)
{
	u_int16_t txtime;
	int ceil, dbps;

	if (RAL_RATE_IS_OFDM(rate)) {
		/*
		 * OFDM TXTIME calculation.
		 * From IEEE Std 802.11a-1999, pp. 37.
		 */
		dbps = rate * 2; /* data bits per OFDM symbol */

		ceil = (16 + 8 * len + 6) / dbps;
		if ((16 + 8 * len + 6) % dbps != 0)
			ceil++;

		txtime = 16 + 4 + 4 * ceil + 6;
	} else {
		/*
		 * High Rate TXTIME calculation.
		 * From IEEE Std 802.11b-1999, pp. 28.
		 */
		ceil = (8 * len * 2) / rate;
		if ((8 * len * 2) % rate != 0)
			ceil++;

		if (rate != 2 && (flags & IEEE80211_F_SHPREAMBLE))
			txtime =  72 + 24 + ceil;
		else
			txtime = 144 + 48 + ceil;
	}

	return txtime;
}

static u_int8_t
ural_plcp_signal(int rate)
{
	switch (rate) {
	/* CCK rates (returned values are device-dependent) */
	case 2:		return 0x0;
	case 4:		return 0x1;
	case 11:	return 0x2;
	case 22:	return 0x3;

	/* OFDM rates (cf IEEE Std 802.11a-1999, pp. 14 Table 80) */
	case 12:	return 0xb;
	case 18:	return 0xf;
	case 24:	return 0xa;
	case 36:	return 0xe;
	case 48:	return 0x9;
	case 72:	return 0xd;
	case 96:	return 0x8;
	case 108:	return 0xc;

	/* unsupported rates (should not get there) */
	default:	return 0xff;
	}
}

static void
ural_setup_tx_desc(struct ural_softc *sc, struct ural_tx_desc *desc,
    u_int32_t flags, int len, int rate)
{
	struct ieee80211com *ic = &sc->sc_ic;
	u_int16_t plcp_length;
	int remainder;

	desc->flags = cpu_to_le32(flags);
	desc->flags |= cpu_to_le32(RAL_TX_NEWSEQ);
	desc->flags |= cpu_to_le32(len << 16);

	if (RAL_RATE_IS_OFDM(rate))
		desc->flags |= cpu_to_le32(RAL_TX_OFDM);

	desc->wme = cpu_to_le16(RAL_LOGCWMAX(5) | RAL_LOGCWMIN(3) |
			        RAL_AIFSN(2));
	desc->wme |= cpu_to_le16(RAL_IVOFFSET(
				sizeof (struct ieee80211_frame)));

	/*
	 * Fill PLCP fields.
	 */
	desc->plcp_service = 4;

	len += 4; /* account for FCS */
	if (RAL_RATE_IS_OFDM(rate)) {
		/*
		 * PLCP length field (LENGTH).
		 * From IEEE Std 802.11a-1999, pp. 14.
		 */
		plcp_length = len & 0xfff;
		desc->plcp_length = cpu_to_le16((plcp_length >> 6) << 8 |
		    (plcp_length & 0x3f));
	} else {
		/*
		 * Long PLCP LENGTH field.
		 * From IEEE Std 802.11b-1999, pp. 16.
		 */
		plcp_length = (8 * len * 2) / rate;
		remainder = (8 * len * 2) % rate;
		if (remainder != 0) {
			if (rate == 22 && (rate - remainder) / 16 != 0)
				desc->plcp_service |= RAL_PLCP_LENGEXT;
			plcp_length++;
		}
		desc->plcp_length = cpu_to_le16(plcp_length);
	}

	desc->plcp_signal = ural_plcp_signal(rate);
	if (rate != 2 && (ic->ic_flags & IEEE80211_F_SHPREAMBLE))
		desc->plcp_signal |= 0x08;

	desc->iv = 0;
	desc->eiv = 0;
}

#define RAL_TX_TIMEOUT	5000

static int
ural_tx_bcn(struct ural_softc *sc, struct sk_buff *skb0,
	   struct ieee80211_node *ni)
{
	struct ural_tx_desc *desc;
	u_int8_t cmd = 0;
	int error;
	u_int8_t *buf;
	int xferlen, rate;
	dma_addr_t dma;

	rate = IEEE80211_IS_CHAN_5GHZ(ni->ni_chan) ? 12 : 4;

	/* xfer length needs to be a multiple of two! */
	xferlen = (RAL_TX_DESC_SIZE + skb0->len + 1) & ~1;

	buf = usb_buffer_alloc(sc->udev, xferlen, GFP_ATOMIC, &dma);
	if (buf == NULL) {
		return ENOMEM;
	}

	if (0 != (error = usb_bulk_msg(sc->udev, sc->sc_tx_pipe, &cmd,
				       sizeof (cmd), NULL, RAL_TX_TIMEOUT))) {
		return error;
	}

	desc = (struct ural_tx_desc *) buf;

        memcpy(buf + RAL_TX_DESC_SIZE, skb0->data, skb0->len);
	ural_setup_tx_desc(sc, desc, RAL_TX_IFS_NEWBACKOFF | RAL_TX_TIMESTAMP,
	    skb0->len, rate);

	DPRINTFN(10, ("%s: sending beacon frame len=%u rate=%u xfer len=%u\n",
		      __func__, skb0->len, rate, xferlen));

	error = usb_bulk_msg(sc->udev, sc->sc_tx_pipe, buf, xferlen,
			     NULL, RAL_TX_TIMEOUT);
	/* XXX               ^^^^ (actual_length)
	 * Check for incomplete transmission */
	return error;
}

static int
ural_tx_mgt(struct ieee80211com *ic, struct sk_buff *skb0)
{
	struct net_device *netdev = ic->ic_dev;
	struct ural_softc *sc = netdev_priv(netdev);
	struct ural_tx_desc *desc;
	struct ieee80211_cb *cb = (struct ieee80211_cb *) skb0->cb;
	struct ieee80211_node *ni = cb->ni;
	struct ieee80211_frame *wh;
	u_int32_t flags = 0;
	u_int16_t dur;
	int error;
	int xferlen, rate;

	if (0 == (netdev->flags & IFF_RUNNING)) {
		DPRINTFN(10, ("%s: discard, invalid flags %x\n",
			      __func__, netdev->flags));
		error = -ENETDOWN;
		goto bad;
	}
	desc = (struct ural_tx_desc *) sc->sc_tx_buf;

	rate = IEEE80211_IS_CHAN_5GHZ(ni->ni_chan) ? 12 : 4;

	sc->sc_tx_skb = skb0;
	sc->sc_tx_ni = ni;

	wh = (struct ieee80211_frame *) skb0->data;

	if (!IEEE80211_IS_MULTICAST(wh->i_addr1)) {
		flags |= RAL_TX_ACK;

		dur = ural_txtime(RAL_ACK_SIZE, rate, ic->ic_flags) + RAL_SIFS;
		*(u_int16_t *)wh->i_dur = cpu_to_le16(dur);

		/* tell hardware to add timestamp for probe responses */
		if ((wh->i_fc[0] &
		     (IEEE80211_FC0_TYPE_MASK | IEEE80211_FC0_SUBTYPE_MASK)) ==
		    (IEEE80211_FC0_TYPE_MGT | IEEE80211_FC0_SUBTYPE_PROBE_RESP))
			flags |= RAL_TX_TIMESTAMP;
	}

	memcpy(sc->sc_tx_buf + RAL_TX_DESC_SIZE, skb0->data, skb0->len);
	ural_setup_tx_desc(sc, desc, flags, skb0->len, rate);

	/* xfer length needs to be a multiple of two! */
	xferlen = (RAL_TX_DESC_SIZE + skb0->len + 1) & ~1;

	DPRINTFN(10, ("%s: sending mgt frame len=%u rate=%u xfer len=%u\n",
		      __func__, skb0->len, rate, xferlen));

	sc->sc_tx_urb->transfer_flags |= URB_ZERO_PACKET;
	usb_fill_bulk_urb(sc->sc_tx_urb, sc->udev, sc->sc_tx_pipe,
			  sc->sc_tx_buf, xferlen, ural_txeof, sc);

	if (0 == (error = usb_submit_urb(sc->sc_tx_urb, GFP_ATOMIC))) {
		return 0;
	}
	/* fall thru... */
bad:
	if (ni && ni != ic->ic_bss) {
		ieee80211_free_node(ic, ni);
	}
	dev_kfree_skb_irq(skb0);
	return error;
}

static int
ural_tx_data(struct ural_softc *sc, struct sk_buff *skb0,
            struct ieee80211_node *ni)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ural_tx_desc *desc;
	struct ieee80211_rateset *rs;
	struct ieee80211_frame *wh;
	struct ieee80211_key *k;
	u_int32_t flags = 0;
	u_int16_t dur;
	int error;
	int xferlen, rate;

	wh = (struct ieee80211_frame *) skb0->data;

	/* XXX this should be reworked */
	if (ic->ic_fixed_rate != -1) {
		if (ic->ic_curmode != IEEE80211_MODE_AUTO) {
			rs = &ic->ic_sup_rates[ic->ic_curmode];
		} else {
			rs = &ic->ic_sup_rates[IEEE80211_MODE_11G];
		}
		rate = rs->rs_rates[ic->ic_fixed_rate];
	} else {
		rs = &ni->ni_rates;
		rate = rs->rs_rates[ni->ni_txrate];
	}
	rate &= IEEE80211_RATE_VAL;

	if (wh->i_fc[1] & IEEE80211_FC1_WEP) {
		k = ieee80211_crypto_encap(ic, ni, skb0);
		if (k == NULL) {
			return ENOBUFS;
		}

		/* packet header may have moved, reset our local pointer */
		wh = (struct ieee80211_frame *) skb0->data;
	}

	desc = (struct ural_tx_desc *) sc->sc_tx_buf;

	sc->sc_tx_skb = skb0;
	sc->sc_tx_ni = ni;

	if (!IEEE80211_IS_MULTICAST(wh->i_addr1)) {
		flags |= RAL_TX_ACK;
		flags |= RAL_TX_RETRY(7);

		dur = ural_txtime(RAL_ACK_SIZE, ural_ack_rate(rate),
		    ic->ic_flags) + RAL_SIFS;
		*(u_int16_t *)wh->i_dur = cpu_to_le16(dur);
	}

	memcpy(sc->sc_tx_buf + RAL_TX_DESC_SIZE, skb0->data, skb0->len);
	ural_setup_tx_desc(sc, desc, flags, skb0->len, rate);

	/* xfer length needs to be a multiple of two! */
	xferlen = (RAL_TX_DESC_SIZE + skb0->len + 1) & ~1;

	DPRINTFN(10, ("%s: sending data frame len=%u rate=%u xfer len=%u\n",
		      __func__, skb0->len, rate, xferlen));

	sc->sc_tx_urb->transfer_flags |= URB_ZERO_PACKET;
	usb_fill_bulk_urb(sc->sc_tx_urb, sc->udev, sc->sc_tx_pipe,
			  sc->sc_tx_buf, xferlen, ural_txeof, sc);

	if (0 != (error = usb_submit_urb(sc->sc_tx_urb, GFP_ATOMIC))) {
		return error;
	}

	return 0;
}

static int
ural_start_xmit(struct sk_buff *skb, struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	struct ieee80211_node *ni = NULL;
	int pktlen;
	int error = 0;

	if (0 == (netdev->flags & IFF_RUNNING)) {
		DPRINTFN(10, ("%s: discard, invalid flags %x\n",
			      __func__, netdev->flags));
		return -ENETDOWN;
	}

	netif_stop_queue(netdev);

	if (ic->ic_state != IEEE80211_S_RUN) {
		DPRINTFN(10, ("%s: discard, state %u\n",
			      __func__, ic->ic_state));
		goto bad;
	}

	skb = ieee80211_encap(ic, skb, &ni);
	if (NULL == skb) {
		DPRINTFN(10, ("%s: discard, encapsulation failure\n",
			      __func__));
		goto bad;
	}
	pktlen = skb->len; /* NB: don't reference skb below */
	if (0 == ural_tx_data(sc, skb, ni)) {
		sc->sc_devstats.tx_packets++;
		sc->sc_devstats.tx_bytes += pktlen;
		return 0;
	}

bad:
	if (NULL != ni && ni != ic->ic_bss) {
		ieee80211_free_node(ic, ni);
	}
	if (NULL != skb) {
		dev_kfree_skb(skb);
		sc->sc_tx_skb = NULL;
	}
	sc->sc_devstats.tx_errors++;
	netif_start_queue(netdev);
	return error; /* NB: return !0 only in a ``hard error condition'' */
}

static void
ural_tx_timeout(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);

	warn("%s: tx timeout\n", netdev->name);
	sc->sc_tx_urb->transfer_flags |= URB_ASYNC_UNLINK;
	usb_unlink_urb(sc->sc_tx_urb);
	sc->sc_devstats.tx_errors++;
	/* XXX: should we prefer _locked here? */
	ural_init(netdev);
}

static void
ural_rx_capture(struct net_device *netdev, struct ural_rx_desc *desc,
	       struct sk_buff *skb)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	wlan_ng_prism2_header *ph;
	int pkt_len;
	
	pkt_len = (le32_to_cpu(desc->flags) >> 16) & 0xfff;
	DPRINTFN(10,
	  ("%s: pkt_len %d - skb -> len %d truesize %d data_len %d\n",
	   __func__, pkt_len, skb->len, skb->truesize, skb->data_len));

	skb->protocol = ETH_P_CONTROL;
	skb->pkt_type = PACKET_OTHERHOST;
	skb_put(skb, pkt_len);

	ph = (wlan_ng_prism2_header *)
		skb_push(skb, sizeof (wlan_ng_prism2_header));
	memset(ph, 0, sizeof (wlan_ng_prism2_header));

	ph->msgcode = DIDmsg_lnxind_wlansniffrm;
	ph->msglen = sizeof (wlan_ng_prism2_header);
	strcpy(ph->devname, netdev->name);

	ph->hosttime.did = DIDmsg_lnxind_wlansniffrm_hosttime;
	ph->hosttime.status = 0;
	ph->hosttime.len = 4;
	ph->hosttime.data = jiffies;

	ph->mactime.did = DIDmsg_lnxind_wlansniffrm_mactime;
	ph->mactime.status = P80211ENUM_msgitem_status_no_value;
	ph->mactime.len = 4;
	ph->mactime.data = 0;
	
	ph->istx.did = DIDmsg_lnxind_wlansniffrm_istx;
	ph->istx.status = 0;
	ph->istx.len = 4;
	ph->istx.data = P80211ENUM_truth_false;

	ph->frmlen.did = DIDmsg_lnxind_wlansniffrm_frmlen;
	ph->frmlen.status = 0;
	ph->frmlen.len = 4;
	ph->frmlen.data = pkt_len;

	ph->channel.did = DIDmsg_lnxind_wlansniffrm_channel;
	ph->channel.status = 0;
	ph->channel.len = 4;
	ph->channel.data = ieee80211_mhz2ieee(ic->ic_ibss_chan->ic_freq, 0);

	ph->rssi.did = DIDmsg_lnxind_wlansniffrm_rssi;
	ph->rssi.status = P80211ENUM_msgitem_status_no_value;
	ph->rssi.len = 4;
	ph->rssi.data = 0;

	ph->signal.did = DIDmsg_lnxind_wlansniffrm_signal;
	ph->signal.status = 0;
	ph->signal.len = 4;
	ph->signal.data = desc->rssi;

	ph->rate.did = DIDmsg_lnxind_wlansniffrm_rate;
	ph->rate.status = 0;
	ph->rate.len = 4;
	ph->rate.data = IEEE80211_IS_CHAN_5GHZ(ic->ic_bss->ni_chan) ? 12 : 4;

	skb->dev = netdev;
	skb->mac.raw = skb->data;
	skb->ip_summed = CHECKSUM_NONE;
	skb->pkt_type = PACKET_OTHERHOST;
	skb->protocol = __constant_htons(0x0019);  /* ETH_P_80211_RAW */

	netif_rx(skb);
}

static void
ural_fill_skb_pool(struct ural_softc *sc)
{
	struct sk_buff *skb;
	int i;

	if (IEEE80211_M_MONITOR == sc->sc_ic.ic_opmode) {
		if (RAL_MONITOR_SKB != (sc->sc_flags & RAL_MONITOR_SKB)) {
			ural_free_skb_pool(sc);
			sc->sc_flags |= RAL_MONITOR_SKB;
		}
		for (i = 0; i < RAL_SKB_POOL_SIZE; i++) {
			if (NULL != sc->rx_skb_pool[i]) {
				continue;
			}
			skb = dev_alloc_skb(sc->sc_rx_buf_size +
					    sizeof (wlan_ng_prism2_header));
			if (NULL == skb) {
				return;
			}
			skb_reserve(skb, sizeof (wlan_ng_prism2_header));
			skb->dev = sc->netdev;
			sc->rx_skb_pool[i] = skb;
		}
		return;
	}

	for (i = 0; i < RAL_SKB_POOL_SIZE; i++) {
		if (NULL != sc->rx_skb_pool[i]) {
			continue;
		}
		if (NULL == (skb = dev_alloc_skb(sc->sc_rx_buf_size))) {
			return;
		}
		skb->dev = sc->netdev;
		sc->rx_skb_pool[i] = skb;
	}
	sc->sc_flags &= ~RAL_MONITOR_SKB;
}

static void
ural_free_skb_pool(struct ural_softc *sc)
{
	int i;

	for (i = 0; i < RAL_SKB_POOL_SIZE; i++) {
		if (NULL != sc->rx_skb_pool[i]) {
			dev_kfree_skb(sc->rx_skb_pool[i]);
			sc->rx_skb_pool[i] = NULL;
		}
	}
}

static struct sk_buff *
ural_pull_skb(struct ural_softc * sc)
{
	struct sk_buff *skb;
	int i;

	for (i = 0; i < RAL_SKB_POOL_SIZE; ++i) {
		if (NULL != sc->rx_skb_pool[i]) {
			skb = sc->rx_skb_pool[i];
			sc->rx_skb_pool[i] = NULL;
			return skb;
		}
	}
	return NULL;
}


/*
 * This function allows for fast channel switching in monitor mode (used by
 * net-mgmt/kismet). In IBSS mode, we must explicitly reset the interface to
 * generate a new beacon frame.
 */
static int
ural_reset(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;

	if (ic->ic_opmode != IEEE80211_M_MONITOR)
		return ural_init_locked(netdev);

	ural_set_chan(sc, ic->ic_ibss_chan);

	return 0;
}

static int
ural_ioctl(struct net_device *netdev, struct ifreq *ifr, int cmd)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	int error;

	RAL_LOCK(sc);

	error = ieee80211_ioctl(ic, ifr, cmd);
	DPRINTFN(10, ("%s: error %d", __func__, error));
#if 0
	if (error == -ENETRESET) {
		if ((netdev->flags & (IFF_UP | IFF_RUNNING)) ==
				(IFF_UP | IFF_RUNNING)) {
			error = ural_init_locked(netdev);
		} else {
			error = 0;
		}
	}
#endif
	RAL_UNLOCK(sc);

	return error;
}

static struct net_device_stats *
ural_getstats(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);

	return &sc->sc_devstats;
}

static struct iw_statistics *
ural_iw_getstats(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	struct iw_statistics *is = &sc->sc_iwstats;

	ieee80211_iw_getstats(ic, is);
	return is;
}

static void
ural_eeprom_read(struct ural_softc *sc, u_int16_t addr, void *buf, int len)
{
	int error;

	error = usb_control_msg(
			sc->udev, usb_rcvctrlpipe(sc->udev, 0),
			RAL_READ_EEPROM, 
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			0, addr, buf, len, HZ * 0.2);
	if (0 > error) {
		err("could not read EEPROM: %d\n", error);
	}
	if (4 == addr)
		printf("ural_eeprom_read: %x -> %s\n", addr, ether_sprintf(buf));
}

static u_int16_t
ural_read(struct ural_softc *sc, u_int16_t reg)
{
	u_int16_t val;

	if (0 > usb_control_msg(
			sc->udev, usb_rcvctrlpipe(sc->udev, 0), RAL_READ_MAC,
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE, 
			0, reg, &val, sizeof (u_int16_t), HZ * 0.2)) {
		err("could not read MAC register: %x\n", reg);
		return 0;
	}
	return le16_to_cpu(val);
}

static void
ural_read_multi(struct ural_softc *sc, u_int16_t reg, void *buf, int len)
{
	if (0 > usb_control_msg(
			sc->udev, usb_rcvctrlpipe(sc->udev, 0),
			RAL_READ_MULTI_MAC, 
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			0, reg, buf, len, HZ * 0.2)) {
		err("could not read MAC register: %x\n", reg);
	}
}

static void
ural_write(struct ural_softc *sc, u_int16_t reg, u_int16_t val)
{
	if (0 > usb_control_msg(
			sc->udev, usb_sndctrlpipe(sc->udev, 0), RAL_WRITE_MAC,
		        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			val, reg, NULL, 0, HZ * 0.2)) {
		err("could not write MAC register: %x\n", reg);
	}
}

static void
ural_write_multi(struct ural_softc *sc, u_int16_t reg, u_int8_t *buf,
		u_int16_t len)
{
	if (0 > usb_control_msg(
			sc->udev, usb_sndctrlpipe(sc->udev, 0),
			RAL_WRITE_MULTI_MAC,
			USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			0, reg, buf, len, HZ * 0.2)) {
		err("could not write MAC register: %x\n", reg);
	}
}

static void
ural_bbp_write(struct ural_softc *sc, u_int8_t reg, u_int8_t val)
{
	u_int16_t tmp;
	int ntries;

	for (ntries = 0; ntries < 5; ntries++) {
		if (!(ural_read(sc, RAL_PHY_CSR8) & RAL_BBP_BUSY))
			break;
	}
	if (ntries == 5) {
		err("could not write to BBP\n");
		return;
	}

	tmp = reg << 8 | val;
	ural_write(sc, RAL_PHY_CSR7, tmp);
}

static u_int8_t
ural_bbp_read(struct ural_softc *sc, u_int8_t reg)
{
	u_int16_t val;
	int ntries;

	val = RAL_BBP_WRITE | reg << 8;
	ural_write(sc, RAL_PHY_CSR7, val);

	for (ntries = 0; ntries < 50; ntries++) {
		if (!(ural_read(sc, RAL_PHY_CSR8) & RAL_BBP_BUSY))
			break;
	}
	if (ntries == 50) {
		err("could not read BBP\n");
		return 0;
	}

	return ural_read(sc, RAL_PHY_CSR7) & 0xff;
}

static void
ural_rf_write(struct ural_softc *sc, u_int8_t reg, u_int32_t val)
{
	u_int32_t tmp;
	int ntries;

	for (ntries = 0; ntries < 5; ntries++) {
		if (!(ural_read(sc, RAL_PHY_CSR10) & RAL_RF_LOBUSY))
			break;
	}
	if (ntries == 5) {
		err("could not write to RF\n");
		return;
	}

	tmp = RAL_RF_BUSY | RAL_RF_20BIT | (val & 0xfffff) << 2 | (reg & 0x3);
	ural_write(sc, RAL_PHY_CSR9,  tmp & 0xffff);
	ural_write(sc, RAL_PHY_CSR10, tmp >> 16);

	/* remember last written value in sc */
	sc->rf_regs[reg] = val;

	DPRINTFN(15, ("RF R[%u] <- 0x%05x\n", reg & 0x3, val & 0xfffff));
}

static void
ural_set_chan(struct ural_softc *sc, struct ieee80211_channel *c)
{
#define N(a)	(sizeof (a) / sizeof ((a)[0]))
	struct ieee80211com *ic = &sc->sc_ic;
	u_int8_t power, tmp;
	u_int i, chan;
	int fixed;

	chan = ieee80211_chan2ieee(ic, c);
	if (chan == 0 || chan == IEEE80211_CHAN_ANY) {
		return;
	}

	fixed = ic->ic_flags & IEEE80211_F_TXPOW_FIXED;
	if (fixed) {
		power = ic->ic_bss->ni_txpower;
	} else if (IEEE80211_IS_CHAN_2GHZ(c)) {
		power = (sc->txpow[chan - 1] > 31) ? 31 : sc->txpow[chan - 1];
	} else {
		power = 31;
	}

	DPRINTFN(7, ("setting channel to %u, %stxpower to %u\n",
		     chan, fixed ? "fixed " : "", power));

	switch (sc->rf_rev) {
	case RAL_RF_2522:
		ural_rf_write(sc, RAL_RF1, 0x00814);
		ural_rf_write(sc, RAL_RF2, ural_rf2522_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x00040);
		break;

	case RAL_RF_2523:
		ural_rf_write(sc, RAL_RF1, 0x08804);
		ural_rf_write(sc, RAL_RF2, ural_rf2523_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x38044);
		ural_rf_write(sc, RAL_RF4, (chan == 14) ? 0x00280 : 0x00286);
		break;

	case RAL_RF_2524:
		ural_rf_write(sc, RAL_RF1, 0x0c808);
		ural_rf_write(sc, RAL_RF2, ural_rf2524_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x00040);
		ural_rf_write(sc, RAL_RF4, (chan == 14) ? 0x00280 : 0x00286);
		break;

	case RAL_RF_2525:
		ural_rf_write(sc, RAL_RF1, 0x08808);
		ural_rf_write(sc, RAL_RF2, ural_rf2525_hi_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x18044);
		ural_rf_write(sc, RAL_RF4, (chan == 14) ? 0x00280 : 0x00286);

		ural_rf_write(sc, RAL_RF1, 0x08808);
		ural_rf_write(sc, RAL_RF2, ural_rf2525_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x18044);
		ural_rf_write(sc, RAL_RF4, (chan == 14) ? 0x00280 : 0x00286);
		break;

	case RAL_RF_2525E:
		ural_rf_write(sc, RAL_RF1, 0x08808);
		ural_rf_write(sc, RAL_RF2, ural_rf2525e_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x18044);
		ural_rf_write(sc, RAL_RF4, (chan == 14) ? 0x00286 : 0x00282);
		break;

	case RAL_RF_2526:
		ural_rf_write(sc, RAL_RF2, ural_rf2526_hi_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF4, (chan & 1) ? 0x00386 : 0x00381);
		ural_rf_write(sc, RAL_RF1, 0x08804);

		ural_rf_write(sc, RAL_RF2, ural_rf2526_r2[chan - 1]);
		ural_rf_write(sc, RAL_RF3, power << 7 | 0x18044);
		ural_rf_write(sc, RAL_RF4, (chan & 1) ? 0x00386 : 0x00381);
		break;

	/* dual-band RF */
	case RAL_RF_5222:
		for (i = 0; i < N(ural_rf5222); i++)
			if (ural_rf5222[i].chan == chan)
				break;

		if (i < N(ural_rf5222)) {
			ural_rf_write(sc, RAL_RF1, ural_rf5222[i].r1);
			ural_rf_write(sc, RAL_RF2, ural_rf5222[i].r2);
			ural_rf_write(sc, RAL_RF3, power << 7 | 0x00040);
			ural_rf_write(sc, RAL_RF4, ural_rf5222[i].r4);
		}
		break;
	}

	if (ic->ic_opmode != IEEE80211_M_MONITOR &&
	    ic->ic_state != IEEE80211_S_SCAN) {
		/* set Japan filter bit for channel 14 */
		tmp = ural_bbp_read(sc, 70);

		tmp &= ~RAL_JAPAN_FILTER;
		if (chan == 14)
			tmp |= RAL_JAPAN_FILTER;

		ural_bbp_write(sc, 70, tmp);

		/* clear CRC errors */
		ural_read(sc, RAL_STA_CSR0);
	}

	ic->ic_bss->ni_txpower = power;
#undef N
}

#if 0
/*
 * Disable RF auto-tuning.
 */
static void
ural_disable_rf_tune(struct ural_softc *sc)
{
	u_int32_t tmp;

	if (sc->rf_rev != RAL_RF_2523) {
		tmp = sc->rf_regs[RAL_RF1] & ~RAL_RF1_AUTOTUNE;
		ural_rf_write(sc, RAL_RF1, tmp);
	}

	tmp = sc->rf_regs[RAL_RF3] & ~RAL_RF3_AUTOTUNE;
	ural_rf_write(sc, RAL_RF3, tmp);

	DPRINTFN(2, ("disabling RF autotune\n"));
}
#endif

/*
 * Refer to IEEE Std 802.11-1999 pp. 123 for more information on TSF
 * synchronization.
 */
static void
ural_enable_tsf_sync(struct ural_softc *sc)
{
	struct ieee80211com *ic = &sc->sc_ic;
	u_int16_t logcwmin, preload, tmp;

	/* first, disable TSF synchronization */
	ural_write(sc, RAL_TXRX_CSR19, 0);

	tmp = (16 * ic->ic_bss->ni_intval) << 4;
	ural_write(sc, RAL_TXRX_CSR18, tmp);

	logcwmin = (ic->ic_opmode == IEEE80211_M_IBSS) ? 2 : 0;
	preload = (ic->ic_opmode == IEEE80211_M_IBSS) ? 320 : 6;
	tmp = logcwmin << 12 | preload;
	ural_write(sc, RAL_TXRX_CSR20, tmp);

	/* finally, enable TSF synchronization */
	if (ic->ic_opmode == IEEE80211_M_STA)
		ural_write(sc, RAL_TXRX_CSR19, RAL_TSF_SYNC_BSS |
		    RAL_TSF_AUTOCOUNT | RAL_BCN_RELOAD);
	else
		ural_write(sc, RAL_TXRX_CSR19, RAL_TSF_SYNC_IBSS |
		    RAL_TSF_AUTOCOUNT | RAL_BCN_RELOAD | RAL_GENERATE_BEACON);

	DPRINTF(("enabling TSF synchronization\n"));
}

static void
ural_set_bssid(struct ural_softc *sc, u_int8_t *bssid)
{
	u_int16_t tmp;

	tmp = bssid[0] | bssid[1] << 8;
	ural_write(sc, RAL_MAC_CSR5, tmp);

	tmp = bssid[2] | bssid[3] << 8;
	ural_write(sc, RAL_MAC_CSR6, tmp);

	tmp = bssid[4] | bssid[5] << 8;
	ural_write(sc, RAL_MAC_CSR7, tmp);

	DPRINTF(("setting BSSID to %s\n", ether_sprintf(bssid)));
}

static int
ural_set_macaddr(struct net_device *netdev, void *addr)
{
	struct ural_softc *sc = netdev_priv(netdev);
	int retval;

	RAL_LOCK(sc);
	retval = ural_set_macaddr(netdev, addr);
	RAL_UNLOCK(sc);
	return retval;
}

static int
ural_set_macaddr_locked(struct net_device *netdev, void *addr)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	struct sockaddr *mac = addr;
	u_int16_t tmp;

	DPRINTF(("ural_set_macaddr: %s\n", ether_sprintf(mac->sa_data)));
	IEEE80211_ADDR_COPY(ic->ic_myaddr, mac->sa_data);
	IEEE80211_ADDR_COPY(netdev->dev_addr, mac->sa_data);

	tmp = mac->sa_data[0] | mac->sa_data[1] << 8;
	ural_write(sc, RAL_MAC_CSR2, tmp);

	tmp = mac->sa_data[2] | mac->sa_data[3] << 8;
	ural_write(sc, RAL_MAC_CSR3, tmp);

	tmp = mac->sa_data[4] | mac->sa_data[5] << 8;
	ural_write(sc, RAL_MAC_CSR4, tmp);

	DPRINTF(("setting MAC address to %s\n", ether_sprintf(mac->sa_data)));
	return 0;
}

static void ural_set_multicast_list(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);

	schedule_work(&sc->sc_mode_init);
}

static void
ural_mode_init(void *arg)
{
	struct ural_softc *sc = arg;
	struct net_device *netdev = sc->netdev;
	u_int16_t tmp;

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return;
	}
	
	netif_stop_queue(netdev);
	tmp = ural_read(sc, RAL_TXRX_CSR2);

        tmp &= ~RAL_DROP_NOT_TO_ME;
	if (!(netdev->flags & IFF_PROMISC))
		tmp |= RAL_DROP_NOT_TO_ME;
	ural_write(sc, RAL_TXRX_CSR2, tmp);
        DPRINTF(("%s promiscuous mode\n", (netdev->flags & IFF_PROMISC) ?
				            "entering" : "leaving"));
	netif_wake_queue(netdev);
}

static const char *
ural_get_rf(int rev)
{
	switch (rev) {
	case RAL_RF_2522:	return "RT2522";
	case RAL_RF_2523:	return "RT2523";
	case RAL_RF_2524:	return "RT2524";
	case RAL_RF_2525:	return "RT2525";
	case RAL_RF_2525E:	return "RT2525e";
	case RAL_RF_2526:	return "RT2526";
	case RAL_RF_5222:	return "RT5222";
	default:		return "unknown";
	}
}

static void
ural_read_eeprom(struct ural_softc *sc)
{
	struct ieee80211com *ic = &sc->sc_ic;
	u_int16_t val;

	ural_eeprom_read(sc, RAL_EEPROM_CONFIG0, &val, 2);
	val = le16_to_cpu(val);
	sc->rf_rev =   (val >> 11) & 0x7;
	sc->hw_radio = (val >> 10) & 0x1;
	sc->led_mode = (val >> 6)  & 0x7;
	sc->rx_ant =   (val >> 4)  & 0x3;
	sc->tx_ant =   (val >> 2)  & 0x3;
	sc->nb_ant =   val & 0x3;

	/* read MAC address */
	ural_eeprom_read(sc, RAL_EEPROM_ADDRESS, ic->ic_myaddr, 6);

	/* read default values for BBP registers */
	ural_eeprom_read(sc, RAL_EEPROM_BBP_BASE, sc->bbp_prom, 2 * 16);

	/* read Tx power for all b/g channels */
	ural_eeprom_read(sc, RAL_EEPROM_TXPOWER, sc->txpow, 14);
}

static int
ural_bbp_init(struct ural_softc *sc)
{
#define N(a)	(sizeof (a) / sizeof ((a)[0]))
	int i, ntries;

	/* wait for BBP to be ready */
	for (ntries = 0; ntries < 100; ntries++) {
		if (ural_bbp_read(sc, RAL_BBP_VERSION) != 0)
			break;
		mdelay(1);
	}
	if (ntries == 100) {
		warn("%s: timeout waiting for BBP\n", sc->netdev->name);
		return EIO;
	}

	/* initialize BBP registers to default values */
	for (i = 0; i < N(ural_def_bbp); i++)
		ural_bbp_write(sc, ural_def_bbp[i].reg, ural_def_bbp[i].val);

#if 0
	/* initialize BBP registers to values stored in EEPROM */
	for (i = 0; i < 16; i++) {
		if (sc->bbp_prom[i].reg == 0xff)
			continue;
		ural_bbp_write(sc, sc->bbp_prom[i].reg, sc->bbp_prom[i].val);
	}
#endif

	return 0;
#undef N
}

static void
ural_set_txantenna(struct ural_softc *sc, int antenna)
{
	u_int16_t tmp;
	u_int8_t tx;

	tx = ural_bbp_read(sc, RAL_BBP_TX) & ~RAL_BBP_ANTMASK;
	if (antenna == 1)
		tx |= RAL_BBP_ANTA;
	else if (antenna == 2)
		tx |= RAL_BBP_ANTB;
	else
		tx |= RAL_BBP_DIVERSITY;

	/* need to force I/Q flip for RF 2525e, 2526 and 5222 */
	if (sc->rf_rev == RAL_RF_2525E || sc->rf_rev == RAL_RF_2526 ||
	    sc->rf_rev == RAL_RF_5222)
		tx |= RAL_BBP_FLIPIQ;

	ural_bbp_write(sc, RAL_BBP_TX, tx);

	/* update values in PHY_CSR5 and PHY_CSR6 */
	tmp = ural_read(sc, RAL_PHY_CSR5) & ~0x7;
	ural_write(sc, RAL_PHY_CSR5, tmp | (tx & 0x7));

	tmp = ural_read(sc, RAL_PHY_CSR6) & ~0x7;
	ural_write(sc, RAL_PHY_CSR6, tmp | (tx & 0x7));
}

static void
ural_set_rxantenna(struct ural_softc *sc, int antenna)
{
	u_int8_t rx;

	rx = ural_bbp_read(sc, RAL_BBP_RX) & ~RAL_BBP_ANTMASK;
	if (antenna == 1)
		rx |= RAL_BBP_ANTA;
	else if (antenna == 2)
		rx |= RAL_BBP_ANTB;
	else
		rx |= RAL_BBP_DIVERSITY;

	/* need to force no I/Q flip for RF 2525e and 2526 */
	if (sc->rf_rev == RAL_RF_2525E || sc->rf_rev == RAL_RF_2526)
		rx &= ~RAL_BBP_FLIPIQ;

	ural_bbp_write(sc, RAL_BBP_RX, rx);
}

static void
ural_init_wq(void *arg) {
	struct net_device *netdev = arg;

	(void) ural_init_locked(netdev);
}

static int
ural_init(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	int retval;

	RAL_LOCK(sc);
	retval = ural_init_locked(netdev);
	RAL_UNLOCK(sc);
	return retval;
}

static int
ural_init_locked(struct net_device *netdev)
{
#define N(a)	(sizeof (a) / sizeof ((a)[0]))
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;
	struct ieee80211_key *wk;
	struct sockaddr mac;
	u_int16_t sta[11], tmp;
	int error = 0;
	int i, ntries;

	ural_stop_locked(netdev);

        /* initialize MAC registers to default values */
	for (i = 0; i < N(ural_def_mac); i++)
		ural_write(sc, ural_def_mac[i].reg, ural_def_mac[i].val);
	
	/* wait for BBP and RF to wake up (this can take a long time!) */
	for (ntries = 0; ntries < 100; ntries++) {
		tmp = ural_read(sc, RAL_MAC_CSR17);
		if ((tmp & (RAL_BBP_AWAKE | RAL_RF_AWAKE)) ==
		    (RAL_BBP_AWAKE | RAL_RF_AWAKE))
			break;
		mdelay(1);
	}
	if (ntries == 100) {
		err("timeout waiting for BBP/RF to wakeup\n");
		error = -ETIMEDOUT;
		goto fail;
	}

	/* we're ready! */
	ural_write(sc, RAL_MAC_CSR1, RAL_HOST_READY);

	/* set supported basic rates (1, 2, 6, 12, 24) */
	ural_write(sc, RAL_TXRX_CSR11, 0x153);

	if (0 != (error = ural_bbp_init(sc))) {
		goto fail;
	}

	/* set default BSS channel */
	ic->ic_bss->ni_chan = ic->ic_ibss_chan;
	ural_set_chan(sc, ic->ic_bss->ni_chan);

	/* clear statistic registers (STA_CSR0 to STA_CSR10) */
	ural_read_multi(sc, RAL_STA_CSR0, sta, sizeof sta);

	/* set default sensitivity */
	ural_bbp_write(sc, 17, 0x48);

	ural_set_txantenna(sc, sc->tx_ant);
	ural_set_rxantenna(sc, sc->rx_ant);

	IEEE80211_ADDR_COPY(mac.sa_data, ic->ic_myaddr);
	ural_set_macaddr_locked(netdev, &mac);

	/*
	 * Copy WEP keys into adapter's memory (SEC_CSR0 to SEC_CSR31).
	 */
	for (i = 0; i < IEEE80211_WEP_NKID; i++) {
		wk = &ic->ic_crypto.cs_nw_keys[i];
		ural_write_multi(sc, wk->wk_keyix * IEEE80211_KEYBUF_SIZE +
		    RAL_SEC_CSR0, wk->wk_key, IEEE80211_KEYBUF_SIZE);
	}

	/*
	 * Open Tx and Rx USB bulk pipes.
	 */
	sc->sc_tx_pipe = usb_sndbulkpipe(sc->udev, sc->sc_tx_no);
	sc->sc_rx_pipe = usb_rcvbulkpipe(sc->udev, sc->sc_rx_no);

	/*
	 * Allocate Tx and Rx URB.
	 */
	if (NULL == (sc->sc_tx_urb = usb_alloc_urb(0, GFP_KERNEL))) {
		err("could not allocate Tx URB\n");
		error = -ENOMEM;
		goto fail;
	}
	if (NULL == (sc->sc_rx_urb = usb_alloc_urb(0, GFP_KERNEL))) {
		err("could not allocate Rx URB\n");
		error = -ENOMEM;
		goto fail;
	}
	sc->sc_flags &= ~RAL_RX_URB_FAILED;

	sc->sc_rx_buf_size = RAL_RX_DESC_SIZE + IEEE80211_MAX_LEN;

	/* Allocate Tx buffer */
	if (NULL == (sc->sc_tx_buf = usb_buffer_alloc(
			sc->udev, RAL_TX_DESC_SIZE + IEEE80211_MAX_LEN,
			GFP_KERNEL, &sc->sc_tx_urb->transfer_dma))) {
		err("could not allocate Tx buffer\n");
		error = -ENOMEM;
		goto fail;
	}

	spin_lock_irq(&sc->rx_pool_lock);

	/* Allocate Rx pool */
	ural_fill_skb_pool(sc);	

	/*
	 * Start up the receive pipe.
	 */
	if (NULL == sc->sc_rx_skb) {
		if (NULL == (sc->sc_rx_skb = ural_pull_skb(sc))) {
			error = -ENOMEM;
			goto fail;
		}
	}

	spin_unlock_irq(&sc->rx_pool_lock);

	usb_fill_bulk_urb(sc->sc_rx_urb, sc->udev, sc->sc_rx_pipe,
			  sc->sc_rx_skb->data, sc->sc_rx_buf_size,
			  ural_rxeof, sc);
	if (0 != (error = usb_submit_urb(sc->sc_rx_urb, GFP_KERNEL))) {
		err("could not start receive %d", i);
		goto fail;
	}

	/* kick Rx */
	tmp = RAL_DROP_PHY_ERROR | RAL_DROP_CRC_ERROR;
	if (ic->ic_opmode != IEEE80211_M_MONITOR) {
		tmp |= RAL_DROP_CTL | RAL_DROP_VERSION_ERROR;
		if (ic->ic_opmode != IEEE80211_M_HOSTAP) {
			tmp |= RAL_DROP_TODS;
		}
		if (!(netdev->flags & IFF_PROMISC)) {
			tmp &= ~RAL_DROP_NOT_TO_ME;
		}
	}
	ural_write(sc, RAL_TXRX_CSR2, tmp);

	/* XXX: Check if right */
	netif_start_queue(netdev);

	netdev->flags |= IFF_RUNNING;

	if (ic->ic_opmode == IEEE80211_M_MONITOR ||
	    ic->ic_roaming == IEEE80211_ROAMING_MANUAL) {
		ieee80211_new_state(ic, IEEE80211_S_RUN, -1);
	} else {
		ieee80211_new_state(ic, IEEE80211_S_SCAN, -1);
	}

	return 0;

fail:	ural_stop_locked(netdev);
	return error;
#undef N
}

static int
ural_stop(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	int retval;

	RAL_LOCK(sc);
	retval = ural_stop_locked(netdev);
	RAL_UNLOCK(sc);
	return retval;
}

static int
ural_stop_locked(struct net_device *netdev)
{
	struct ural_softc *sc = netdev_priv(netdev);
	struct ieee80211com *ic = &sc->sc_ic;

	netif_stop_queue(netdev);
	netdev->flags &= ~IFF_RUNNING;

	if (RAL_UNPLUG == (sc->sc_flags & RAL_UNPLUG)) {
		return 0;
	}

	ieee80211_new_state(ic, IEEE80211_S_INIT, -1);

	/* disable Rx */
	ural_write(sc, RAL_TXRX_CSR2, RAL_DISABLE_RX);

	/* reset ASIC and BBP (but won't reset MAC registers!) */
	ural_write(sc, RAL_MAC_CSR1, RAL_RESET_ASIC | RAL_RESET_BBP);
	ural_write(sc, RAL_MAC_CSR1, 0);

	if (sc->sc_rx_urb != NULL) {
		usb_kill_urb(sc->sc_rx_urb);
		usb_free_urb(sc->sc_rx_urb);
		sc->sc_rx_urb = NULL;
	}
	spin_lock_irq(&sc->rx_pool_lock);
	ural_free_skb_pool(sc);
	spin_lock_irq(&sc->rx_pool_lock);

	if (sc->sc_tx_urb != NULL) {
		usb_kill_urb(sc->sc_tx_urb);
		usb_free_urb(sc->sc_tx_urb);
		sc->sc_tx_urb = NULL;
	}
	if (sc->sc_tx_skb != NULL) {
		dev_kfree_skb(sc->sc_tx_skb);
		sc->sc_tx_skb = NULL;
	}
	if (sc->sc_tx_ni != NULL && sc->sc_tx_ni != sc->sc_ic.ic_bss) {
		ieee80211_free_node(&sc->sc_ic, sc->sc_tx_ni);
		sc->sc_tx_ni = NULL;
	}

	sc->sc_rx_pipe = 0;
	sc->sc_tx_pipe = 0;

	return 0;
}

static char const driver_name[] = DRIVER_NAME;

static struct usb_driver ural_driver = {
	.owner =	THIS_MODULE,
	.name =		driver_name,
	.probe =	ural_probe,
	.disconnect =	ural_disconnect,
	.id_table =	ural_table
};

static int __init ural_module_init(void)
{
	info(DRIVER_DESC " " DRIVER_VERSION);
	return usb_register(&ural_driver);
}

static void __exit ural_module_exit(void)
{
	usb_deregister(&ural_driver);
}


module_init(ural_module_init);
module_exit(ural_module_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("BSD");

/* vim: noet sw=8
 */
