#ifndef __ORINOCO_KCOMPAT_H
#define __ORINOCO_KCOMPAT_H

#include <linux/version.h>

/********************************************************************/
/* General backward compatibility                                   */
/********************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 19)
#define BUG_ON(condition) do { if ((condition)!=0) BUG(); } while(0)
#endif

#ifndef WARN_ON
#define WARN_ON(condition) BUG_ON(condition)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define in_atomic in_interrupt
#endif

#ifndef __devexit_p
#define __devexit_p(x) x
#endif

#ifndef ALIGN
#define ALIGN(x,a) (((x)+(a)-1)&~((a)-1))
#endif

/* Work queue / task queue compatibility */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,41)
#include <linux/tqueue.h>
#define work_struct tq_struct
#define INIT_WORK(x,y,z) INIT_TQUEUE(x,y,z)
#define schedule_work(x) schedule_task(x)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 25)
#define module_param(name,type,perm) MODULE_PARM(name,"i")
#endif

#ifndef _set_ver
#define _set_ver(sym) sym
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
#include <linux/pci.h>
#undef pci_save_state
#define pci_save_state(pdev) \
	_set_ver(pci_save_state)(pdev, \
		((struct orinoco_pci_card *)priv->card)->pci_state)

#undef pci_restore_state
#define pci_restore_state(pdev) \
	_set_ver(pci_restore_state)(pdev, \
		((struct orinoco_pci_card *)priv->card)->pci_state)
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10) */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
typedef u32 pm_message_t;
#define PCI_D3hot 0
#endif

/* Interrupt handler backwards compatibility */
#ifndef IRQ_NONE
#define IRQ_NONE
#define IRQ_HANDLED
typedef void irqreturn_t;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,29))
static inline void msleep(unsigned int msecs)
{
	current->state = TASK_UNINTERRUPTIBLE;
	schedule_timeout(msecs * HZ / 1000);
}

static inline void ssleep(unsigned int seconds)
{
	msleep(seconds * 1000);
}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20))
static inline void list_move(struct list_head *list, struct list_head *head)
{
        __list_del(list->prev, list->next);
        list_add(list, head);
}

static inline void list_move_tail(struct list_head *list,
				  struct list_head *head)
{
        __list_del(list->prev, list->next);
        list_add_tail(list, head);
}
#endif

#ifndef HAVE_NETDEV_PRIV
#define netdev_priv(dev)	((dev)->priv)
#endif /* ! HAVE_NETDEV_PRIV */

/********************************************************************/
/* I/O backward compatibility                                   */
/********************************************************************/

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))

/* IA-64 has its own PIO_OFFSET, PIO_MASK and PIO_RESERVED */
#ifdef CONFIG_IA64
#error "2.4 kernels for IA-64 are not supported"
#endif

#define PIO_OFFSET	0x10000UL
#define PIO_MASK	0x0ffffUL
#define PIO_RESERVED	0x40000UL

#define VERIFY_PIO(port) BUG_ON((port & ~PIO_MASK) != PIO_OFFSET)

#define IO_COND(addr, is_pio, is_mmio) do {			\
	unsigned long port = (unsigned long)addr;		\
	if (port < PIO_RESERVED) {				\
		VERIFY_PIO(port);				\
		port &= PIO_MASK;				\
		is_pio;						\
	} else {						\
		is_mmio;					\
	}							\
} while (0)

static inline void mmio_insw(void __iomem *addr, u16 *dst, int count)
{
	while (--count >= 0) {
		u16 data = __raw_readw(addr);
		*dst = data;
		dst++;
	}
}
static inline void mmio_outsw(void __iomem *addr, const u16 *src, int count)
{
	while (--count >= 0) {
		__raw_writew(*src, addr);
		src++;
	}
}
static inline unsigned int fastcall ioread16(void __iomem *addr)
{
	IO_COND(addr, return inw(port), return readw(addr));
}
static inline void fastcall iowrite16(u16 val, void __iomem *addr)
{
	IO_COND(addr, outw(val,port), writew(val, addr));
}
static inline void fastcall ioread16_rep(void __iomem *addr, void *dst,
					 unsigned long count)
{
	IO_COND(addr, insw(port,dst,count), mmio_insw(addr, dst, count));
}
static inline void fastcall iowrite16_rep(void __iomem *addr,
					  const void *src,
					  unsigned long count)
{
	IO_COND(addr, outsw(port, src, count), mmio_outsw(addr, src, count));
}

/* Create a virtual mapping cookie for an IO port range */
static inline void __iomem *ioport_map(unsigned long port, unsigned int nr)
{
	if (port > PIO_MASK)
		return NULL;
	return (void __iomem *) (unsigned long) (port + PIO_OFFSET);
}

static inline void ioport_unmap(void __iomem *addr)
{
	/* Nothing to do */
}

/* Create a virtual mapping cookie for a PCI BAR (memory or IO) */
static inline void __iomem *pci_iomap(struct pci_dev *dev, int bar,
				      unsigned long maxlen)
{
	unsigned long start = pci_resource_start(dev, bar);
	unsigned long len = pci_resource_len(dev, bar);
	unsigned long flags = pci_resource_flags(dev, bar);

	if (!len || !start)
		return NULL;
	if (maxlen && len > maxlen)
		len = maxlen;
	if (flags & IORESOURCE_IO)
		return ioport_map(start, len);
	if (flags & IORESOURCE_MEM) {
		if (flags & IORESOURCE_CACHEABLE)
			return ioremap(start, len);
		return ioremap_nocache(start, len);
	}
	/* What? */
	return NULL;
}

static inline void pci_iounmap(struct pci_dev *dev, void __iomem * addr)
{
	IO_COND(addr, /* nothing */, iounmap(addr));
}
#endif

/********************************************************************/
/* Network backward compatibility                                   */
/********************************************************************/

/* Device type while in monitor mode, appeared in Linux 2.4.6 */
#ifndef ARPHRD_IEEE80211
#define ARPHRD_IEEE80211 801
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#ifndef HAVE_FREE_NETDEV
#define free_netdev(dev) kfree(dev)
#endif
#endif

#ifndef SET_NETDEV_DEV
#define SET_NETDEV_DEV(x,y)
#endif


/********************************************************************/
/* Wireless extensions backward compatibility                       */
/********************************************************************/

#if defined(WIRELESS_EXT) && (WIRELESS_EXT <= 12)
/* Taken verbatim from iw_handler.h */
struct iw_request_info
{
	u16 cmd;	/* Wireless Extension command */
	u16 flags;	/* More to come ;-) */
};

#ifndef SIOCIWFIRSTPRIV
#define SIOCIWFIRSTPRIV	0x8BE0
#endif

#endif	/* WIRELESS_EXT <= 12 */

#ifndef IW_MODE_MONITOR
#define IW_MODE_MONITOR 6
#endif


/********************************************************************/
/* USB backward compatibility                                       */
/********************************************************************/

#ifdef __LINUX_USB_H

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define return_probe(priv_data, retval) return priv_data
#define usb_set_intfdata(intf, priv_data)
#define usb_get_desc(iface) (iface)

#undef usb_match_id
#define usb_match_id(intf, table) _set_ver(usb_match_id)(udev, intf, table)

#undef usb_submit_urb
#define usb_submit_urb(urb, mem_flags) _set_ver(usb_submit_urb)(urb)

#undef usb_alloc_urb
#define usb_alloc_urb(iso_packets, mem_flags) _set_ver(usb_alloc_urb)(iso_packets)

#define ezusb_bulk_in_callback(urb, pt_regs) ezusb_bulk_in_callback(urb)
#define ezusb_request_out_callback(urb, pt_regs) ezusb_request_out_callback(urb)
#define prism_usb_bulkout_cb(urb, pt_regs) prism_usb_bulkout_cb(urb)

#define URB_ASYNC_UNLINK USB_ASYNC_UNLINK
#define fw_dev_param "usb:" DRIVER_NAME
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0) */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
#define usb_fill_bulk_urb(a,b,c,d,e,f,g) FILL_BULK_URB(a,b,c,d,e,f,g)
#endif

#endif /* __LINUX_USB_H */


/********************************************************************/
/* PCI backward compatibility                                       */
/********************************************************************/

#ifdef LINUX_PCI_H
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 22)
static inline char *pci_name(struct pci_dev *pdev)
{
	        return pdev->slot_name;
}
#endif
#endif /* LINUX_PCI_H */


/********************************************************************/
/* PCMCIA backward compatibility                                    */
/********************************************************************/

#ifdef _LINUX_DS_H

/* pcmcia_core module from pcmcia-cs doesn't export pcmcia_* functions
 * and requires using CardServices instead.  Taken from hostap driver. */
#ifdef __IN_PCMCIA_PACKAGE__
#undef pcmcia_access_configuration_register
#define pcmcia_access_configuration_register(handle, reg) \
	CardServices(AccessConfigurationRegister, handle, reg)

#undef pcmcia_register_client
#define pcmcia_register_client(handle, reg) \
	CardServices(RegisterClient, handle, reg)

#undef pcmcia_deregister_client
#define pcmcia_deregister_client(handle) \
	CardServices(DeregisterClient, handle)

#undef pcmcia_get_first_tuple
#define pcmcia_get_first_tuple(handle, tuple) \
	CardServices(GetFirstTuple, handle, tuple)

#undef pcmcia_get_next_tuple
#define pcmcia_get_next_tuple(handle, tuple) \
	CardServices(GetNextTuple, handle, tuple)

#undef pcmcia_get_tuple_data
#define pcmcia_get_tuple_data(handle, tuple) \
	CardServices(GetTupleData, handle, tuple)

#undef pcmcia_parse_tuple
#define pcmcia_parse_tuple(handle, tuple, parse) \
	CardServices(ParseTuple, handle, tuple, parse)

#undef pcmcia_get_configuration_info
#define pcmcia_get_configuration_info(handle, config) \
	CardServices(GetConfigurationInfo, handle, config)

#undef pcmcia_request_io
#define pcmcia_request_io(handle, req) \
	CardServices(RequestIO, handle, req)

#undef pcmcia_request_irq
#define pcmcia_request_irq(handle, req) \
	CardServices(RequestIRQ, handle, req)

#undef pcmcia_request_configuration
#define pcmcia_request_configuration(handle, req) \
	CardServices(RequestConfiguration, handle, req)

#undef pcmcia_release_configuration
#define pcmcia_release_configuration(handle) \
	CardServices(ReleaseConfiguration, handle)

#undef pcmcia_release_io
#define pcmcia_release_io(handle, req) \
	CardServices(ReleaseIO, handle, req)

#undef pcmcia_release_irq
#define pcmcia_release_irq(handle, req) \
	CardServices(ReleaseIRQ, handle, req)

#undef pcmcia_release_window
#define pcmcia_release_window(win) \
	CardServices(ReleaseWindow, win)

#undef pcmcia_get_card_services_info
#define pcmcia_get_card_services_info(info) \
	CardServices(GetCardServicesInfo, info)

#undef pcmcia_reset_card
#define pcmcia_reset_card(handle, err) \
	CardServices(ResetCard, handle, err)

#undef pcmcia_validate_cis
#define pcmcia_validate_cis(handle, reg) \
	CardServices(ValidateCIS, handle, reg)
#endif /* __IN_PCMCIA_PACKAGE__ */


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,70)

/* Old style pcmcia registration */

struct device_driver {
	char			name[DEV_NAME_LEN];
};

struct pcmcia_driver {
	dev_link_t		*(*attach)(void);
	void			(*detach)(dev_link_t *);
	struct module		*owner;
	struct device_driver	drv;
};

static int __init pcmcia_register_driver(struct pcmcia_driver *driver)
{
	servinfo_t serv;

	CardServices(GetCardServicesInfo, &serv);
	if (serv.Revision != CS_RELEASE_CODE) {
		printk(KERN_NOTICE "orinoco_cs: Card Services release "
		       "does not match!\n");
		return -1;
	}

	register_pccard_driver((dev_info_t *) &driver->drv.name,
			       driver->attach, driver->detach);
	return 0;
}


static void __exit pcmcia_unregister_driver(struct pcmcia_driver *driver)
{
	unregister_pccard_driver((dev_info_t *) &driver->drv.name);
}

static void
cs_error(client_handle_t handle, int func, int ret)
{
	error_info_t err = { func, ret };
	CardServices(ReportError, handle, &err);
}

#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,5,70) */

#endif /* _LINUX_DS_H */

#endif /* __ORINOCO_KCOMPAT_H */
